/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CProtMpiSplited.cpp
 *       MPI based communication using the split module (CProtMpiSplitModule.cpp).
 *       MPI_COMM_WORLD is split into multiple disjunct sets of processes, set number 0
 *       is used for the application, the remaining sets are used for levels of tool
 *       processes. This protocol establishes a MPI call based communication between
 *       these process sets.
 *
 * @author Tobias Hilbrich
 * @date 28.07.2009
 */

#include <assert.h>
#include <stdio.h>
#include <pnmpimod.h>
#include <algorithm>
#include <mpi.h>
#include "CProtMpiSplited.h"
#include "GtiTraceDefines.h"
#include "GtiMacros.h"

using namespace gti;

/*Name of split module @todo move to some common place, currently present twice*/
#define GTI_SPLIT_MODULE_NAME "split_processes"
#define TAG_1 				   666
#define TAG_2 				   777

/**
 * Global variables for MPI state.
 */
static bool gMpiIsInitialized = false;
static bool gMpiIsFinalized = false;

//=============================
// Basic Module functions
//=============================
mGET_INSTANCE_FUNCTION (CProtMPISplited);
mFREE_INSTANCE_FUNCTION (CProtMPISplited);
mPNMPI_REGISTRATIONPOINT_FUNCTION(CProtMPISplited);

//=============================
// MPI_Init
//=============================
int MPI_Init (int *pArgc, char ***pArgv)
{
	int ret;
	ret = PMPI_Init (pArgc, pArgv);

	std::map <std::string, CProtMPISplited*> instances = CProtMPISplited::getActiveInstances ();
	std::map <std::string, CProtMPISplited*>::iterator i;

	//save state of MPI
	gMpiIsInitialized = true;

	//notify all existing communication strategies
	for (i = instances.begin(); i != instances.end(); i++)
	{
		CProtMPISplited* prot = i->second;

		if (prot)
			prot->notifyMpiInit();
	}

	return ret;
}

//=============================
// MPI_Finalize
//=============================
int MPI_Finalize (void)
{
	int ret;

	//notify all existing communication strategies
	std::map <std::string, CProtMPISplited*> instances = CProtMPISplited::getActiveInstances ();
	std::map <std::string, CProtMPISplited*>::iterator i;

	for (i = instances.begin(); i != instances.end(); i++)
	{
		CProtMPISplited* prot = i->second;
		if (prot)
			prot->notifyMpiFinalize();
	}

	ret = PMPI_Finalize ();

	//save state of MPI
	gMpiIsFinalized = true;

	return ret;
}


//=============================
// CProtMPISplitedRequest
//=============================
CProtMPISplitedRequest::CProtMPISplitedRequest (unsigned int id, MPI_Request mpi_request, unsigned long num_bytes, unsigned long channel, bool isRecv)
	: id(id), mpi_request(mpi_request), num_bytes (num_bytes), channel(channel), isRecv(isRecv)
{
	//Nothing else to do
}

//=============================
// CProtMPISplited
//=============================
CProtMPISplited::CProtMPISplited (const char* instanceName)
: ModuleBase<CProtMPISplited, I_CommProtocol> (instanceName),
  myNextRequestID (0),
  myInterComm (MPI_COMM_WORLD),
  myIsConnected (false),
  myWasShutdownCalled (false),
  myRequests (),
  myMapRankToChannel (),
  myPartnerIdsOTF (),
  myPartnerRanks (),
  myIsTop (false),
  myNextRoundRobin(0)
{
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //Needs no sub modules
    assert (subModInstances.size() == 0);

    //Startup if necessary
    //   If tracable done as part of setupTracing
    //   Otherwise done right now
    if (!this->isTraceable())
        startup ();
}

//=============================
// ~CProtMPISplited
//=============================
CProtMPISplited::~CProtMPISplited (void)
{
    OTF_Tracer::writeEnter (COMM_PROTOCOL__Destructor);

    //nothing to do ...

    OTF_Tracer::writeLeave (COMM_PROTOCOL__Destructor);
}

//=============================
// isTraceable
//=============================
bool CProtMPISplited::isTraceable (void)
{
    return true;
}

//=============================
// setupTracing
//=============================
GTI_RETURN CProtMPISplited::setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData)
{
    if (init == GTI_TRACE_INIT_COPY)
    {
        OTF_Tracer::setup ((OTF_Tracer*) pSetupData);
    }
    else if (init == GTI_TRACE_INIT_CHILD)
    {
        uint32_t    parent = *((uint32_t*)pSetupData),
                    proc_id = parent;

        //Configuration data
        std::map<std::string, std::string> data = getData();
        if (data.find("otf_proc_offset") != data.end ())
            proc_id += atoi(data.find("otf_proc_offset")->second.c_str());
        else
            proc_id += 1; //TODO should warn or abort ?

        OTF_Tracer::setup (proc_id, parent, "cprot_mpi_splitted");
    }
    else
    {
        /*This is not a root module*/
        assert (0);
    }

    //Setup Functions
    OTF_Tracer::defFunction (COMM_PROTOCOL__setupTracing, "setupTracing", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (COMM_PROTOCOL__Destructor, "CommProtocol Destructor", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (COMM_PROTOCOL__startup, "startup", FUNCTION_GROUP_ID__PROCESSING);

    OTF_Tracer::defFunction (I_COMM_PROTOCOL__shutdown, "shutdown", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__getNumChannels, "getNumChannels", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__ssend, "ssend", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__isend, "isend", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__recv, "recv", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__irecv, "irecv", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__test_msg, "test_msg", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__wait_msg, "wait_msg", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__removeOutstandingRequests, "removeOutstandingRequests", FUNCTION_GROUP_ID__COMMUNICATE);

    OTF_Tracer::writeEnter (COMM_PROTOCOL__setupTracing);

    //Setup sub-modules
    //none ...

    //Start-up deferred (to initialize tracing first)
    if (this->isTraceable())
        startup ();

    OTF_Tracer::writeLeave (COMM_PROTOCOL__setupTracing);
    return GTI_SUCCESS;
}

//=============================
// isConnected
//=============================
bool CProtMPISplited::isConnected (void)
{
	return myIsConnected && isInitialized() && !isFinalized();
}

//=============================
// isInitialized
//=============================
bool CProtMPISplited::isInitialized (void)
{
	return gMpiIsInitialized;
}

//=============================
// isFinalized
//=============================
bool CProtMPISplited::isFinalized (void)
{
	return gMpiIsFinalized;
}

//=============================
// notifyMpiInit
//=============================
void CProtMPISplited::notifyMpiInit (void)
{
	if (!myIsConnected && !myWasShutdownCalled)
		startup ();
}

//=============================
// notifyMpiFinalize
//=============================
void CProtMPISplited::notifyMpiFinalize (void)
{
	if (myIsConnected)
		shutdown ();
}

//=============================
// startup
//=============================
GTI_RETURN CProtMPISplited::startup (void)
{
    OTF_Tracer::writeEnter (COMM_PROTOCOL__startup);

    unsigned int        commId;
    char                commSide;
    int 				err;
    int 				this_set_size,
						other_set_size,
						other_set_start_rank;
    MPI_Comm 			this_set_comm;

    // === (0) Check whether MPI is initialized
    if (!gMpiIsInitialized)
    {
    	OTF_Tracer::writeLeave (COMM_PROTOCOL__startup);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    // === Check whether already connected
    if (myIsConnected)
    	return GTI_SUCCESS;

    // === (1) Read data ===
    //General protocol data
    std::map<std::string, std::string> data = getData ();
    std::map<std::string, std::string>::iterator i;

    i=data.find("comm_id");
    assert (i != data.end ());
    commId = atoi (i->second.c_str());

    i=data.find("side");
    assert (i != data.end ());
    assert (i->second.length() == 1); //just one letter !
    commSide = i->second.c_str()[0];
    assert ((commSide == 't') || (commSide == 'b')); //either "t" or "b"

    // === (2) Query for services of the split module ===
    PNMPI_modHandle_t handle;
    PNMPI_Service_descriptor_t service;
    PNMPI_Service_Fct_t fct;

#ifdef PNMPI_FIXED
    err = PNMPI_Service_GetModuleByName(GTI_SPLIT_MODULE_NAME, &handle);
#else
    char string[512];
    sprintf (string, "%s",GTI_SPLIT_MODULE_NAME);
    err = PNMPI_Service_GetModuleByName(string, &handle);
#endif
    if (err != PNMPI_SUCCESS)
    {
    	std::cerr
    	<< "Failed to get a handle for the P^nMPI module \""
    	<< GTI_SPLIT_MODULE_NAME
    	<< "\""
    	<< std::endl
    	<< "(Failed in module \""
    	<< getName() << "::" << __FUNCTION__
    	<< "\")"
    	<< std::endl;
    	assert (0);
    }

    err = PNMPI_Service_GetServiceByName(handle, "SplitMod_getMySetSize", "p", &service);
    assert (err == PNMPI_SUCCESS);
    ((int(*)(int*)) service.fct) (&this_set_size);

    err = PNMPI_Service_GetServiceByName(handle, "SplitMod_getMySetComm", "p", &service);
    assert (err == PNMPI_SUCCESS);
    ((int(*)(void*)) service.fct) (&this_set_comm);

    err = PNMPI_Service_GetServiceByName(handle, "SplitMod_getSetInfo", "ipp", &service);
    assert (err == PNMPI_SUCCESS);
    ((int(*)(int, int*, int*)) service.fct) (commId,&other_set_size,&other_set_start_rank);

    // === (3) Calculate ranks of comm. partner processes ===
    int rank;
    int num_partners, remaining, start_index;

    if (commSide == 't')
    {
    		//Top side partner calculation
    		myIsTop = true;
		err = PMPI_Comm_rank (this_set_comm, &rank);
		assert (err == MPI_SUCCESS);

		num_partners = other_set_size / this_set_size;
		remaining = other_set_size - num_partners * this_set_size;

		start_index = /*other_set_start_rank*/0 + rank * num_partners + std::min(remaining,rank);

		if (rank < remaining) //does this rank get one of the remaining processes ?
			num_partners++;

		myPartnerRanks.resize(num_partners);
		for (int i = 0; i < num_partners; i++)
		{
			myPartnerRanks[i] = start_index + i; /**< Ranks of comm. partners. */
			myMapRankToChannel.insert (std::make_pair(start_index + i, i));
		}

		err = PMPI_Intercomm_create (this_set_comm, 0, MPI_COMM_WORLD, other_set_start_rank, TAG_2, &myInterComm);
		assert (err == MPI_SUCCESS);
    }
    else
    {
    	//bottom side partner calculation
    	err = PMPI_Comm_rank (this_set_comm, &rank);
    	assert (err == MPI_SUCCESS);

    	num_partners = this_set_size / other_set_size;
    	remaining = this_set_size - num_partners * other_set_size;

    	myPartnerRanks.resize(1);
    	for (int i = 1; i < other_set_size + 2; i++)
    	{
    		if (rank < i * num_partners + std::min(remaining,i))
    		{
    			myPartnerRanks[0] = /*other_set_start_rank*/0 + i - 1;
    			myMapRankToChannel.insert (std::make_pair(myPartnerRanks[0], 0));
    			break;
    		}
    	}

    	PMPI_Intercomm_create (this_set_comm, 0, MPI_COMM_WORLD, other_set_start_rank, TAG_2, &myInterComm);
    	assert (err == MPI_SUCCESS);
    }

    //DEBUG START
/*
    int myrank;
    err = PMPI_Comm_rank(MPI_COMM_WORLD,&myrank);
    assert (err == MPI_SUCCESS);
    std::cout << "I am rank: " << myrank
    << "My Partners are: {";
    for (int i = 0; i < myPartnerRanks.size(); i++)
    {
    	std::cout << myPartnerRanks[i] << ", ";
    }
    std::cout << "}" << std::endl;
*/
    //DEBUG END

    // === (4) Get OTF process ids from partners ===
    myPartnerIdsOTF.resize (myPartnerRanks.size());
    for (int i = 0; i < myPartnerRanks.size(); i++)
    {
    	int otf_id = OTF_Tracer::getProcessId (),
			other_id;
    	MPI_Request request;
    	MPI_Status status;

    	err = PMPI_Isend (&otf_id, 1, MPI_INT, myPartnerRanks[i], TAG_1, myInterComm, &request);
    	assert (err == MPI_SUCCESS);
    	err = PMPI_Recv (&other_id, 1, MPI_INT, myPartnerRanks[i], TAG_1, myInterComm, &status);
    	assert (err == MPI_SUCCESS);
    	err = PMPI_Wait (&request,&status);
    	assert (err == MPI_SUCCESS);

    	myPartnerIdsOTF[i] = other_id;
    }

    myIsConnected = true;

    OTF_Tracer::writeLeave (COMM_PROTOCOL__startup);
    return GTI_SUCCESS;
}

//=============================
// shutdown
//=============================
GTI_RETURN CProtMPISplited::shutdown (void)
{
    myWasShutdownCalled = true;

    //Does a connection exists ?
    if (!myIsConnected)
    	return GTI_SUCCESS;

    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__shutdown);

    /*Take care of any outstanding requests*/
    int err = PMPI_Comm_free(&myInterComm);
    assert (err == MPI_SUCCESS);

    std::map<unsigned int, CProtMPISplitedRequest >::iterator iter;
    for (iter = myRequests.begin(); iter != myRequests.end(); iter++)
    {
    	err = PMPI_Cancel(&(iter->second.mpi_request));
    	assert (err == MPI_SUCCESS);
    	err = PMPI_Request_free(&(iter->second.mpi_request));
    	assert (err == MPI_SUCCESS);
    }
    myRequests.clear();

    //Set as unconnected
    myIsConnected = false;

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__shutdown);
    return GTI_SUCCESS;
}

//=============================
// removeOutstandingRequests
//=============================
GTI_RETURN CProtMPISplited::removeOutstandingRequests (void)
{
	int err;

	if (!myIsConnected)
		return GTI_SUCCESS;

	OTF_Tracer::writeEnter (I_COMM_PROTOCOL__removeOutstandingRequests);

	std::map<unsigned int, CProtMPISplitedRequest >::iterator iter;
	for (iter = myRequests.begin(); iter != myRequests.end(); iter++)
	{
		err = PMPI_Cancel(&(iter->second.mpi_request));
		assert (err == MPI_SUCCESS);
		err = PMPI_Request_free(&(iter->second.mpi_request));
		assert (err == MPI_SUCCESS);
	}
	myRequests.clear();

	/*TODO: This is MPI-2, we need to ifdef here for an MPI-1 alternative.*/
#ifndef GTI_NO_MPI2
	err = PMPI_Barrier (myInterComm);
	assert (err == MPI_SUCCESS);
#else
	for (int i = 0; i < myPartnerRanks.size(); i++)
	{
		MPI_Status status;
		if (myIsTop)
			PMPI_Recv (NULL, 0, MPI_INT, myPartnerRanks[i], 12345, myInterComm, &status);
		else
			PMPI_Ssend (NULL, 0, MPI_INT, myPartnerRanks[i], 12345, myInterComm);
	}
#endif

	OTF_Tracer::writeLeave (I_COMM_PROTOCOL__removeOutstandingRequests);

    return GTI_SUCCESS;
}

//=============================
// getNumChannels
//=============================
GTI_RETURN CProtMPISplited::getNumChannels (
        unsigned long* out_numChannels)
{
	if (!myIsConnected)
		return GTI_ERROR_NOT_INITIALIZED;

    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__getNumChannels);

    if (out_numChannels)
    	*out_numChannels = myPartnerRanks.size();

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__getNumChannels);
    return GTI_SUCCESS;
}

//=============================
// ssend
//=============================
GTI_RETURN CProtMPISplited::ssend (
        void* buf,
        unsigned long num_bytes,
        unsigned long channel
        )
{
	int err;

	if (!myIsConnected)
		return GTI_ERROR_NOT_INITIALIZED;

    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__ssend);

    //Sanity
    assert (channel >= 0 && channel < myPartnerRanks.size());

    //OTF Send event
    OTF_Tracer::writeSendMsg (myPartnerIdsOTF[channel], 0, num_bytes);

    //Send!
    err = PMPI_Ssend (buf, num_bytes, MPI_BYTE, myPartnerRanks[channel], TAG_1, myInterComm);
    assert (err == MPI_SUCCESS);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__ssend);
    return GTI_SUCCESS;
}

//=============================
// isend
//=============================
GTI_RETURN CProtMPISplited::isend (
        void* buf,
        unsigned long num_bytes,
        unsigned int *out_request,
        unsigned long channel
        )
{
	if (!myIsConnected)
		return GTI_ERROR_NOT_INITIALIZED;

    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__isend);

    MPI_Request mpi_request;
    int err;

    //Sanity
    assert (channel >= 0 && channel < myPartnerRanks.size());

    //OTF Send event
    OTF_Tracer::writeSendMsg (myPartnerIdsOTF[channel], 0, num_bytes);

    //ISend
    err = PMPI_Isend (buf, num_bytes, MPI_BYTE, myPartnerRanks[channel], TAG_1, myInterComm, &mpi_request);
    assert (err == MPI_SUCCESS);

    //Store and return request (map MPI request to an int)
    myRequests.insert (std::make_pair(
        	myNextRequestID,
        	CProtMPISplitedRequest(myNextRequestID, mpi_request, num_bytes, channel, false)));
    *out_request = myNextRequestID;
    myNextRequestID++;

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__isend);
    return GTI_SUCCESS;
}

//=============================
// recv
//=============================
GTI_RETURN CProtMPISplited::recv (
        void* out_buf,
        unsigned long num_bytes,
        unsigned long* out_length,
        unsigned long channel,
        unsigned long *out_channel
        )
{
	if (!myIsConnected)
		return GTI_ERROR_NOT_INITIALIZED;

    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__recv);

    MPI_Status status;
    int err;

    if (channel != RECV_ANY_CHANNEL)
    	assert (channel >= 0 && channel < myPartnerRanks.size());

    //do the receive
    if (channel == RECV_ANY_CHANNEL)
    {
    		//To avoid starvation of some ranks in a scenario where the receiver side is a bottleneck and
    	    	//the connected processes overfood this process, we use an MPI_IProbe in a round robin fashion
    	    	//to avoid always receiving from the same rank
    	    	MPI_Status probeStatus;
    	    	int probeFlag;
    	    	err = PMPI_Iprobe (myPartnerRanks[myNextRoundRobin], TAG_1, myInterComm, &probeFlag, &probeStatus);

    	    	int sourceToUse = MPI_ANY_SOURCE;
   	    	if (probeFlag)
   	    		sourceToUse = myPartnerRanks[myNextRoundRobin];
   	    	myNextRoundRobin=(myNextRoundRobin+1)%myPartnerRanks.size();

    		err = PMPI_Recv (out_buf, num_bytes, MPI_BYTE, sourceToUse, TAG_1, myInterComm, &status);
		assert (err == MPI_SUCCESS);
    }
    else
    {
    	err = PMPI_Recv (out_buf, num_bytes, MPI_BYTE, myPartnerRanks[channel], TAG_1, myInterComm, &status);
    	assert (err == MPI_SUCCESS);
    }

    //save the used channel and the message length
    unsigned long used_channel = channel;
    if (used_channel == RECV_ANY_CHANNEL)
    {
		std::map<int, int>::iterator rank_iter = myMapRankToChannel.find(status.MPI_SOURCE);
		assert (rank_iter != myMapRankToChannel.end());
		used_channel = rank_iter->second;
    }

    if (out_channel)
    	*out_channel = used_channel;

    int count;
    err = PMPI_Get_count (&status, MPI_BYTE, &count);
    assert (err == MPI_SUCCESS);

    if (out_length)
    	*out_length = count;

    //Write OTF receive event
    OTF_Tracer::writeRecvMsg (myPartnerIdsOTF[used_channel], 0, count);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__recv);
    return GTI_SUCCESS;
}

//=============================
// irecv
//=============================
GTI_RETURN CProtMPISplited::irecv (
        void* out_buf,
        unsigned long num_bytes,
        unsigned int *out_request,
        unsigned long channel
        )
{
	if (!myIsConnected)
		return GTI_ERROR_NOT_INITIALIZED;

	OTF_Tracer::writeEnter (I_COMM_PROTOCOL__irecv);

    MPI_Request request;
    int err;

    if (channel != RECV_ANY_CHANNEL)
    	assert (channel >= 0 && channel < myPartnerRanks.size());

    //do the receive
    if (channel == RECV_ANY_CHANNEL)
    {
    		//To avoid starvation of some ranks in a scenario where the receiver side is a bottleneck and
    	    //the connected processes overfood this process, we use an MPI_IProbe in a round robin fashion
    	    //to avoid always receiving from the same rank
    	    MPI_Status probeStatus;
    	    int probeFlag;
    	    err = PMPI_Iprobe (myPartnerRanks[myNextRoundRobin], TAG_1, myInterComm, &probeFlag, &probeStatus);

    	    int sourceToUse = MPI_ANY_SOURCE;
    	    if (probeFlag)
    	    		sourceToUse = myPartnerRanks[myNextRoundRobin];
    	    myNextRoundRobin=(myNextRoundRobin+1)%myPartnerRanks.size();

		err = PMPI_Irecv (out_buf, num_bytes, MPI_BYTE, sourceToUse, TAG_1, myInterComm, &request);
		assert (err == MPI_SUCCESS);
    }
    else
    {
			err = PMPI_Irecv (out_buf, num_bytes, MPI_BYTE, myPartnerRanks[channel], TAG_1, myInterComm, &request);
			assert (err == MPI_SUCCESS);
    }

    //Store and return request (map MPI request to an int)
    myRequests.insert (std::make_pair(
            	myNextRequestID,
            	CProtMPISplitedRequest(myNextRequestID, request, num_bytes, channel, true)));
    if (out_request)
    	*out_request = myNextRequestID;
    myNextRequestID++;

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__irecv);
    return GTI_SUCCESS;
}

//=============================
// test_msg
//=============================
GTI_RETURN CProtMPISplited::test_msg (
        unsigned int request,
        int* out_completed,
        unsigned long* out_receive_length,
        unsigned long* out_channel
        )
{
	if (!myIsConnected)
		return GTI_ERROR_NOT_INITIALIZED;

    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__test_msg);

    int err;
    std::map<unsigned int, CProtMPISplitedRequest >::iterator iter;
    iter = myRequests.find(request);

	assert (iter != myRequests.end()); //has to be valid

    MPI_Status status;
    int flag;
    err = PMPI_Test (&(iter->second.mpi_request), &flag, &status);
    assert (err == MPI_SUCCESS);

    if (flag)
    {
    	if (out_completed)
    		*out_completed = 1;

    	//save the used channel and the message length
    	int count;
    	unsigned long channel_used;

    	if (iter->second.isRecv)
    	{
    		std::map<int, int>::iterator rank_iter;
    		rank_iter = myMapRankToChannel.find(status.MPI_SOURCE);
			assert (rank_iter != myMapRankToChannel.end());
			channel_used = rank_iter->second;

			err = PMPI_Get_count (&status, MPI_BYTE, &count);
    		assert (err == MPI_SUCCESS);
    	}
    	else
    	{
    		channel_used = iter->second.channel;
    		count = iter->second.num_bytes;
    	}

    	if (out_channel)
    		*out_channel = channel_used;

    	if (out_receive_length)
    		*out_receive_length = count;

    	//Write OTF receive event
    	if (iter->second.isRecv)
    	{
    		OTF_Tracer::writeRecvMsg (myPartnerIdsOTF[channel_used], 0, count);
    	}

    	myRequests.erase (iter);
    }
    else
    {
    	if (out_completed)
    		*out_completed = 0;
    }

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__test_msg);
    return GTI_SUCCESS;
}

//=============================
// wait_msg
//=============================
GTI_RETURN CProtMPISplited::wait_msg (
        unsigned int request,
        unsigned long* out_receive_length,
        unsigned long* out_channel
        )
{
	if (!myIsConnected)
		return GTI_ERROR_NOT_INITIALIZED;

    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__wait_msg);

    std::map<unsigned int, CProtMPISplitedRequest >::iterator iter;
    iter = myRequests.find(request);

    assert (iter != myRequests.end()); //has to be valid

    MPI_Status status;
    int err;

    err = PMPI_Wait (&(iter->second.mpi_request), &status);
    assert (err == MPI_SUCCESS);

   	//save the used channel and the message length
    int count;
    unsigned long channel_used;

    if (iter->second.isRecv)
    {
    	std::map<int, int>::iterator rank_iter;
    	rank_iter = myMapRankToChannel.find(status.MPI_SOURCE);
    	assert (rank_iter != myMapRankToChannel.end());
    	channel_used = rank_iter->second;

    	err = PMPI_Get_count (&status, MPI_BYTE, &count);
    	assert (err == MPI_SUCCESS);
    }
    else
    {
    	channel_used = iter->second.channel;
    	count = iter->second.num_bytes;
    }

    if (out_channel)
    	*out_channel = channel_used;

    if (out_receive_length)
    	*out_receive_length = count;

   	//Write OTF receive event
    if (iter->second.isRecv)
    {
    	OTF_Tracer::writeRecvMsg (myPartnerIdsOTF[channel_used], 0, count);
    }

    myRequests.erase (iter);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__wait_msg);
    return GTI_SUCCESS;
}

/*EOF*/
