/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CProtSimpleTcp.cpp
 *       Simple TCP based implementation of the communication
 *       protocol interface.
 *
 * @author Tobias Hilbrich
 * @date 21.04.2009
 */

#include <assert.h>
#include <stdio.h>
#include <pnmpimod.h>
#include "CProtSimpleTcp.h"
#include "GtiTraceDefines.h"
#include "GtiMacros.h"

#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

using namespace gti;

mGET_INSTANCE_FUNCTION(CProtSimpleTCP);
mFREE_INSTANCE_FUNCTION(CProtSimpleTCP);
mPNMPI_REGISTRATIONPOINT_FUNCTION(CProtSimpleTCP);

const unsigned long CProtSimpleTCP::TOKEN_MSG_SIZE_HEADER = 0xFFFFFFFF;
const unsigned long CProtSimpleTCP::TOKEN_SYNC_MSG        = 0xFFFFFFFE;

/**
 * Function to perform a blocking receive with sockets.
 */
int blocking_recv(int socket, void* buffer, size_t bytes)
{
    int retBytes;
    do
    {
        retBytes = ::recv (socket,buffer,bytes, MSG_WAITALL);

        if (retBytes < 0)
        {
            int myErrno = errno;
            printf ("CRITICAL: blocking_recv failed with errno=%d\n",myErrno);
            printf ("          Error: %s\n",strerror(myErrno));
            assert (0);
        }
        else
        {
            usleep(1); //TODO specifiy something more meaningful
        }
    }
    while (retBytes == 0);

    return retBytes;
}

/**
 * Function to perform a blocking send with sockets.
 */
int all_send(int socket, void* buffer, size_t bytes)
{
    int sentBytes;
    do
    {
        sentBytes = ::send (socket,buffer,bytes, 0);

        if (sentBytes < 0)
        {
            int myErrno = errno;
            printf ("CRITICAL: all_send failed with errno=%d\n",myErrno);
            printf ("          Error: %s\n",strerror(myErrno));
            assert (0);
        }
        else
        {
            usleep(1); //TODO specifiy something more meaningful
        }
    }
    while (sentBytes == 0);

    return sentBytes;
}

CProtSimpleTCP::CProtSimpleTCP (const char* instanceName)
: ModuleBase<CProtSimpleTCP, I_CommProtocol> (instanceName),
  myIsShutdown (false)
{
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //A Comm Protocol likely needs no sub modules
    assert (subModInstances.size() == 0);

    //Init data
    myNumSockets = 0;
    mySockets = NULL;
    mySocketIDsOTF = NULL;
    myOutstandingMsgs.clear();
    myActiveRequests.clear();
    myCompletedRequests.clear();

    //Startup if necessary
    if (!this->isTraceable())
    {
    		GTI_RETURN ret = startup ();
        assert (ret == GTI_SUCCESS);
    }
}

CProtSimpleTCP::~CProtSimpleTCP (void)
{
    OTF_Tracer::writeEnter (COMM_PROTOCOL__Destructor);

    //Closing of sockets is done in shutdown !

    if (mySockets)
        delete mySockets;
    mySockets = NULL;

    if (mySocketIDsOTF)
        delete mySocketIDsOTF;
    mySocketIDsOTF = NULL;

    OTF_Tracer::writeLeave (COMM_PROTOCOL__Destructor);
}

bool CProtSimpleTCP::isTraceable (void)
{
    return true;
}

GTI_RETURN CProtSimpleTCP::setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData)
{
    if (init == GTI_TRACE_INIT_COPY)
    {
        OTF_Tracer::setup ((OTF_Tracer*) pSetupData);
    }
    else if (init == GTI_TRACE_INIT_CHILD)
    {
        uint32_t    parent = *((uint32_t*)pSetupData),
                    proc_id = parent;

        //Configuration data
        std::map<std::string, std::string> data = getData();
        if (data.find("otf_proc_offset") != data.end ())
            proc_id += atoi(data.find("otf_proc_offset")->second.c_str());
        else
            proc_id += 1; //TODO should warn or abort ?

        OTF_Tracer::setup (proc_id, parent, "commProtocolTemplate");
    }
    else
    {
        /*This is not a root module*/
        assert (0);
    }

    //Setup Functions
    OTF_Tracer::defFunction (COMM_PROTOCOL__setupTracing, "setupTracing", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (COMM_PROTOCOL__Destructor, "~CommProtocolTemplate", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (COMM_PROTOCOL__startup, "startup", FUNCTION_GROUP_ID__PROCESSING);

    OTF_Tracer::defFunction (I_COMM_PROTOCOL__shutdown, "shutdown", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__getNumChannels, "getNumChannels", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__ssend, "ssend", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__isend, "isend", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__recv, "recv", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__irecv, "irecv", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__test_msg, "test_msg", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__wait_msg, "wait_msg", FUNCTION_GROUP_ID__COMMUNICATE);

    OTF_Tracer::writeEnter (COMM_PROTOCOL__setupTracing);

    //Setup sub-modules
    //none ...

    //Start-up deferred (to initialize tracing first)
    if (this->isTraceable())
    {
    		GTI_RETURN ret = startup ();
        assert (ret == GTI_SUCCESS);
    }

    OTF_Tracer::writeLeave (COMM_PROTOCOL__setupTracing);
    return GTI_SUCCESS;
}

bool CProtSimpleTCP::isConnected (void)
{
	return isInitialized() && !isFinalized();
}

bool CProtSimpleTCP::isInitialized (void)
{
	/*Constructor either succeeds in connecting or dies with an assert*/
	return true;
}

bool CProtSimpleTCP::isFinalized (void)
{
	return myIsShutdown;
}

GTI_RETURN CProtSimpleTCP::startup (void)
{
    OTF_Tracer::writeEnter (COMM_PROTOCOL__startup);

    unsigned int        commId;
    char                commSide;
    unsigned long       tierSize;
    unsigned long       selfID;
    unsigned short int  serverPort;
    struct sockaddr_in  serverAddr;
    int                 serverSocket;
    static bool isConnected = false;

    if (isConnected)
    		return GTI_SUCCESS;
    isConnected = true;
    int ret;

    assert (!myIsShutdown);

    // === (1) Read data ===
    //General protocol data
    std::map<std::string, std::string> data = getData ();
    std::map<std::string, std::string>::iterator i;

    i=data.find("comm_id");
    assert (i != data.end ());
    commId = atoi (i->second.c_str());

    i=data.find("side");
    assert (i != data.end ());
    assert (i->second.length() == 1); //just one letter !
    commSide = i->second.c_str()[0];
    assert ((commSide == 't') || (commSide == 'b')); //either "t" or "b"

    i=data.find("tier_size");
    assert (i != data.end ());
    tierSize = atol (i->second.c_str());

    i=data.find("id");
    assert (i != data.end ());
    selfID = atol (i->second.c_str());

    //TCP specific stuff
    i=data.find("server_port");
    assert (i != data.end ());
    serverPort = atoi (i->second.c_str());

    i=data.find("server_ip");
    assert (i != data.end ());
    memset(&serverAddr, 0, sizeof(serverAddr));
    inet_aton(i->second.c_str(), &(serverAddr.sin_addr));
    serverAddr.sin_port = serverPort;
    serverAddr.sin_family = AF_INET;

    // === (2) Connect to server ===
    serverSocket = ::socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    assert (serverSocket >= 0);

    ret = ::connect(serverSocket,
                    (struct sockaddr *) &serverAddr,
                    sizeof(serverAddr)
                   );
    assert (ret >= 0);

    // === (3) Send own connection info to server ===
    // We send:
    //   commId, commSide, tierSize, selfID, [OTF_ProcID]
    unsigned long int connectionInfoBuffer[4];
    unsigned int len;

    connectionInfoBuffer[0] = (unsigned long int) commId;
    if (commSide == 't')
        connectionInfoBuffer[1] = (unsigned long int) 0;
    else
        connectionInfoBuffer[1] = (unsigned long int) 1;
    connectionInfoBuffer[2] = (unsigned long int) tierSize;
    connectionInfoBuffer[3] = (unsigned long int) selfID;

    len = sizeof(unsigned long int) * 4;

    ret = ::all_send(serverSocket, connectionInfoBuffer, len);
    assert (ret == len);

    // === (4) Receive partner bundle from server ===
    int bytes = 0;

    //Receive num bundles
    bytes = ::blocking_recv(serverSocket, connectionInfoBuffer, sizeof(unsigned long int));
    assert (bytes == sizeof (unsigned long int));
    assert (connectionInfoBuffer[0] >= 0); //At least one partner needed !

    //Receive bundle contents
    len = sizeof (unsigned long) * connectionInfoBuffer[0] * 2;
    unsigned long *connectionPartners = new unsigned long[connectionInfoBuffer[0]*2];

    bytes = ::blocking_recv(serverSocket, connectionPartners, len);
    assert (bytes == len);

    // === (5) Open connection to partners ===
    myNumSockets = connectionInfoBuffer[0];
    myOutstandingMsgs.resize (myNumSockets);
    myActiveRequests.resize (myNumSockets+1);

    assert (!mySockets && !mySocketIDsOTF);
    mySockets = new int[myNumSockets];
    mySocketIDsOTF = new unsigned int[myNumSockets];

    for (unsigned int i = 0; i < myNumSockets; i++)
    {
        struct sockaddr_in  partnerAddr;
        memset(&partnerAddr, 0, sizeof(partnerAddr));
        partnerAddr.sin_addr.s_addr = connectionPartners[2*i + 0];
        partnerAddr.sin_port = connectionPartners[2*i + 1];
        partnerAddr.sin_family = AF_INET;

        printf("Establishing connection: ip=%s port=%d\n",inet_ntoa(partnerAddr.sin_addr),partnerAddr.sin_port);

        mySockets[i] = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
        assert (mySockets[i] >= 0);

        if (commSide == 'b')
        {
            int listenSocket = mySockets[i];

            //only one sock on bottom side
            assert (myNumSockets == 1);

            //Bottom side listens
            ret = bind(
                    listenSocket,
                    (struct sockaddr*) &partnerAddr,
                    sizeof(partnerAddr)
                    );
            assert(ret  >= 0);

            ret = listen(listenSocket, 1);
            assert (ret >= 0);

            //Send "I-am-listing" to the server
            bytes = ::all_send(serverSocket, connectionInfoBuffer, sizeof(unsigned long int));
            assert (bytes == sizeof(unsigned long int));

            unsigned int addrLen = sizeof(partnerAddr);
            mySockets[i] = accept(listenSocket, (struct sockaddr *) &partnerAddr, &addrLen);
            assert (mySockets[i] >= 0);

            close (listenSocket);
        }
        else
        {
            //wait for a "good-to-connect" from the server
            bytes = ::blocking_recv(serverSocket, connectionInfoBuffer, sizeof(unsigned long int));
            assert (bytes == sizeof(unsigned long int));

            //Top side connects
            ret = connect(mySockets[i],
                    (struct sockaddr *) &partnerAddr,
                    sizeof(partnerAddr)
                   );
            assert (ret >= 0);
        }
    }
    delete (connectionPartners);

    // === (6) Send a done connecting to the server ===
    len = sizeof (unsigned long);
    ret = all_send(serverSocket, connectionInfoBuffer, len);
    assert (ret == len);

    // === (7) Close connection to server ===
    close (serverSocket);

    // === (8) Get OTF process ids from partners ===
    for (unsigned int i = 0; i < myNumSockets; i++)
    {
        unsigned long procID = OTF_Tracer::getProcessId ();

        if (commSide == 'b')
        {
        		ret = ::all_send(mySockets[i], &procID, sizeof(procID));
            assert (ret == sizeof(procID));
            ret = ::blocking_recv(mySockets[i],
                    &(mySocketIDsOTF[i]),
                    sizeof(unsigned long));
            assert (ret == sizeof(unsigned long));

            ////DEBUG
            //printf ("Bottom: OTF: i=%d procID=%ld",i,mySocketIDsOTF[i]);
        }
        else
        {
        		ret = ::blocking_recv(mySockets[i],
                        &(mySocketIDsOTF[i]),
                        sizeof(unsigned long));
            assert (ret == sizeof(unsigned long));
            ret = ::all_send(mySockets[i], &procID, sizeof(procID));
            assert (ret == sizeof(procID));

            ////DEBUG
            //printf ("Top: OTF: i=%d procID=%ld",i,mySocketIDsOTF[i]);
        }
    }

    OTF_Tracer::writeLeave (COMM_PROTOCOL__startup);
    return GTI_SUCCESS;
}

GTI_RETURN CProtSimpleTCP::shutdown (void)
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__shutdown);

    for (unsigned int i = 0; i < myNumSockets; i++)
    {
        if (mySockets[i] >= 0)
            close (mySockets[i]);
        mySockets[i] = -1;
    }

    myIsShutdown = true;

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__shutdown);
    return GTI_SUCCESS;
}

GTI_RETURN CProtSimpleTCP::removeOutstandingRequests (void)
{

    //num sockets +1 for any receives ...
    for (unsigned long i = 0; i <= myNumSockets; i++)
    {
        myActiveRequests[i].clear ();
    }

    return GTI_SUCCESS;
}

GTI_RETURN CProtSimpleTCP::getNumChannels (
        unsigned long* out_numChannels)
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__getNumChannels);

    *out_numChannels = myNumSockets;

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__getNumChannels);
    return GTI_SUCCESS;
}

GTI_RETURN CProtSimpleTCP::ssend (
        void* buf,
        unsigned long num_bytes,
        unsigned long channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__ssend);

    ssize_t sentBytes;
    unsigned long header[2];
    ssize_t headerLen = sizeof (unsigned long) * 2;

    assert ((channel >= 0) && (channel < myNumSockets));

    OTF_Tracer::writeSendMsg (mySocketIDsOTF[channel], 0, num_bytes);

    //Send the size of the message to partner
    header[0] = TOKEN_MSG_SIZE_HEADER;
    header[1] = num_bytes;
    sentBytes = ::all_send (mySockets[channel], &header, headerLen);
    assert (sentBytes == headerLen);

    //Send the actual message
    sentBytes = ::all_send (mySockets[channel], buf, num_bytes);
    assert (sentBytes == num_bytes);

    //Receive a reply
    while (header[0] != TOKEN_SYNC_MSG)
    {
        sentBytes = ::blocking_recv (mySockets[channel], header, headerLen);
        assert (sentBytes == headerLen);

        if (header[0] != TOKEN_SYNC_MSG)
        {
            assert (header[0] == TOKEN_MSG_SIZE_HEADER);

            //Handle receival of msg
            recv_msg_conntent (channel, header[1], NULL, NULL, NULL, NULL);
        }
    }

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__ssend);
    return GTI_SUCCESS;
}

GTI_RETURN CProtSimpleTCP::isend (
        void* buf,
        unsigned long num_bytes,
        unsigned int *out_request,
        unsigned long channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__recv);

    /*Not Supported !!*/
    assert (0);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__recv);
    return GTI_SUCCESS;
}

GTI_RETURN CProtSimpleTCP::recv (
        void* out_buf,
        unsigned long num_bytes,
        unsigned long* out_length,
        unsigned long channel,
        unsigned long *out_channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__recv);

    unsigned long header[2];
    ssize_t sentBytes;
    ssize_t headerLen = sizeof (unsigned long) * 2;

    assert ( ((channel >= 0) && (channel < myNumSockets)) ||
             (channel == RECV_ANY_CHANNEL)                  );

    //Did we already receive any messages
    unsigned long i_start = channel;
    unsigned long i_end = channel+1;
    if (channel == RECV_ANY_CHANNEL)
    {
        i_start = 0;
        i_end = myNumSockets;
        channel = myNumSockets;
    }

    for (unsigned int i = i_start;i < i_end;i++)
    {
        if (myOutstandingMsgs[i].size() > 0)
        {
            std::pair<unsigned long, void*> elem = myOutstandingMsgs[i].front();
            myOutstandingMsgs[i].pop_front();

            assert (elem.first <= num_bytes);
            *out_length = elem.first;
            if (out_channel)
                *out_channel = i;
            memcpy (out_buf,elem.second,elem.first);
            free (elem.second);

            OTF_Tracer::writeLeave (I_COMM_PROTOCOL__recv);
            return GTI_SUCCESS;
        }
    }

    //Regular receive
    bool bWasReceived = false;
    do {
        unsigned long channel_used;
        if (channel != myNumSockets)
        {
            //Receive from specified channel
            sentBytes = ::blocking_recv (mySockets[channel], header, headerLen);
            assert (sentBytes == headerLen);
            channel_used = channel;
        }
        else
        {
            //probe all channels
            bool bGotHeader = false;

            do
            {
                for (unsigned long i = 0; i < myNumSockets; i++)
                {
                    sentBytes = ::recv (mySockets[i], header, headerLen, MSG_DONTWAIT);
                    if (sentBytes < 0)
                        assert (errno == EAGAIN);
                    if (sentBytes == headerLen)
                    {
                        channel_used = i;
                        bGotHeader = true;
                        break;
                    }
                }

                if (!bGotHeader) usleep (5);
            } while (!bGotHeader);
        }
        assert (sentBytes == headerLen);
        assert (header[0] == TOKEN_MSG_SIZE_HEADER);

        recv_msg_conntent (channel_used, header[1], out_buf, &bWasReceived, NULL, NULL);
        if (bWasReceived)
        {
            assert (header[1] <= num_bytes);
            *out_length = header[1];
            if (out_channel)
                *out_channel = channel_used;
        }
    } while (!bWasReceived);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__recv);
    return GTI_SUCCESS;
}

GTI_RETURN CProtSimpleTCP::irecv (
        void* out_buf,
        unsigned long num_bytes,
        unsigned int *out_request,
        unsigned long channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__irecv);

    static unsigned int nextRequestID = 1;
    static unsigned long order = 0;

    assert ( ((channel >= 0) && (channel < myNumSockets)) ||
             (channel == RECV_ANY_CHANNEL) );

    //Did we already receive any messages
    unsigned long i_start = channel;
    unsigned long i_end = channel+1;
    if (channel == RECV_ANY_CHANNEL)
    {
        i_start = 0;
        i_end = myNumSockets;
        channel = myNumSockets;
    }

    for (unsigned int i = i_start;i < i_end;i++)
    {
        if (myOutstandingMsgs[i].size() > 0)
        {
            std::pair<unsigned long, void*> elem = myOutstandingMsgs[i].front();
            myOutstandingMsgs[i].pop_front();

            assert (elem.first <= num_bytes);
            memcpy (out_buf,elem.second,elem.first);
            free (elem.second);

            *out_request = nextRequestID;
            myCompletedRequests.insert(
                    std::make_pair(
                            nextRequestID,
                            std::make_pair (i, elem.first)));
            nextRequestID++;

            OTF_Tracer::writeLeave (I_COMM_PROTOCOL__irecv);
            return GTI_SUCCESS;
        }
    }

    myActiveRequests[channel].push_back (
            std::make_pair (
                    nextRequestID,
                    std::make_pair (
                            order,
                            std::make_pair (out_buf,num_bytes)
                            )
                    )
            );
    *out_request = nextRequestID;
    nextRequestID++;
    order++;

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__irecv);
    return GTI_SUCCESS;
}

GTI_RETURN CProtSimpleTCP::test_msg (
        unsigned int request,
        int* out_completed,
        unsigned long* out_receive_length,
        unsigned long* out_channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__test_msg);

    int bytesReceived;
    unsigned long header[2];
    unsigned long headerLen = sizeof (unsigned long) * 2;

    //(1) see whether request already in completed requests
    std::map <unsigned int, CompletionInfo>::iterator iter
        = myCompletedRequests.find(request);
    if (iter != myCompletedRequests.end())
    {
        *out_receive_length = iter->second.second;
        *out_channel = iter->second.first;
        *out_completed = 1;
        myCompletedRequests.erase (iter);
        OTF_Tracer::writeLeave (I_COMM_PROTOCOL__test_msg);
        return GTI_SUCCESS;
    }

    //(2) probe all channels and receive all available
    //    messages.
    //    Stop if either the request in question was
    //    completed, or all channels have no more messages
    for (unsigned int i = 0; i < myNumSockets; i++)
    {
        do
        {
            bytesReceived = ::recv (mySockets[i], header, headerLen, MSG_DONTWAIT);

            if (bytesReceived == headerLen)
            {
                assert (header[0] == TOKEN_MSG_SIZE_HEADER);
                unsigned int requestUsed;
                bool wasRequestUsed;

                recv_msg_conntent (i, header[1], NULL, NULL, &wasRequestUsed, &requestUsed);

                if (wasRequestUsed)
                {
                    if (requestUsed == request)
                    {
                        *out_completed = 1;
                        *out_receive_length = header[1];
                        *out_channel = i;
                        myCompletedRequests.erase(
                                myCompletedRequests.find(request));

                        OTF_Tracer::writeLeave (I_COMM_PROTOCOL__test_msg);
                        return GTI_SUCCESS;
                    }
                }
            }
        } while (bytesReceived == headerLen);
    }

    //not completed ...
    *out_completed = 0;

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__test_msg);
    return GTI_SUCCESS;
}

GTI_RETURN CProtSimpleTCP::wait_msg (
        unsigned int request,
        unsigned long* out_receive_length,
        unsigned long* out_channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__wait_msg);

    int bytesReceived;
    unsigned long header[2];
    unsigned long headerLen = sizeof (unsigned long) * 2;

    //(1) see whether request already in completed requests
    std::map <unsigned int, CompletionInfo>::iterator iter
    = myCompletedRequests.find(request);
    if (iter != myCompletedRequests.end())
    {
        *out_receive_length = iter->second.second;
        *out_channel = iter->second.first;
        myCompletedRequests.erase (iter);
        OTF_Tracer::writeLeave (I_COMM_PROTOCOL__wait_msg);
        return GTI_SUCCESS;
    }

    //(2) probe all channels and receive all available
    //    messages.
    //    Stop if either the request in question was
    //    completed, or all channels have no more messages
    unsigned int requestUsed;
    bool wasRequestUsed;

    do
    {
        for (unsigned int i = 0; i < myNumSockets; i++)
        {
            do
            {
                bytesReceived = ::recv (mySockets[i], header, headerLen, MSG_DONTWAIT);

                if (bytesReceived == headerLen)
                {
                    assert (header[0] == TOKEN_MSG_SIZE_HEADER);

                    recv_msg_conntent (i, header[1], NULL, NULL, &wasRequestUsed, &requestUsed);

                    if (wasRequestUsed)
                    {
                        if (requestUsed == request)
                        {
                            *out_receive_length = header[1];
                            *out_channel = i;
                            myCompletedRequests.erase(
                                    myCompletedRequests.find(request));

                            OTF_Tracer::writeLeave (I_COMM_PROTOCOL__wait_msg);
                            return GTI_SUCCESS;
                        }
                    }
                }
            } while (bytesReceived == headerLen);
        }
    }while (requestUsed != request);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__wait_msg);
    return GTI_SUCCESS;
}

GTI_RETURN CProtSimpleTCP::recv_msg_conntent (unsigned long channel, unsigned long size, void* target_buffer, bool *target_was_used, bool *request_was_used, unsigned int *request_finished)
{
    ssize_t sentBytes;
    unsigned long header[2];
    ssize_t headerLen = sizeof (unsigned long) * 2;

    if (target_was_used)
        *target_was_used = false;
    if (request_was_used)
        *request_was_used = false;

    //Can the message fill an active request ?
    if ((myActiveRequests[channel].size() == 0     ) &&
        (myActiveRequests[myNumSockets].size() == 0)    )
    {
        //No active request present
        if (target_buffer == NULL)
        {
            void *tempBuffer = malloc (size);
            sentBytes = ::blocking_recv (mySockets[channel], tempBuffer, size);
            assert (sentBytes == size);

            myOutstandingMsgs[channel].push_back (std::make_pair(size,tempBuffer));
        }
        else
        {
            sentBytes = ::blocking_recv (mySockets[channel], target_buffer, size);
            assert (sentBytes == size);

            if (target_was_used)
                *target_was_used = true;
        }
    }
    else
    {
        Request r;

        //Satisfy an active request
        if (myActiveRequests[channel].size() &&
                !myActiveRequests[myNumSockets].size())
        {
            r = myActiveRequests[channel].front();
            myActiveRequests[channel].pop_front();
        }
        else
            if (!myActiveRequests[channel].size() &&
                    myActiveRequests[myNumSockets].size())
            {
                r = myActiveRequests[myNumSockets].front();
                myActiveRequests[myNumSockets].pop_front();
            }
            else
                if (myActiveRequests[channel].front().second.first <
                    myActiveRequests[myNumSockets].front().second.first)
                {
                    r = myActiveRequests[channel].front();
                    myActiveRequests[channel].pop_front();
                }
                else
                {
                    r = myActiveRequests[myNumSockets].front();
                    myActiveRequests[myNumSockets].pop_front();
                }

        assert (r.second.second.second >= size);
        sentBytes = ::blocking_recv (mySockets[channel], r.second.second.first, size);
        assert (sentBytes == size);

        if (request_finished)
            *request_finished = r.first;

        if (request_was_used)
            *request_was_used = true;

        myCompletedRequests.insert (
                std::make_pair (r.first,
                        std::make_pair(channel, size)
                )
        );
    }

    //Send a was-received reply
    header[0] = TOKEN_SYNC_MSG;
    header[1] = 0;
    sentBytes = ::all_send (mySockets[channel], &header, headerLen);
    assert (sentBytes == headerLen);
    OTF_Tracer::writeRecvMsg (mySocketIDsOTF[channel], 0, size);

    return GTI_SUCCESS;
}

/*EOF*/
