/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CommProtocolTemplate.cpp
 *       Template for implementations of the communication
 *       protocol interface.
 *
 * All functions of the interface are offered, but
 * will lead to an assertion.
 *
 */

#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <pnmpimod.h>
#include <stdlib.h>

#include "CommProtocolTemplate.h"
#include "GtiTraceDefines.h"
#include "GtiMacros.h"

using namespace gti;

mGET_INSTANCE_FUNCTION(CommProtocolTemplate)
mFREE_INSTANCE_FUNCTION(CommProtocolTemplate)
mPNMPI_REGISTRATIONPOINT_FUNCTION(CommProtocolTemplate)

//=============================
// CommProtocolTemplate
//=============================
CommProtocolTemplate::CommProtocolTemplate (const char* intanceName)
: ModuleBase<CommProtocolTemplate, I_CommProtocol> (intanceName)
{
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //A Comm Protocol likely needs no sub modules
    assert (subModInstances.size() == 0);
}

//=============================
// ~CommProtocolTemplate
//=============================
CommProtocolTemplate::~CommProtocolTemplate (void)
{
    OTF_Tracer::writeEnter (COMM_PROTOCOL__Destructor);

    //TODO

    OTF_Tracer::writeLeave (COMM_PROTOCOL__Destructor);
}

//=============================
// isTraceable
//=============================
bool CommProtocolTemplate::isTraceable (void)
{
    return true;
}

//=============================
// setupTracing
//=============================
GTI_RETURN CommProtocolTemplate::setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData)
{
    if (init == GTI_TRACE_INIT_COPY)
    {
        OTF_Tracer::setup ((OTF_Tracer*) pSetupData);
    }
    else if (init == GTI_TRACE_INIT_CHILD)
    {
        uint32_t    parent = *((uint32_t*)pSetupData),
                    proc_id = parent;

        //Configuration data
        std::map<std::string, std::string> data = getData();
        if (data.find("otf_proc_offset") != data.end ())
            proc_id += atoi(data.find("otf_proc_offset")->second.c_str());
        else
            proc_id += 1; //TODO should warn or abort ?

        OTF_Tracer::setup (proc_id, parent, "commProtocolTemplate");
    }
    else
    {
        /*This is not a root module*/
        assert (0);
    }

    //Setup Functions
    OTF_Tracer::defFunction (COMM_PROTOCOL__setupTracing, "setupTracing", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (COMM_PROTOCOL__Destructor, "~CommProtocolTemplate", FUNCTION_GROUP_ID__PROCESSING);

    OTF_Tracer::defFunction (I_COMM_PROTOCOL__shutdown, "shutdown", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__getNumChannels, "getNumChannels", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__ssend, "ssend", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__isend, "isend", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__recv, "recv", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__irecv, "irecv", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__test_msg, "test_msg", FUNCTION_GROUP_ID__COMMUNICATE);
    OTF_Tracer::defFunction (I_COMM_PROTOCOL__wait_msg, "wait_msg", FUNCTION_GROUP_ID__COMMUNICATE);

    OTF_Tracer::writeEnter (COMM_PROTOCOL__setupTracing);

    //Setup sub-modules
    //none ...

    OTF_Tracer::writeLeave (COMM_PROTOCOL__setupTracing);
    return GTI_SUCCESS;
}

//=============================
// isConnected
//=============================
bool CommProtocolTemplate::isConnected (void)
{
	return false;
}

//=============================
// isInitialized
//=============================
bool CommProtocolTemplate::isInitialized (void)
{
	return false;
}

//=============================
// isFinalized
//=============================
bool CommProtocolTemplate::isFinalized (void)
{
	return false;
}

//=============================
// shutdown
//=============================
GTI_RETURN CommProtocolTemplate::shutdown (void)
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__shutdown);

    assert (0);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__shutdown);
    return GTI_SUCCESS;
}

//=============================
// getNumChannels
//=============================
GTI_RETURN CommProtocolTemplate::getNumChannels (
        unsigned long* out_numChannels)
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__getNumChannels);

    assert (0);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__getNumChannels);
    return GTI_SUCCESS;
}

//=============================
// removeOutstandingRequests
//=============================
GTI_RETURN CommProtocolTemplate::removeOutstandingRequests (void)
{
    return GTI_SUCCESS;
}

//=============================
// ssend
//=============================
GTI_RETURN CommProtocolTemplate::ssend (
        void* buf,
        unsigned long num_bytes,
        unsigned long channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__ssend);

    assert (0);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__ssend);
    return GTI_SUCCESS;
}

//=============================
// isend
//=============================
GTI_RETURN CommProtocolTemplate::isend (
        void* buf,
        unsigned long num_bytes,
        unsigned int *out_request,
        unsigned long channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__recv);

    assert (0);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__recv);
    return GTI_SUCCESS;
}

//=============================
// recv
//=============================
GTI_RETURN CommProtocolTemplate::recv (
        void* out_buf,
        unsigned long num_bytes,
        unsigned long* out_length,
        unsigned long channel,
        unsigned long *out_channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__recv);

    assert (0);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__recv);
    return GTI_SUCCESS;
}

//=============================
// irecv
//=============================
GTI_RETURN CommProtocolTemplate::irecv (
        void* out_buf,
        unsigned long num_bytes,
        unsigned int *out_request,
        unsigned long channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__irecv);

    assert (0);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__irecv);
    return GTI_SUCCESS;
}

//=============================
// test_msg
//=============================
GTI_RETURN CommProtocolTemplate::test_msg (
        unsigned int request,
        int* out_completed,
        unsigned long* out_receive_length,
        unsigned long* out_channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__test_msg);

    assert (0);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__test_msg);
    return GTI_SUCCESS;
}

//=============================
// wait_msg
//=============================
GTI_RETURN CommProtocolTemplate::wait_msg (
        unsigned int request,
        unsigned long* out_receive_length,
        unsigned long* out_channel
        )
{
    OTF_Tracer::writeEnter (I_COMM_PROTOCOL__wait_msg);

    assert (0);

    OTF_Tracer::writeLeave (I_COMM_PROTOCOL__wait_msg);
    return GTI_SUCCESS;
}

/*EOF*/
