/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CStratSimpleUp.cpp
 *        Simple implementation for the I_CommStrategyUp interface.
 *
 * Provides the interface functionality with a simple implementation.
 *
 * @author Tobias Hilbrich
 * @date 16.04.2009
 */

#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <pnmpimod.h>
#include <stdlib.h>

#include "CStratSimpleUp.h"
#include "GtiTraceDefines.h"
#include "GtiMacros.h"

using namespace gti;

const unsigned long CStratSimpleUp::myTokenShutdownSync = 0xFFFFFFFF;
const unsigned long CStratSimpleUp::myTokenMessage = 0xFFFFFFFE;
const unsigned long CStratSimpleUp::myTokenAck = 0xFFFFFFFD;

mGET_INSTANCE_FUNCTION(CStratSimpleUp);
mFREE_INSTANCE_FUNCTION(CStratSimpleUp);
mPNMPI_REGISTRATIONPOINT_FUNCTION(CStratSimpleUp);

//=============================
// buf_free_function
//=============================
GTI_RETURN buf_free_function (void* free_data, unsigned long num_bytes, void* buf)
{
    delete [] (unsigned long*) buf;
    return GTI_SUCCESS;
}

//=============================
// CStratSimpleUp
//=============================
CStratSimpleUp::CStratSimpleUp (const char* instanceName)
    : ModuleBase<CStratSimpleUp, CStratUpQueue> (instanceName),
      myMsgQueue()
{
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //A Comm Strategy either uses a regular communication protocol or a regular and a
    //fall-back one.
    assert ((subModInstances.size() == 1) || (subModInstances.size() == 2));

    protocol = (I_CommProtocol*) subModInstances[0];
    if (subModInstances.size() == 2)
        fallback = (I_CommProtocol*) subModInstances[1];
    else
        fallback = NULL;

    //Set up own data
    myRequest = 0xFFFFFFFF;
    myBuf[0] = myBuf[1] = 0; //of no importance
    myGotPing = false;
}

//=============================
// ~CStratSimpleUp
//=============================
CStratSimpleUp::~CStratSimpleUp (void)
{
    OTF_Tracer::writeEnter (COMM_STRATEGY_UP__Destructor);

    /*Destroy sub modules*/
    if (protocol)
    	destroySubModuleInstance ((I_Module*)protocol);
    protocol = NULL;

    if (fallback)
        destroySubModuleInstance ((I_Module*)fallback);
    fallback = NULL;

    OTF_Tracer::writeLeave (COMM_STRATEGY_UP__Destructor);
}

//=============================
// isTraceable
//=============================
bool CStratSimpleUp::isTraceable (void)
{
    return true;
}

//=============================
// setupTracing
//=============================
GTI_RETURN CStratSimpleUp::setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData)
{
    if (init == GTI_TRACE_INIT_COPY)
    {
        OTF_Tracer::setup ((OTF_Tracer*) pSetupData);
    }
    else if (init == GTI_TRACE_INIT_CHILD)
    {
        uint32_t    parent = *((uint32_t*)pSetupData),
        proc_id = parent;

        //Configuration data
        std::map<std::string, std::string> data = getData();
        if (data.find("otf_proc_offset") != data.end ())
            proc_id += atoi(data.find("otf_proc_offset")->second.c_str());
        else
            proc_id += 1; //TODO should warn or abort ?

        OTF_Tracer::setup (proc_id, parent, "commStrategyUpTemplate");
    }
    else
    {
        /*This is not a root module*/
        assert (0);
    }

    //Setup Functions
    OTF_Tracer::defFunction (COMM_STRATEGY_UP__setupTracing, "setupTracing", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (COMM_STRATEGY_UP__Destructor, "~CommStrategyUpTemplate", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__shutdown, "shutdown", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__send, "send", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__test, "test", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__wait, "wait", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__flush, "flush", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__enforce_blocking_comm, "enforce_blocking_comm", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__activate_fallback_comm, "enforce_blocking_comm", FUNCTION_GROUP_ID__PROCESSING);

    OTF_Tracer::writeEnter (COMM_STRATEGY_UP__setupTracing);

    //Setup sub-modules
    if (protocol->isTraceable())
        protocol->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    if ((fallback) && (fallback->isTraceable()))
        fallback->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    OTF_Tracer::writeLeave (COMM_STRATEGY_UP__setupTracing);
    return GTI_SUCCESS;
}

//=============================
// shutdown
//=============================
GTI_RETURN CStratSimpleUp::shutdown (
        GTI_FLUSH_TYPE flush_behavior,
        GTI_SYNC_TYPE sync_behavior)
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__shutdown);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__shutdown);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==shutdown
    if (flush_behavior == GTI_FLUSH)
        flush ();

    // Remove protocol requests, to avoid sync msgs beeing catched by
    // outstanding requests !
    protocol->removeOutstandingRequests ();

    if (sync_behavior == GTI_SYNC)
    {
        unsigned long buf[2] = {0,0};
        unsigned long length;

        //Receive the ping (if not already received)
        while ((buf[0] != myTokenShutdownSync) && (!myGotPing))
        {
            protocol->recv(buf,sizeof(unsigned long)*2,&length,0,NULL);
            assert(length == sizeof(unsigned long)*2);

            if (buf[0] != myTokenShutdownSync)
            {
                //TODO give a warning here, received some other outstanding
                //     message even though this strategy is being closed
                //Receive the message, and discard it
                std::cerr << "WARNING: In shutdown sync (StrategyUp) "
                          << "received an outstanding message!"
                          << std::endl;
                unsigned long *temp_buf;
                length = buf[1];
                if (length % sizeof(unsigned long)) length = ((length/sizeof(unsigned long))+1)*sizeof(unsigned long);
                temp_buf = new unsigned long[length/sizeof(unsigned long)];
                protocol->recv(temp_buf,buf[1],&length,0, NULL);
                delete temp_buf;
            }
        }

        //Send the pong
        buf[0] = myTokenShutdownSync;
        buf[1] = 0; //of no importance
        protocol->ssend(buf,sizeof(unsigned long)*2,0);
    }

    //Shutdown the protocols
    protocol->shutdown();
    if (fallback)
        fallback->shutdown();

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__shutdown);
    return GTI_SUCCESS;
}

//=============================
// send
//=============================
GTI_RETURN CStratSimpleUp::send (
        void* buf,
        unsigned long num_bytes,
        void* free_data,
        GTI_RETURN (*buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__send);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	AddToQueue (buf, num_bytes, free_data, buf_free_function);
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__send);
    	return GTI_SUCCESS;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==send
    unsigned long temp_buf[2] = {myTokenMessage, num_bytes};
    protocol->ssend(temp_buf,sizeof(unsigned long)*2,0);
    protocol->ssend(buf,num_bytes,0);

    //Wait for the acknowledgment
    bool hadAck = false;
    while (!hadAck)
    {
    		unsigned long length;
    		protocol->recv(temp_buf, sizeof(unsigned long)*2, &length, 0, NULL);

    		if (temp_buf[0] == myTokenShutdownSync)
        {
    			//We got a shutdown ping instead
    			assert (myGotPing == false);
            myGotPing = true;
        }
    		else if (temp_buf[0] == myTokenMessage)
    		{
    			//We got a regular message instead
    			unsigned long buf_size, recv_size;
    			if (temp_buf[1] % sizeof(unsigned long) != 0)
    				buf_size = (temp_buf[1]/sizeof(unsigned long))+1;
    			else
    				buf_size = (temp_buf[1]/sizeof(unsigned long));

    			unsigned long* tempBuf = new unsigned long[buf_size];

    			protocol->recv(tempBuf,temp_buf[1],&recv_size,0, NULL);

    			assert (recv_size == temp_buf[1]); //must be the size specified in first message

    			myMsgQueue.push_back(std::make_pair(recv_size, tempBuf));
    		}
    		else
    		{
    			//We got the acknowledgement
    			assert (temp_buf[0] == myTokenAck);
    			hadAck = true;
    		}
    }

    if (buf_free_function)
        (*buf_free_function) (free_data,num_bytes,buf);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__send);
    return GTI_SUCCESS;
}

//=============================
// test
//=============================
GTI_RETURN CStratSimpleUp::test (
        int* out_flag,
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__test);

    //Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__test);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //Did we receive messages earlier on (when waiting for an ack)
    if (myMsgQueue.size())
    {
		*out_flag = 1;
		*out_num_bytes = myMsgQueue.front().first;
		*out_buf = myMsgQueue.front().second;
		*out_free_data = NULL;
		*out_buf_free_function = buf_free_function;

		myMsgQueue.pop_front();

		OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__test);
		return GTI_SUCCESS;
    }

    //test
    if (myRequest == 0xFFFFFFFF)
    {
        protocol->irecv(myBuf,sizeof(unsigned long)*2,&myRequest,0);
    }

    int completed;
    unsigned long numbytes, channel;

    protocol->test_msg(myRequest,&completed,&numbytes,&channel);

    if (completed)
    {
        //Save request as used up
        myRequest = 0xFFFFFFFF;

        assert (numbytes == sizeof(unsigned long)*2); //has to be a token,length message

        //Did we receive the ping used for ping-pong shutdown sync ?
        if (myBuf[0] == myTokenShutdownSync)
        {
            assert (myGotPing == false);
            myGotPing = true;
            //Retry, maybe another message available ...
            OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__test);
            return test (out_flag, out_num_bytes, out_buf, out_free_data, out_buf_free_function);
        }

        assert (myBuf[0] == myTokenMessage); //has to be a message

        unsigned long* buf;
        unsigned long  buf_size, recv_size;

        if (myBuf[1] % sizeof(unsigned long) != 0)
            buf_size = (myBuf[1]/sizeof(unsigned long))+1;
        else
            buf_size = (myBuf[1]/sizeof(unsigned long));

        buf = new unsigned long[buf_size];

        protocol->recv(buf,myBuf[1],&recv_size,channel,NULL);

        assert (recv_size == myBuf[1]); //must be the size specified in first message

        //save values
        *out_flag = 1;
        *out_num_bytes = recv_size;
        *out_buf = buf;
        *out_free_data = NULL;
        *out_buf_free_function = buf_free_function;
    }
    else
    {
        //No message received
        *out_flag = 0;
    }

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__test);
    return GTI_SUCCESS;
}

//=============================
// wait
//=============================
GTI_RETURN CStratSimpleUp::wait (
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__wait);

    //Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__wait);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //Did we receive messages earlier on (when waiting for an ack)
    if (myMsgQueue.size())
    {
			*out_num_bytes = myMsgQueue.front().first;
			*out_buf = myMsgQueue.front().second;
			*out_free_data = NULL;
			*out_buf_free_function = buf_free_function;

			myMsgQueue.pop_front();

			OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__wait);
			return GTI_SUCCESS;
    }

    //wait
    unsigned long numbytes, channel;

    if (myRequest != 0xFFFFFFFF)
    {
        protocol->wait_msg(myRequest,&numbytes,&channel);
    }
    else
    {
        protocol->recv(myBuf,sizeof(unsigned long)*2,&numbytes,0,&channel);
    }

    //Save request as used up
    myRequest = 0xFFFFFFFF;

    //Sanity
    assert (numbytes == sizeof(unsigned long)*2); //has to be a token,length message

    //Did we receive the ping used for ping-pong shutdown sync ?
    if (myBuf[0] == myTokenShutdownSync)
    {
        assert (myGotPing == false);
        myGotPing = true;
        //Retry, maybe another message available ...
        OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__wait);
        return wait (out_num_bytes, out_buf, out_free_data, out_buf_free_function);
    }

    //Regular message
    assert (myBuf[0] == myTokenMessage); //has to be a message

    unsigned long* buf;
    unsigned long  buf_size, recv_size;

    if (myBuf[1] % sizeof(unsigned long) != 0)
        buf_size = (myBuf[1]/sizeof(unsigned long))+1;
    else
        buf_size = (myBuf[1]/sizeof(unsigned long));

    buf = new unsigned long[buf_size];

    protocol->recv(buf,myBuf[1],&recv_size,channel, NULL);

    assert (recv_size == myBuf[1]); //must be the size specified in first message

    //save values
    *out_num_bytes = recv_size;
    *out_buf = buf;
    *out_free_data = NULL;
    *out_buf_free_function = buf_free_function;

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__wait);
    return GTI_SUCCESS;
}

//=============================
// flush
//=============================
GTI_RETURN CStratSimpleUp::flush (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__flush);

    //Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //Flush
    //No action needed
    //->send uses ssend

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__flush);
    return GTI_SUCCESS;
}

//=============================
// enforce_blocking_comm
//=============================
GTI_RETURN CStratSimpleUp::enforce_blocking_comm (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__enforce_blocking_comm);

    //Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //Enforce_blocking_comm
    //No action needed
    //->send uses ssend

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__enforce_blocking_comm);
    return GTI_SUCCESS;
}

//=============================
// activate_fallback_comm
//=============================
GTI_RETURN CStratSimpleUp::activate_fallback_comm (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__activate_fallback_comm);

    //not supported
    assert (0);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__activate_fallback_comm);
    return GTI_SUCCESS;
}

/*EOF*/
