/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CStratThreadedDown.cpp
 *        Application phase aware communication strategy.
 *
 * Provides the interface functionality with an implementation
 * that tries to avoid any comm. during phases in which the
 * application is communicating.
 *
 * @author Tobias Hilbrich
 * @date 24.08.2009
 */

#include <assert.h>
#include <stdio.h>
#include <mpi.h>
#include <pnmpimod.h>
#include "CStratThreadedDown.h"
#include "GtiTraceDefines.h"
#include "GtiMacros.h"
#include <stdlib.h>

using namespace gti;

mGET_INSTANCE_FUNCTION(CStratThreadedDown);
mFREE_INSTANCE_FUNCTION(CStratThreadedDown);
mPNMPI_REGISTRATIONPOINT_FUNCTION(CStratThreadedDown);

//=============================
// my_buf_free_function
//=============================
GTI_RETURN my_buf_free_function (void* free_data, unsigned long num_bytes, void* buf)
{
	CStratAggregateInfo* aggre_info = (CStratAggregateInfo*)free_data;
	assert (aggre_info->num_in_use); //must be 1 or higher!
	aggre_info->num_in_use--;

	if (aggre_info->num_in_use == 0 && aggre_info->num_msgs_left == 0)
	{
		//aggregate completely processed!
		aggre_info->instance->used_aggregates.erase(aggre_info);
		aggre_info->instance->free_aggregates.push_back (aggre_info);
	}

    return GTI_SUCCESS;
}

//=============================
// my_long_buf_free_function
//=============================
GTI_RETURN my_long_buf_free_function (void* free_data, unsigned long num_bytes, void* buf)
{
	delete [] (unsigned long*) buf;
	return GTI_SUCCESS;
}

//=============================
// CStratAggregateInfo
//=============================
CStratAggregateInfo::CStratAggregateInfo(unsigned long buf_size)
{
	buf = new char[buf_size];
	current_position = 0;
	num_msgs_left = 0;
	num_in_use = 0;
	channel = 0;
}

//=============================
// ~CStratAggregateInfo
//=============================
CStratAggregateInfo::~CStratAggregateInfo()
{
	delete [] buf;
}

//=============================
// CStratThreadedDown
//=============================
CStratThreadedDown::CStratThreadedDown (const char* instanceName)
: ModuleBase<CStratThreadedDown, CStratDownQueue> (instanceName),
  free_aggregates(), open_aggregate(NULL), used_aggregates(), test_aggregate(NULL)
{
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //Either uses a regular comm protocol or a regular and a
    //fall-back one.
    assert ((subModInstances.size() == 1) || (subModInstances.size() == 2));

    protocol = (I_CommProtocol*) subModInstances[0];
    if (subModInstances.size() == 2)
        fallback = (I_CommProtocol*) subModInstances[1];
    else
        fallback = NULL;

    //Set up internal data
    test_request = 0xFFFFFFFF;
}

//=============================
// ~CStratThreadedDown
//=============================
CStratThreadedDown::~CStratThreadedDown (void)
{
    OTF_Tracer::writeEnter (COMM_STRATEGY_DOWN__Destructor);

    /*Destroy sub modules*/
    if (protocol)
    	destroySubModuleInstance ((I_Module*)protocol);

    if (fallback)
        destroySubModuleInstance ((I_Module*)fallback);

    //Free internal data
    std::map<CStratAggregateInfo*, int>::iterator iter;
    for (iter = used_aggregates.begin(); iter != used_aggregates.end (); iter++)
    	delete (iter->first);
    used_aggregates.clear();

    while (free_aggregates.size())
    {
    	CStratAggregateInfo* a = free_aggregates.front();
		delete (a);
		free_aggregates.pop_front();
    }

    if (open_aggregate)
    	delete (open_aggregate);

    if (test_aggregate)
    	delete (test_aggregate);

    OTF_Tracer::writeLeave (COMM_STRATEGY_DOWN__Destructor);
}

//=============================
// isTraceable
//=============================
bool CStratThreadedDown::isTraceable (void)
{
    return true;
}

//=============================
// setupTracing
//=============================
GTI_RETURN CStratThreadedDown::setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData)
{
    if (init == GTI_TRACE_INIT_COPY)
    {
        OTF_Tracer::setup ((OTF_Tracer*) pSetupData);
    }
    else if (init == GTI_TRACE_INIT_CHILD)
    {
        uint32_t    parent = *((uint32_t*)pSetupData),
                    proc_id = parent;

        //Configuration data
        std::map<std::string, std::string> data = getData();
        if (data.find("otf_proc_offset") != data.end ())
            proc_id += atoi(data.find("otf_proc_offset")->second.c_str());
        else
            proc_id += 1; //TODO should warn or abort ?

        OTF_Tracer::setup (proc_id, parent, "cstrat_asend_down");
    }
    else
    {
        /*This is not a root module*/
        assert (0);
    }

    //Setup Functions
    OTF_Tracer::defFunction (COMM_STRATEGY_DOWN__setupTracing, "setupTracing", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (COMM_STRATEGY_DOWN__Destructor, "~CommStrategyDown", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__shutdown, "shutdown", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__getNumClients, "getNumClients", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__broadcast, "broadcast", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__test, "test", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__wait, "wait", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__flush, "flush", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__enforce_blocking_comm, "enforce_blocking_comm", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__activate_fallback_comm, "activate_fallback_comm", FUNCTION_GROUP_ID__PROCESSING);

    OTF_Tracer::writeEnter (COMM_STRATEGY_DOWN__setupTracing);

    //Setup sub-modules
    if (protocol->isTraceable())
        protocol->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    if ((fallback) && (fallback->isTraceable()))
        fallback->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    OTF_Tracer::writeLeave (COMM_STRATEGY_DOWN__setupTracing);
    return GTI_SUCCESS;
}

//=============================
// shutdown
//=============================
GTI_RETURN CStratThreadedDown::shutdown (
        GTI_FLUSH_TYPE flush_behavior,
        GTI_SYNC_TYPE sync_behavior
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__shutdown);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__shutdown);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==shutdown
    if (flush_behavior == GTI_FLUSH)
        flush ();

    // Remove protocol requests, to avoid sync msgs beeing catched by
    // outstanding requests !
    protocol->removeOutstandingRequests ();

    if (sync_behavior == GTI_SYNC)
    {
        unsigned long num,i;
        protocol->getNumChannels(&num);
        for (i = 0; i < num; i++)
        {
            unsigned long buf[2] = {myTokenShutdownSync,0};
            unsigned long length;

            //Send the ping
            protocol->ssend(buf,sizeof(unsigned long)*2,i);

            //receive the pong
            do
            {
            	CStratAggregateInfo* aggregate = get_free_aggregate();
                protocol->recv(aggregate->buf, BUF_LENGTH, &length, i, NULL);
                assert (length >= sizeof(unsigned long)*2);
                if (((unsigned long*)aggregate->buf)[0] != myTokenShutdownSync)
                {
                    //TODO give a warning here, received some other outstanding
                    //     message even though this strategy is being closed
                    //Receive the message, and discard it
                    std::cerr << "WARNING: In shutdown sync (StrategDown) "
                              << "received an outstanding message!"
                              << std::endl;

                    if (((unsigned long*)aggregate->buf)[0] != myTokenLongMsg)
                    {
                    	unsigned long *temp_buf;
                    	length = ((unsigned long*)aggregate->buf)[1];
                    	if (length % sizeof(unsigned long)) length = ((length/sizeof(unsigned long))+1)*sizeof(unsigned long);
                    	temp_buf = new unsigned long[length/sizeof(unsigned long)];
                    	protocol->recv(temp_buf, ((unsigned long*)aggregate->buf)[1], &length, i, NULL);
                    	delete temp_buf;
                    }

                    free_aggregates.push_back(aggregate);
                }
            } while (buf[0] != myTokenShutdownSync);
        }
    }

    //Shutdown the protocols
    protocol->shutdown();
    if (fallback)
        fallback->shutdown();

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__shutdown);
    return GTI_SUCCESS;
}

//=============================
// broadcast
//=============================
GTI_RETURN CStratThreadedDown::broadcast (
        void* buf,
        unsigned long num_bytes,
        void* free_data,
        GTI_RETURN (*buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__broadcast);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	AddToQueue (buf, num_bytes, free_data, buf_free_function);
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__broadcast);
    	return GTI_SUCCESS;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==broadcast
    unsigned long num,i;
    protocol->getNumChannels(&num);
    for (i = 0; i < num; i++)
    {
        unsigned long temp_buf[2] = {myTokenMessage,
                                num_bytes};
        protocol->ssend(temp_buf,sizeof(unsigned long)*2,i);
        protocol->ssend(buf,num_bytes,i);
    }
    if (buf_free_function)
        (*buf_free_function) (free_data, num_bytes, buf);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__broadcast);
    return GTI_SUCCESS;
}

//=============================
// getNumClients
//=============================
GTI_RETURN CStratThreadedDown::getNumClients (unsigned long *outNumClients)
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__getNumClients);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__getNumClients);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==getNumClients
    if (outNumClients)
        protocol->getNumChannels(outNumClients);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__getNumClients);
    return GTI_SUCCESS;
}

//=============================
// test
//=============================
GTI_RETURN CStratThreadedDown::test (
        int* out_flag,
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
        unsigned long *outChannel
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__test);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__test);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==test

    //**
    //** (1) we still have an open aggregate with remaining messages
    //**
    if (open_aggregate)
    {
			GTI_RETURN ret = msg_from_open_aggregate (out_flag, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);
			OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__test);
			return ret;
    }

    //**
    //** (2) Do we already have a request for test calls ?
    //**
    if (test_request == 0xFFFFFFFF)
    {
    		test_aggregate = get_free_aggregate();
        protocol->irecv(test_aggregate->buf, BUF_LENGTH, &test_request, RECV_ANY_CHANNEL);
    }

    //**
    //** (3) Test for completion of the request
    //**
    int completed;
    unsigned long numbytes, channel;

    protocol->test_msg(test_request,&completed,&numbytes,&channel);

    //**
    //** (4) Handle new aggregate if the test succeeded
    //**
    if (completed)
    {
        assert (numbytes >= sizeof(unsigned long)*2); //each message has at least two entries
        assert (((unsigned long*)test_aggregate->buf)[0] == myTokenMessage ||
        		((unsigned long*)test_aggregate->buf)[0] == myTokenLongMsg    ); //has to be a message

        //Save request as used up
        test_request = 0xFFFFFFFF;

        if (((unsigned long*)test_aggregate->buf)[0] == myTokenMessage)
        {
				//Received an aggregate
				open_aggregate = test_aggregate;
				open_aggregate->current_position = 2*sizeof(unsigned long);
				open_aggregate->num_msgs_left = ((unsigned long*)open_aggregate->buf)[1];
				open_aggregate->num_in_use = 0;
				open_aggregate->channel = channel;
				open_aggregate->instance = this;
				test_aggregate = NULL;

				//return a message from the new aggregate
				GTI_RETURN ret = msg_from_open_aggregate (out_flag, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);
				OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__test);
				return ret;
        }
        else
        {
				//Received info for a long message
				GTI_RETURN ret = long_msg_from_info (
						((unsigned long*)test_aggregate->buf)[1], channel,
						out_flag, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);

				free_aggregates.push_back (test_aggregate);
				test_aggregate = NULL;

				OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__test);
				return ret;
        }
    }
    else
    {
        //No message received
        if (out_flag)
        	*out_flag = 0;
    }

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__test);
    return GTI_SUCCESS;
}

//=============================
// wait
//=============================
GTI_RETURN CStratThreadedDown::wait (
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
        unsigned long *outChannel
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__wait);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__wait);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==wait

    //**
    //** (1) we still have an open aggregate with remaining messages ?
    //**
    if (open_aggregate)
    {
			GTI_RETURN ret = msg_from_open_aggregate (NULL, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);
			OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__test);
			return ret;
	}

    //**
    //** (2) Wait for a message
    //**
    unsigned long numbytes, channel;

    if (test_request != 0xFFFFFFFF)
    {
        protocol->wait_msg(test_request,&numbytes,&channel);
    }
    else
    {
    		test_aggregate = get_free_aggregate();
        protocol->recv(test_aggregate->buf, BUF_LENGTH, &numbytes, RECV_ANY_CHANNEL, &channel);

        //DEBUG
        //std::cout << "RECVD NEW AGGREGATE: [" << numbytes << "]" << std::endl;
        //ENDDEBUG
    }

    //**
    //** (3) Handle new aggregate
    //**
    assert (numbytes >= sizeof(unsigned long)*2); //each message has at least two entries
    assert (((unsigned long*)test_aggregate->buf)[0] == myTokenMessage ||
    		((unsigned long*)test_aggregate->buf)[0] == myTokenLongMsg    ); //has to be a message

    //Save request as used up
    test_request = 0xFFFFFFFF;

    GTI_RETURN ret;
    if (((unsigned long*)test_aggregate->buf)[0] == myTokenMessage)
    {
			//Received an aggregate
			open_aggregate = test_aggregate;
			open_aggregate->current_position = 2*sizeof(unsigned long);
			open_aggregate->num_msgs_left = ((unsigned long*)open_aggregate->buf)[1];
			open_aggregate->num_in_use = 0;
			open_aggregate->channel = channel;
			open_aggregate->instance = this;
			test_aggregate = NULL;

			//return a message from the new aggregate
			ret = msg_from_open_aggregate (NULL, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);
    }
    else
    {
			//Received info for a long message
			ret = long_msg_from_info (
					((unsigned long*)test_aggregate->buf)[1], channel,
					NULL, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);

			free_aggregates.push_back (test_aggregate);
			test_aggregate = NULL;
    }

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__wait);
    return ret;
}

//=============================
// acknowledge
//=============================
GTI_RETURN CStratThreadedDown::acknowledge (
		unsigned long channel
        )
{
	/*Nothing to do*/
	return GTI_SUCCESS;
}

//=============================
// flush
//=============================
GTI_RETURN CStratThreadedDown::flush (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__flush);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__flush);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
      	ProcessQueue ();

    //==flush
    //No action needed
    //->broadcast uses ssend

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__flush);
    return GTI_SUCCESS;
}

//=============================
// enforce_blocking_comm
//=============================
GTI_RETURN CStratThreadedDown::enforce_blocking_comm (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__enforce_blocking_comm);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__enforce_blocking_comm);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==enforce_blocking_comm
    //No action needed
    //->broadcast uses ssend

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__enforce_blocking_comm);
    return GTI_SUCCESS;
}

//=============================
// activate_fallback_comm
//=============================
GTI_RETURN CStratThreadedDown::activate_fallback_comm (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__activate_fallback_comm);

    //not supported
    assert (0);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__activate_fallback_comm);
    return GTI_SUCCESS;
}

//=============================
// get_free_aggregate
//=============================
CStratAggregateInfo* CStratThreadedDown::get_free_aggregate (void)
{
	if (free_aggregates.size() == 0)
	{
		//create a new buffer
		free_aggregates.push_back (new CStratAggregateInfo (BUF_LENGTH));
	}

	CStratAggregateInfo* ret = free_aggregates.front();
	free_aggregates.pop_front();
	return ret;
}

//=============================
// msg_from_open_aggregate
//=============================
GTI_RETURN CStratThreadedDown::msg_from_open_aggregate (
        int* out_flag,
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
        unsigned long *outChannel
        )
{
	unsigned long msg_length = ((unsigned long*)open_aggregate->buf)[open_aggregate->current_position/sizeof(unsigned long)];
	void *buffer_start = open_aggregate->buf + open_aggregate->current_position + sizeof(unsigned long);
	void *free_data =  open_aggregate;

	if (out_flag)
		*out_flag = 1;
	if (out_num_bytes)
		*out_num_bytes = msg_length;
	if (out_buf)
		*out_buf = buffer_start;
	if (out_free_data)
		*out_free_data = free_data;
	if (out_buf_free_function)
		*out_buf_free_function = my_buf_free_function;
	if (outChannel)
		*outChannel = open_aggregate->channel;

	open_aggregate->num_in_use++;
	open_aggregate->num_msgs_left--;

	if (open_aggregate->num_msgs_left == 0)
	{
		//last message from aggregate -> add it to used
		used_aggregates.insert (std::make_pair(open_aggregate,0));
		open_aggregate = NULL;
	}
	else
	{
		//set position to next message in aggregate
		open_aggregate->current_position +=  sizeof(unsigned long) + msg_length;

		if (open_aggregate->current_position % sizeof(unsigned long))
		    open_aggregate->current_position += sizeof(unsigned long) - open_aggregate->current_position % sizeof(unsigned long);
	}

	//DEBUG
	//std::cout << "RECV: [" << *out_num_bytes << "]" << "id=" << ((unsigned long*) *out_buf)[0] << std::endl;
	//ENDDEBUG

	return GTI_SUCCESS;
}

//=============================
// long_msg_from_info
//=============================
GTI_RETURN CStratThreadedDown::long_msg_from_info (
		unsigned long in_length,
		unsigned long channel,
        int* out_flag,
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
        unsigned long *outChannel
        )
{
	//Received info for a long message
	unsigned long* buf;
	unsigned long  buf_size, recv_size;

	if (in_length % sizeof(unsigned long) != 0)
		buf_size = (in_length/sizeof(unsigned long))+1;
	else
		buf_size = (in_length/sizeof(unsigned long));

	buf = new unsigned long[buf_size];

	protocol->recv(buf, in_length ,&recv_size, channel, NULL);

	assert (recv_size == in_length); //must be the size specified in first message

	//save values
	if (out_flag)
		*out_flag = 1;
	if (out_num_bytes)
		*out_num_bytes = recv_size;
	if (out_buf)
		*out_buf = buf;
	if (out_free_data)
		*out_free_data = NULL;
	if (out_buf_free_function)
		*out_buf_free_function = my_long_buf_free_function;
	if (outChannel)
		*outChannel = channel;

	//DEBUG
	//std::cout << "LONG MSG RECV: [" << *out_num_bytes << "]" << "id=" << ((unsigned long*) *out_buf)[0] << std::endl;
	//ENDDEBUG

	return GTI_SUCCESS;
}

/*EOF*/
