/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CStratThreadedDown.h
 *        Application phase aware communication strategy.
 *
 * Provides the interface functionality with an implementation
 * that tries to avoid any comm. during phases in which the
 * application is communicating.
 *
 * @author Tobias Hilbrich
 * @date 24.08.2009
 */

#include "GtiDefines.h"
#include "GtiEnums.h"
#include "GtiTypes.h"
#include "I_CommStrategyDown.h"
#include "I_CommProtocol.h"
#include "ModuleBase.h"
#include "OtfTracer.h"
#include "CStratThreaded.h"
#include "CStratQueue.h"

#ifndef CSTRAT_THREADED_DOWN_H
#define CSTRAT_THREADED_DOWN_H

namespace gti
{
	class CStratThreadedDown;

	/**
	 * Class with aggregate information.
	 */
	class CStratAggregateInfo
	{
	public:
		char* buf;
		unsigned long current_position; //byte index for buf
		unsigned long num_msgs_left;
		unsigned long num_in_use;
		unsigned long channel;
		CStratThreadedDown *instance;

		CStratAggregateInfo(unsigned long buf_size);
		~CStratAggregateInfo(void);
	};

    /**
     * Class that describes the communication protocol interface.
     */
    class CStratThreadedDown : public ModuleBase<CStratThreadedDown, CStratDownQueue>, public CStratThreaded, public OTF_Tracer
    {
    public:
        std::map<CStratAggregateInfo*, int> used_aggregates;
        std::list<CStratAggregateInfo*> free_aggregates;

    protected:
        /** The communication protocol. */
        I_CommProtocol *protocol;

        /** The fallback communication protocol. */
        I_CommProtocol *fallback;

        unsigned int  test_request;

        CStratAggregateInfo* open_aggregate;
        CStratAggregateInfo* test_aggregate;

        /**
         * Returns a free aggregate from free_aggregates or
         * creates a new one if necessary.
         * @return a free aggregate.
         */
        CStratAggregateInfo* get_free_aggregate (void);

        /**
         * Handles receival of a message from the currently open
         * aggregate (open aggregate required).
         * @see I_CommStrategyDown::test
         */
        GTI_RETURN msg_from_open_aggregate (
                int* out_flag,
                unsigned long* out_num_bytes,
                void** out_buf,
                void** out_free_data,
                GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
                unsigned long *outChannel
                );

        /**
         * Handles receival of a long message.
         * @param in_length the length of the msg (received with the long msg token).
         * @param channel the channel to use.
         * @param out_flag flag for testing.
         * @param out_num_bytes number of bytes received.
         * @param out_buf buffer with received content.
         * @param out_free_data associated data for freeing out_buffer.
         * @param out_buf_free_function function to use for freeing out_buffer.
         * @see I_CommStrategyDown::test
         */
        GTI_RETURN long_msg_from_info (
        		unsigned long in_length,
        		unsigned long channel,
        		int* out_flag,
        		unsigned long* out_num_bytes,
        		void** out_buf,
        		void** out_free_data,
        		GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
        		unsigned long *outChannel
             );

    public:

        /**
         * Constructor.
         * @ref ModConf see for details on module configuration.
         * @param instanceName name of the instance.
         */
    	CStratThreadedDown (const char* instanceName);

        /**
         *Destructor.
         */
        ~CStratThreadedDown (void);

        /**
         * @see I_Module::isTraceable
         */
        bool isTraceable (void);

        /**
         * @see I_Traceable::setupTracing
         */
        GTI_RETURN setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData);

        /**
         * @see I_CommStrategyDown::getNumClients
         */
        GTI_RETURN getNumClients (unsigned long *outNumClients);

        /**
         * @see I_CommStrategyDown::shutdown
         */
        GTI_RETURN shutdown (
                GTI_FLUSH_TYPE flush_behavior,
                GTI_SYNC_TYPE sync_behavior);

        /**
         * @see I_CommStrategyDown::broadcast
         */
        GTI_RETURN broadcast (
                void* buf,
                unsigned long num_bytes,
                void* free_data,
                GTI_RETURN (*buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
                );

        /**
         * @see I_CommStrategyDown::test
         */
        GTI_RETURN test (
                int* out_flag,
                unsigned long* out_num_bytes,
                void** out_buf,
                void** out_free_data,
                GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
                unsigned long *outChannel
                );

        /**
         * @see I_CommStrategyDown::wait
         */
        GTI_RETURN wait (
                unsigned long* out_num_bytes,
                void** out_buf,
                void** out_free_data,
                GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
                unsigned long *outChannel
                );

        /**
         * @see I_CommStrategyDown::acknowledge
         */
        GTI_RETURN acknowledge (
        		unsigned long channel
        );

        /**
         * @see I_CommStrategyDown::flush
         */
        GTI_RETURN flush (
                void
                );

        /**
         * @see I_CommStrategyDown::enforce_blocking_comm
         */
        GTI_RETURN enforce_blocking_comm (
                void
                );

        /**
         * @see I_CommStrategyDown::activate_fallback_comm
         */
        GTI_RETURN activate_fallback_comm (
                void
                );
    };
}

#endif /* CSTRAT_THREADED_DOWN_H */
