/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CStratThreadedUp.cpp
 *        Application phase aware communication strategy.
 *
 * Provides the interface functionality with an implementation
 * that tries to avoid any comm. during phases in which the
 * application is communicating.
 *
 * @author Tobias Hilbrich
 * @date 24.08.2009
 */

#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <pnmpimod.h>
#include "CStratThreadedUp.h"
#include "GtiTraceDefines.h"
#include "GtiMacros.h"
#include <stdlib.h>
#include <strings.h>

using namespace gti;

mGET_INSTANCE_FUNCTION(CStratThreadedUp);
mFREE_INSTANCE_FUNCTION(CStratThreadedUp);
mPNMPI_REGISTRATIONPOINT_FUNCTION(CStratThreadedUp);

//=============================
// my_buf_free_function
//=============================
GTI_RETURN my_buf_free_function (void* free_data, unsigned long num_bytes, void* buf)
{
	delete [] (unsigned long*) buf;
	return GTI_SUCCESS;
}

//=============================
// CStratThreadedUp
//=============================
CStratThreadedUp::CStratThreadedUp (const char* instanceName)
    : ModuleBase<CStratThreadedUp, CStratUpQueue> (instanceName),
      myRequests(),
      myFreeBufs()
{
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //A Comm Strategy either uses a regular communication protocol or a regular and a
    //fall-back one.
    assert ((subModInstances.size() == 1) || (subModInstances.size() == 2));

    protocol = (I_CommProtocol*) subModInstances[0];
    if (subModInstances.size() == 2)
        fallback = (I_CommProtocol*) subModInstances[1];
    else
        fallback = NULL;

    /*Removed, as dummy comm protocols will return 0*/
    //Check correctness of comm-protocol
    //unsigned long num_channels;
    //protocol->getNumChannels(&num_channels);
    //assert (num_channels == 1); //upwards only one channel !

    //Set up own data
    myRequest = 0xFFFFFFFF;
    myGotPing = false;

    my_bufferA = new char [BUF_LENGTH];
    //my_bufferB = new char [BUF_LENGTH];
    my_pCommBuff = NULL;
    my_pAggregateBuf = my_bufferA;
    //my_requestCommBuf = 0xFFFFFFFF; //unused
    myMaxNumReqs = 0;

    prepareAggregateBuffer();
}

//=============================
// ~CStratThreadedUp
//=============================
CStratThreadedUp::~CStratThreadedUp (void)
{
    OTF_Tracer::writeEnter (COMM_STRATEGY_UP__Destructor);

    /*Destroy sub modules*/
    if (protocol)
    	destroySubModuleInstance ((I_Module*)protocol);
    protocol = NULL;

    if (fallback)
        destroySubModuleInstance ((I_Module*)fallback);
    fallback = NULL;

    if (my_pAggregateBuf)
    	delete [] my_pAggregateBuf;
    //if (my_bufferB)
    //	delete (my_bufferB);

    while (myFreeBufs.size())
    {
    	delete [] myFreeBufs.front();
    	myFreeBufs.pop_front();
    }

    ////What about outstanding requests ??? assert if present ?
    //std::list<std::pair<char*, unsigned int> > myRequests;

#ifdef GTI_DEBUG
    //DEBUG OUT
    std::cout << "MaxReqs: " << myMaxNumReqs << std::endl;
#endif

    OTF_Tracer::writeLeave (COMM_STRATEGY_UP__Destructor);
}

//=============================
// isTraceable
//=============================
bool CStratThreadedUp::isTraceable (void)
{
    return true;
}

//=============================
// setupTracing
//=============================
GTI_RETURN CStratThreadedUp::setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData)
{
    if (init == GTI_TRACE_INIT_COPY)
    {
        OTF_Tracer::setup ((OTF_Tracer*) pSetupData);
    }
    else if (init == GTI_TRACE_INIT_CHILD)
    {
        uint32_t    parent = *((uint32_t*)pSetupData),
        proc_id = parent;

        //Configuration data
        std::map<std::string, std::string> data = getData();
        if (data.find("otf_proc_offset") != data.end ())
            proc_id += atoi(data.find("otf_proc_offset")->second.c_str());
        else
            proc_id += 1; //TODO should warn or abort ?

        OTF_Tracer::setup (proc_id, parent, "cstrat_asend_up");
    }
    else
    {
        /*This is not a root module*/
        assert (0);
    }

    //Setup Functions
    OTF_Tracer::defFunction (COMM_STRATEGY_UP__setupTracing, "setupTracing", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (COMM_STRATEGY_UP__Destructor, "~CommStrategyUp", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__shutdown, "shutdown", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__send, "send", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__test, "test", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__wait, "wait", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__flush, "flush", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__enforce_blocking_comm, "enforce_blocking_comm", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_UP__activate_fallback_comm, "enforce_blocking_comm", FUNCTION_GROUP_ID__PROCESSING);

    OTF_Tracer::writeEnter (COMM_STRATEGY_UP__setupTracing);

    //Setup sub-modules
    if (protocol->isTraceable())
        protocol->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    if ((fallback) && (fallback->isTraceable()))
        fallback->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    OTF_Tracer::writeLeave (COMM_STRATEGY_UP__setupTracing);
    return GTI_SUCCESS;
}

//=============================
// shutdown
//=============================
GTI_RETURN CStratThreadedUp::shutdown (
        GTI_FLUSH_TYPE flush_behavior,
        GTI_SYNC_TYPE sync_behavior)
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__shutdown);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__shutdown);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==shutdown
    if (flush_behavior == GTI_FLUSH)
    {
        flush ();

        //remove outstanding requests
        while (myRequests.size())
        {
        	std::pair<char*, unsigned int> req = myRequests.front();
        	protocol->wait_msg(req.second,NULL,NULL);
        	delete []req.first;
        	myRequests.pop_front();
        }
    }

    // Remove protocol requests, to avoid sync msgs beeing catched by
    // outstanding requests !
    protocol->removeOutstandingRequests ();

    if (sync_behavior == GTI_SYNC)
    {
    	unsigned long buf[2] = {0,0};
    	unsigned long length;

    	//Receive the ping (if not already received)
    	while ((buf[0] != myTokenShutdownSync) && (!myGotPing))
    	{
    		protocol->recv(buf,sizeof(unsigned long)*2,&length,0,NULL);
    		assert(length == sizeof(unsigned long)*2);

    		if (buf[0] != myTokenShutdownSync)
    		{
    			//TODO give a warning here, received some other outstanding
    			//     message even though this strategy is being closed
    			//Receive the message, and discard it
    			std::cerr << "WARNING: In shutdown sync (StrategyUp) "
    			<< "received an outstanding message!"
    			<< std::endl;
    			unsigned long *temp_buf;
    			length = buf[1];
    			if (length % sizeof(unsigned long)) length = ((length/sizeof(unsigned long))+1)*sizeof(unsigned long);
    			temp_buf = new unsigned long[length/sizeof(unsigned long)];
    			protocol->recv(temp_buf,buf[1],&length,0, NULL);
    			delete temp_buf;
    		}
    	}

    	//Send the pong
    	buf[0] = myTokenShutdownSync;
    	buf[1] = 0; //of no importance
    	protocol->ssend(buf,sizeof(unsigned long)*2,0);
    }

    //Shutdown the protocols
    protocol->shutdown();
    if (fallback)
        fallback->shutdown();

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__shutdown);
    return GTI_SUCCESS;
}

//=============================
// send
//=============================
GTI_RETURN CStratThreadedUp::send (
        void* buf,
        unsigned long num_bytes,
        void* free_data,
        GTI_RETURN (*buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__send);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	AddToQueue (buf, num_bytes, free_data, buf_free_function);
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__send);
    	return GTI_SUCCESS;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==send

    //DEBUG
    //std::cout << "SEND: [" << num_bytes << "]" << "id=" << ((unsigned long*) buf)[0] << std::endl;
    //ENDDEBUG

    //Is it a long message ?
    if (num_bytes + 3*sizeof (unsigned long) > BUF_LENGTH) //3 -> message token + num records + one record length
    {
    	GTI_RETURN ret;
    	ret = send_long_message (buf, num_bytes, free_data, buf_free_function);
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__send);
    	return ret;
    }

    //Does it still fit into the aggregate buffer ?
    if (my_currAggregateLen + num_bytes + sizeof(unsigned long) > BUF_LENGTH)
    {
    	//Its too long ...
    	unsigned long comm_buf_length = my_currAggregateLen;
    	unsigned int request;
    	swap(); //finish up outstanding communication of last buffer and swap buffers

    	//DEBUG
    	//std::cout << "SEND AGGREGATE: [" << comm_buf_length << "]" << std::endl;
    	//ENDDEBUG

    	protocol->isend(my_pCommBuff, comm_buf_length, &request, 0);
    	myRequests.push_back (std::make_pair(my_pCommBuff,request));
    	my_pCommBuff = NULL;
    }

    //Add the new message to the aggregate buffer
    unsigned long startIndex = (my_currAggregateLen/sizeof(unsigned long));
    ((unsigned long*)my_pAggregateBuf)[1] = ((unsigned long*)my_pAggregateBuf)[1] + 1; //increase count of msgs
    ((unsigned long*)my_pAggregateBuf)[startIndex] = num_bytes;
    my_currAggregateLen += sizeof (unsigned long);
    bcopy (buf, my_pAggregateBuf+my_currAggregateLen, num_bytes);
    my_currAggregateLen += num_bytes;

    //pad to a multiple of "sizeof(unsigned long)"
    if (my_currAggregateLen % sizeof(unsigned long))
    	my_currAggregateLen += sizeof(unsigned long) - my_currAggregateLen % sizeof(unsigned long);

    //remove padding if buffer becomes too long due to that
    if (my_currAggregateLen > BUF_LENGTH)
    	my_currAggregateLen = BUF_LENGTH;

    //free given buffer
    buf_free_function (free_data, num_bytes, buf);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__send);
    return GTI_SUCCESS;
}

//=============================
// test
//=============================
GTI_RETURN CStratThreadedUp::test (
        int* out_flag,
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__test);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__test);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==test
    if (myRequest == 0xFFFFFFFF)
    {
        protocol->irecv(myBuf,sizeof(unsigned long)*2,&myRequest,0);
    }

    int completed;
    unsigned long numbytes, channel;

    protocol->test_msg(myRequest,&completed,&numbytes,&channel);

    if (completed)
    {
        //Save request as used up
        myRequest = 0xFFFFFFFF;

        assert (numbytes == sizeof(unsigned long)*2); //has to be a token,length message

        //Did we receive the ping used for ping-pong shutdown sync ?
        if (myBuf[0] == myTokenShutdownSync)
        {
            assert (myGotPing == false);
            myGotPing = true;
            //Retry, maybe another message available ...
            OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__test);
            return test (out_flag, out_num_bytes, out_buf, out_free_data, out_buf_free_function);
        }

        assert (myBuf[0] == myTokenMessage); //has to be a message

        unsigned long* buf;
        unsigned long  buf_size, recv_size;

        if (myBuf[1] % sizeof(unsigned long) != 0)
            buf_size = (myBuf[1]/sizeof(unsigned long))+1;
        else
            buf_size = (myBuf[1]/sizeof(unsigned long));

        buf = new unsigned long[buf_size];

        protocol->recv(buf,myBuf[1],&recv_size,channel,NULL);

        assert (recv_size == myBuf[1]); //must be the size specified in first message

        //save values
        *out_flag = 1;
        *out_num_bytes = recv_size;
        *out_buf = buf;
        *out_free_data = NULL;
        *out_buf_free_function = my_buf_free_function;
    }
    else
    {
        //No message received
        *out_flag = 0;
    }

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__test);
    return GTI_SUCCESS;
}

//=============================
// wait
//=============================
GTI_RETURN CStratThreadedUp::wait (
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__wait);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__wait);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==wait
    unsigned long numbytes, channel;

    if (myRequest != 0xFFFFFFFF)
    {
        protocol->wait_msg(myRequest,&numbytes,&channel);
    }
    else
    {
        protocol->recv(myBuf,sizeof(unsigned long)*2,&numbytes,0,&channel);
    }

    //Save request as used up
    myRequest = 0xFFFFFFFF;

    //Sanity
    assert (numbytes == sizeof(unsigned long)*2); //has to be a token,length message

    //Did we receive the ping used for ping-pong shutdown sync ?
    if (myBuf[0] == myTokenShutdownSync)
    {
        assert (myGotPing == false);
        myGotPing = true;
        //Retry, maybe another message available ...
        OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__wait);
        return wait (out_num_bytes, out_buf, out_free_data, out_buf_free_function);
    }

    //Regular message
    assert (myBuf[0] == myTokenMessage); //has to be a message

    unsigned long* buf;
    unsigned long  buf_size, recv_size;

    if (myBuf[1] % sizeof(unsigned long) != 0)
        buf_size = (myBuf[1]/sizeof(unsigned long))+1;
    else
        buf_size = (myBuf[1]/sizeof(unsigned long));

    buf = new unsigned long[buf_size];

    protocol->recv(buf,myBuf[1],&recv_size,channel, NULL);

    assert (recv_size == myBuf[1]); //must be the size specified in first message

    //save values
    *out_num_bytes = recv_size;
    *out_buf = buf;
    *out_free_data = NULL;
    *out_buf_free_function = my_buf_free_function;

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__wait);
    return GTI_SUCCESS;
}

//=============================
// flush
//=============================
GTI_RETURN CStratThreadedUp::flush (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__flush);

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__flush);
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==flush
    int completed;

    //Test for completion of existing send operations
    unsigned long old_len = my_currAggregateLen;
    if (old_len > 2*sizeof(unsigned long))
    {
    	unsigned int request;
    	swap(); //finish sending old buffer and swap buffers

    	//DEBUG
    	//std::cout << "SEND SHUTDOWN AGGREGATE: [" << old_len << "]" << std::endl;
    	//ENDDEBUG

    	protocol->isend(my_pCommBuff, old_len, &request, 0);
    	myRequests.push_back (std::make_pair(my_pCommBuff,request));
       	my_pCommBuff = NULL;
    }

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__flush);
    return GTI_SUCCESS;
}

//=============================
// enforce_blocking_comm
//=============================
GTI_RETURN CStratThreadedUp::enforce_blocking_comm (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__enforce_blocking_comm);

    //TODO
    assert(0);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__enforce_blocking_comm);
    return GTI_SUCCESS;
}

//=============================
// activate_fallback_comm
//=============================
GTI_RETURN CStratThreadedUp::activate_fallback_comm (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_UP__activate_fallback_comm);

    //not supported
    assert (0);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_UP__activate_fallback_comm);
    return GTI_SUCCESS;
}

//=============================
// swap
//=============================
void CStratThreadedUp::swap (void)
{
	/*
	 * - wait till curr comm buffer is received (if necessary)
	 * - switch buffers
	 * - prepare the new aggregate buf (set len = 0)
	 */
	my_pCommBuff = my_pAggregateBuf;
	my_pAggregateBuf = NULL;

	int completed=myRequests.size();

	if (completed > myMaxNumReqs)
		myMaxNumReqs = completed;

	while (completed)
	{
		std::pair<char*,unsigned int> req = myRequests.front();

		if (completed != MAX_NUM_MSGS)
		{
			protocol->test_msg(req.second,&completed,NULL,NULL);
		}
		else
		{
			protocol->wait_msg(req.second,NULL,NULL);
			completed=1;
		}

		if (completed)
		{
			if (!my_pAggregateBuf)
				my_pAggregateBuf = req.first;
			else
				myFreeBufs.push_back(req.first);

			myRequests.pop_front();

			completed = myRequests.size();
		}
	}

	if (!my_pAggregateBuf)
	{
		if (myFreeBufs.size())
		{
			my_pAggregateBuf = myFreeBufs.front();
			myFreeBufs.pop_front();
		}
		else
		{
			my_pAggregateBuf = new char [BUF_LENGTH];
		}
	}

	/*
	if (my_requestCommBuf != 0xFFFFFFFF)
		protocol->wait_msg(my_requestCommBuf,NULL,NULL);
	my_requestCommBuf = 0xFFFFFFFF;

	char *temp = my_pCommBuff;
	my_pCommBuff = my_pAggregateBuf;
	my_pAggregateBuf = temp;
	*/

	prepareAggregateBuffer();
}

//=============================
// prepareAggregateBuffer
//=============================
void CStratThreadedUp::prepareAggregateBuffer (void)
{
	//Prepare the aggregate buffer if it is still fresh
	my_currAggregateLen = 2*sizeof (unsigned long);
	((unsigned long*)my_pAggregateBuf)[0] = myTokenMessage;
	((unsigned long*)my_pAggregateBuf)[1] = 0; //curr num messages
}

//=============================
// send_long_message
//=============================
GTI_RETURN CStratThreadedUp::send_long_message (
        void* buf,
        unsigned long num_bytes,
        void* free_data,
        GTI_RETURN (*buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
	/*
	 * - send old aggregate (swap and isend)
	 * - finish sending current aggregate
	 * - send ping (overlong message token + length)
	 * - send the long message as one buffer (isend)
	 * - wait till sending is finished
	 */
	//unsigned long old_len = my_currAggregateLen;
	//swap(); //finish sending old buffer and swap buffers
	//if (old_len > 2*sizeof(unsigned long))
	//	protocol->ssend(my_pCommBuff, old_len, 0);

	unsigned long old_len = my_currAggregateLen;
	if (old_len > 2*sizeof(unsigned long))
	{
		unsigned int request;
		swap(); //finish sending old buffer and swap buffers
		protocol->isend(my_pCommBuff, old_len, &request, 0);
		myRequests.push_back (std::make_pair(my_pCommBuff,request));
		my_pCommBuff = NULL;
	}

	unsigned long temp_buf[2];
	temp_buf[0] = myTokenLongMsg;
	temp_buf[1] = num_bytes;
	protocol->ssend(temp_buf, 2*sizeof(unsigned long), 0);

	protocol->ssend(buf, num_bytes, 0);

	//free given buffer
	buf_free_function (buf, num_bytes, free_data);

	return GTI_SUCCESS;
}
/*EOF*/
