/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CommStrategyDownTemplate.cpp
 *       Template for implementations of the communication
 *       protocol interface.
 *
 * All functions of the interface are offered
 * and will lead to an assertion.
 *
 */

#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <pnmpimod.h>
#include "CommStrategyDownTemplate.h"
#include "GtiTraceDefines.h"
#include "GtiMacros.h"

using namespace gti;

mGET_INSTANCE_FUNCTION(CommStrategyDownTemplate)
mFREE_INSTANCE_FUNCTION(CommStrategyDownTemplate)
mPNMPI_REGISTRATIONPOINT_FUNCTION(CommStrategyDownTemplate)

CommStrategyDownTemplate::CommStrategyDownTemplate (const char* instanceName)
: ModuleBase<CommStrategyDownTemplate, I_CommStrategyDown> (instanceName)
{
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //A Comm Strategy either uses a regular communication protocol or a regular and a
    //fall-back one.
    //actual implementations should assert here
    ////assert ((subModInstances.size() == 1) || (subModInstances.size() == 2));

    if (subModInstances.size() >= 1)
        protocol = (I_CommProtocol*) subModInstances[0];
    else
        protocol = NULL;

    if (subModInstances.size() == 2)
        fallback = (I_CommProtocol*) subModInstances[1];
    else
        fallback = NULL;
}

CommStrategyDownTemplate::~CommStrategyDownTemplate (void)
{
    OTF_Tracer::writeEnter (COMM_STRATEGY_DOWN__Destructor);

    /*Destroy sub modules*/
    if (protocol)
        destroySubModuleInstance ((I_Module*)protocol);
    protocol = NULL;

    if (fallback)
        destroySubModuleInstance ((I_Module*)fallback);
    fallback = NULL;

    OTF_Tracer::writeLeave (COMM_STRATEGY_DOWN__Destructor);
}

bool CommStrategyDownTemplate::isTraceable (void)
{
    return true;
}

GTI_RETURN CommStrategyDownTemplate::setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData)
{
    if (init == GTI_TRACE_INIT_COPY)
    {
        OTF_Tracer::setup ((OTF_Tracer*) pSetupData);
    }
    else if (init == GTI_TRACE_INIT_CHILD)
    {
        uint32_t    parent = *((uint32_t*)pSetupData),
                    proc_id = parent;

        //Configuration data
        std::map<std::string, std::string> data = getData();
        if (data.find("otf_proc_offset") != data.end ())
            proc_id += atoi(data.find("otf_proc_offset")->second.c_str());
        else
            proc_id += 1; //TODO should warn or abort ?

        OTF_Tracer::setup (proc_id, parent, "commStrategyDownTemplate");
    }
    else
    {
        /*This is not a root module*/
        assert (0);
    }

    //Setup Functions
    OTF_Tracer::defFunction (COMM_STRATEGY_DOWN__setupTracing, "setupTracing", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (COMM_STRATEGY_DOWN__Destructor, "~CommStrategyDownTemplate", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__shutdown, "shutdown", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__getNumClients, "getNumClients", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__broadcast, "broadcast", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__test, "test", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__wait, "wait", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__flush, "flush", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__enforce_blocking_comm, "enforce_blocking_comm", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (I_COMM_STRATEGY_DOWN__activate_fallback_comm, "activate_fallback_comm", FUNCTION_GROUP_ID__PROCESSING);

    OTF_Tracer::writeEnter (COMM_STRATEGY_DOWN__setupTracing);

    //Setup sub-modules
    if ((protocol) && (protocol->isTraceable()))
        protocol->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    if ((fallback) && (fallback->isTraceable()))
        fallback->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    OTF_Tracer::writeLeave (COMM_STRATEGY_DOWN__setupTracing);
    return GTI_SUCCESS;
}

GTI_RETURN CommStrategyDownTemplate::shutdown (
        GTI_FLUSH_TYPE flush_behavior,
        GTI_SYNC_TYPE sync_behavior)
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__shutdown);


    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__shutdown);
    return GTI_SUCCESS;
}

GTI_RETURN CommStrategyDownTemplate::getNumClients (unsigned long *outNumClients)
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__getNumClients);

    *outNumClients = 0;

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__getNumClients);
    return GTI_SUCCESS;
}

GTI_RETURN CommStrategyDownTemplate::broadcast (
        void* buf,
        unsigned long num_bytes,
        void* free_data,
        GTI_RETURN (*buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__broadcast);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__broadcast);
    return GTI_SUCCESS;
}

GTI_RETURN CommStrategyDownTemplate::test (
        int* out_flag,
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
        unsigned long *outChannel
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__test);

    *out_flag=false;

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__test);
    return GTI_SUCCESS;
}

GTI_RETURN CommStrategyDownTemplate::wait (
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
        unsigned long *outChannel
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__wait);

    /*???*/
    while( 1 == 1) {}

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__wait);
    return GTI_SUCCESS;
}

GTI_RETURN CommStrategyDownTemplate::acknowledge (
	unsigned long channel
    )
{

	return GTI_SUCCESS;
}

GTI_RETURN CommStrategyDownTemplate::flush (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__flush);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__flush);
    return GTI_SUCCESS;
}

GTI_RETURN CommStrategyDownTemplate::enforce_blocking_comm (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__enforce_blocking_comm);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__enforce_blocking_comm);
    return GTI_SUCCESS;
}

GTI_RETURN CommStrategyDownTemplate::activate_fallback_comm (
        void
        )
{
    OTF_Tracer::writeEnter (I_COMM_STRATEGY_DOWN__activate_fallback_comm);

    OTF_Tracer::writeLeave (I_COMM_STRATEGY_DOWN__activate_fallback_comm);
    return GTI_SUCCESS;
}

/*EOF*/
