/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file I_PlaceReceival.h
 *       Interface for receival of trace and log data on a tool place.
 *
 * This interface receives trace and logging data received on a tool
 * place. Implementations will usually be auto generated.
 *
 * @author Tobias Hilbrich
 * @date 15.04.2009
 */

#include "GtiDefines.h"
#include "GtiEnums.h"
#include "GtiTypes.h"

#include "I_Module.h"
#include "I_ChannelId.h"

#ifndef I_PLACE_RECEIVAL_H
#define I_PLACE_RECEIVAL_H

namespace gti
{
    /**
     * Class that describes the communication protocol interface.
     */
    class I_PlaceReceival : public I_Module
    {
    public:

    	/**
    	 * Virtual destructor.
    	 */
    	virtual ~I_PlaceReceival () {}

        /**
         * Receives a record, and analyzes it.
         * The call may cause communication calls,
         * depending on the used forwarding communication
         * strategies used for this module.
         *
         * The given buffer may be freed when this call
         * returns.
         *
         * @param buf buffer that contains the trace record.
         * @param num_bytes length of the buffer in bytes.
         * @param free_data data associated with the buffer,
         *        used to store state for it.
         * @param buf_free_function function to call when the
         *        buffer may be freed.
         * @param numRemainingClients count of remaining
         *        places that are connected in the downwards
         *        direction, this number must be decremented
         *        when a shutdown token arrives, if it was
         *        the shutdown token of the last remaining
         *        connected place the receival must forward
         *        the shutdown event to all upwards connected
         *        places.
         * @param recordChannelId channel id of the incoming record
         *        must have been retrieved with a prior call to
         *        getUpdatedChannelId, the id will be freed at some point
         *        after the call returns, if the id is needed for a longer
         *        time its user must copy it.
         * @param outWasChannelIdSuspended pointer to storage for a
         *        boolean value, is set to true if a reduction was triggered for
         *        the record and still needs further data to complete.
         * @param outChannelIdsToOpen pointer to storage for a list of channel ids.
         *        Is used to store which channel ids were
         *        removed from suspension due to a successful reduction.
         *        The channel ids in the list must be freed by the caller of this function.
         *        The list must be empty when calling the function.
         */
        virtual GTI_RETURN ReceiveRecord (
                void* buf,
                unsigned long num_bytes,
                void* free_data,
                GTI_RETURN (*buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
                unsigned long *numRemainingClients,
                I_ChannelId *recordChannelId,
                bool *outWasChannelIdSuspended,
                std::list<I_ChannelId*> *outChannelIdsToOpen
                ) = 0;

        /**
         * Updates and returns the channel id of a newly received record.
         * The placement driver needs to pass the channel number from which it received
         * the record in order to allow the update.
         * The returned channel id should be used by the placement driver to suspend
         * and reactivate channels.
         * @param buf of serialized record
         * @param num_bytes size of buffer
         * @param channel from which the placement driver received the record.
         * @param numChannels number of channels connected to this place (total, not number of still active channels!)
         * @param pOutChannelId pointer to storage of channel id interface pointer,
         *               is going to be set to a pointer to the read and updated channel id.
         * @return GTI_SUCCESS if successful.
         */
        virtual GTI_RETURN getUpdatedChannelId (
        		void* buf,
        		unsigned long num_bytes,
        		unsigned long channel,
        		unsigned long numChannels,
        		I_ChannelId** pOutChannelId
        		) = 0;

        /**
         * Notifies the receival module that the placement driver stopped
         * blocking channels with ongoing reductions and that all ongoing
         * reductions must be notified that they must not succeed in their
         * ongoing reduction. The queued forwards must be executed as
         * a restult.
         *
         * Note: for reductions with semantic restrictions, e.g. a barrier
         *          is executed on ALL processes, the reduction should still
         *          track for the semantic completeness of the aborted reduction
         *          and start its next working reduction afterwards.
         *
         * @return GTI_SUCCESS if successful.
         */
        virtual GTI_RETURN timeOutReductions (void) = 0;
    };
}

#endif /* I_PLACE_RECEIVAL_H */
