/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file SuspensionBufferTree.cpp
 *       @see gti::SuspensionBufferTree
 *
 * @author Tobias Hilbrich
 * @date 13.12.2009
 */

#include <assert.h>
#include <iostream>
#include <sstream>

#include "SuspensionBufferTree.h"

using namespace gti;

//=============================
// SuspensionBufferTree
//=============================
SuspensionBufferTree::SuspensionBufferTree (int subIdIndex, long numChannels)
 : I_ChannelTree<SuspensionBufferTree> (subIdIndex, numChannels),
   myIsSuspended (false),
   myQRecords (),
   myParent (NULL),
   myNumChildRecords (0),
   myNumSuspendedChilds (0)
{
	/*Nothing to do*/
}

//=============================
// SuspensionBufferTree
//=============================
SuspensionBufferTree::SuspensionBufferTree (int subIdIndex, long numChannels, SuspensionBufferTree* parent)
 : I_ChannelTree<SuspensionBufferTree> (subIdIndex, numChannels),
   myIsSuspended (false),
   myQRecords (),
   myParent (parent),
   myNumChildRecords (0),
   myNumSuspendedChilds (0)
{
	/*Nothing to do*/
}

//=============================
// ~SuspensionBufferTree
//=============================
SuspensionBufferTree::~SuspensionBufferTree (void)
{
	//Free all enqueued records
	while (!myQRecords.empty())
	{
		RecordInfo info = myQRecords.front();
		myQRecords.pop_front();

		info.buf_free_function (info.free_data, info.num_bytes, info.buf);
	}

	//Do not delete the parent
	myParent = NULL;
}

//=============================
// getNode
//=============================
SuspensionBufferTree* SuspensionBufferTree::getNode (I_ChannelId* id, SuspensionBufferTree **outFirstSuspendedNode, SuspensionBufferTree **outFirstNonEmptyQueue)
{
	//Is this the first suspended node along the way ?
	if (myIsSuspended && outFirstSuspendedNode && *outFirstSuspendedNode == NULL)
		*outFirstSuspendedNode = this;

	//Is this the first node with a non-empty record queue along the way ?
	if (!myQRecords.empty() && outFirstNonEmptyQueue && *outFirstNonEmptyQueue == NULL)
		*outFirstNonEmptyQueue = this;

	//Get node for this channel if
	SuspensionBufferTree *ret = getChildForChannel (id);

	//was it invalid or is this the right node for the channel id ?
	if (ret == NULL || ret == this)
		return this;

	//recurse
	return ret->getNode(id, outFirstSuspendedNode, outFirstNonEmptyQueue);
}

//=============================
// isSuspended
//=============================
bool SuspensionBufferTree::isSuspended (void)
{
	return myIsSuspended;
}

//=============================
// setSuspension
//=============================
void SuspensionBufferTree::setSuspension (bool suspended)
{
	if (!suspended && myIsSuspended)
		if (myParent) myParent->decChildSuspensions();
	if (suspended && !myIsSuspended)
		if (myParent) myParent->incChildSuspensions();

	myIsSuspended = suspended;
}

//=============================
// getQueueSize
//=============================
size_t SuspensionBufferTree::getQueueSize(void)
{
	return myQRecords.size();
}

//=============================
// popFront
//=============================
RecordInfo SuspensionBufferTree::popFront (void)
{
	assert (!myQRecords.empty());
	RecordInfo ret = myQRecords.front();
	myQRecords.pop_front();
	if (myParent) myParent->decChildRecordNum();

	//If this is not a leaf: Push all records that are enqueued and do not belong to this node,
	//to the respective childs until the record queue is empty or the front element
	//belongs to this node (e.g. has a -1 as sub-id)
	while (!myQRecords.empty() && mySubIdIndex >= 0)
	{
		RecordInfo info = myQRecords.front();

		long int subId = getChannelForId(info.recordChannelId);

		//If first element is for this node: abort the redistribution loop
		if (subId == -1)
			break;

		//This element can be redistributed
		myQRecords.pop_front();
		/**
		 * @todo This is somewhat inefficent, the decrementing and incrementing
		 * resulting from this is absolutely unnecessary.
		 * A simple (but somewhat dirty) solution would be setting myParent
		 * to NULL and storing its old value. Afterwards do the requeuing and
		 * set myParent back to the old value afterward.
		 */
		if (myParent) myParent->decChildRecordNum();

		SuspensionBufferTree *target, *firstNonEmptyQ = NULL;
		target = getChildForChannel(info.recordChannelId)->getNode (info.recordChannelId, NULL, &firstNonEmptyQ);

		if (firstNonEmptyQ)
			target = firstNonEmptyQ;

		target->pushBack(info);
	}

	return ret;
}

//=============================
// pushBack
//=============================
void SuspensionBufferTree::pushBack(RecordInfo info)
{
	if (myParent) myParent->incChildRecordNum();
	myQRecords.push_back (info);
}

//=============================
// getQueuedRecord
//=============================
bool SuspensionBufferTree::getQueuedRecord (RecordInfo *outRecord)
{
	return getQueuedRecordRecurse (outRecord, NULL);
}

//=============================
// getQueuedRecord
//=============================
bool SuspensionBufferTree::getQueuedRecordRecurse (RecordInfo *outRecord, bool *outHadASuspension)
{
	bool noSuspensionInChilds = true;

	//Is this node suspended, if so no suitable record present!
	if (myIsSuspended)
	{
		if (outHadASuspension)
			*outHadASuspension = true;
		return false;
	}

	//Is there anything at all in this sub-tree ? If not don't recurse further!
	if (myQRecords.empty() && myNumChildRecords == 0)
	{
		//There is some suspended child if myNumSuspendedChilds is > 0
		if (outHadASuspension)
			*outHadASuspension = (myNumSuspendedChilds > 0); //myIsSuspended is already handled above, so no need to put this into the calculation
		return false;
	}

	//Look in the child nodes for a suitable record
	//Their queued records have priority over this nodes records!
	std::map<long int, SuspensionBufferTree*>::iterator iter;
	for (iter = myChilds.begin(); iter != myChilds.end(); iter++)
	{
		SuspensionBufferTree *child = iter->second;
		bool hadASuspension = false;

		if (child->getQueuedRecordRecurse(outRecord, &hadASuspension))
			return true;

		if (hadASuspension)
			noSuspensionInChilds = false;
	}

	//Set outHadASuspension correctly
	if (outHadASuspension)
		*outHadASuspension = !noSuspensionInChilds;

	//If no suspension in childs, we may use queued records of this node
	if (noSuspensionInChilds && !myQRecords.empty())
	{
		if (outRecord)
			*outRecord = popFront();

		return true;
	}

	return false;
}

//=============================
// isCompletlySuspended
//=============================
bool SuspensionBufferTree::isCompletlySuspended (void)
{
	if (myIsSuspended)
		return true;

	//not all sub ids present ? (if so, not completely suspended !)
	if (myChilds.size() != myNumChannels)
		return false;

	std::map<long int, SuspensionBufferTree*>::iterator iter;
	for (iter = myChilds.begin(); iter != myChilds.end(); iter++)
	{
		SuspensionBufferTree *child = iter->second;

		if (!child->isCompletlySuspended())
			return false;
	}

	return true;
}

//=============================
// hasAnySuspension
//=============================
bool SuspensionBufferTree::hasAnySuspension (void)
{
	if (myIsSuspended)
		return true;

	if (myNumSuspendedChilds > 0)
		return true;

	return false;
}

//=============================
// getChildQueueSize
//=============================
int SuspensionBufferTree::getChildQueueSize(void)
{
    return myNumChildRecords;
}

//=============================
// getNodeName
//=============================
std::string SuspensionBufferTree::getNodeName (void)
{
	char temp[128];
	sprintf (temp, "%d", mySubIdIndex);
	return temp;
}

//=============================
// getNodeColor
//=============================
std::string SuspensionBufferTree::getNodeColor (void)
{
	//Is suspended itself
	if (myIsSuspended)
		return "red";

	//Is indirectly suspended, due to some child being suspended
	if (myNumSuspendedChilds > 0)
		return "yellow";

	//Is open for processing
	return "green";
}

//=============================
// getNodeExtraRows
//=============================
std::string SuspensionBufferTree::getNodeExtraRows (void)
{
	std::stringstream stream;
	stream
		<< "NumRecordsInChilds: " << myNumChildRecords << "|"
		<< "NumSuspendedChilds: " << myNumSuspendedChilds	;

	std::deque<RecordInfo> temp = myQRecords;
	while (!temp.empty())
	{
		RecordInfo info = temp.front();
		temp.pop_front();
		stream << "|" << info.recordChannelId->toString() << ": " << ((unsigned long*)info.buf)[0];
	}

	return stream.str();
}

//=============================
// removeAllSuspensions
//=============================
void SuspensionBufferTree::removeAllSuspensions (void)
{
	myIsSuspended = false;
	myNumSuspendedChilds = 0;

	std::map<long int, SuspensionBufferTree*>::iterator iter;
	for (iter = myChilds.begin(); iter != myChilds.end(); iter++)
	{
		SuspensionBufferTree *child = iter->second;

		child->removeAllSuspensions();
	}
}

//=============================
// allocateChild
//=============================
SuspensionBufferTree* SuspensionBufferTree::allocateChild (int subIdIndex, long numChannels)
{
	return new SuspensionBufferTree (subIdIndex, numChannels, this);
}

//=============================
// incChildRecordNum
//=============================
void SuspensionBufferTree::incChildRecordNum (void)
{
	myNumChildRecords++;
	if (myParent) myParent->incChildRecordNum();
}

//=============================
// decChildRecordNum
//=============================
void SuspensionBufferTree::decChildRecordNum (void)
{
	myNumChildRecords--;
	assert (myNumChildRecords >= 0);
	if (myParent) myParent->decChildRecordNum();
}

//=============================
// incChildSuspensions
//=============================
void SuspensionBufferTree::incChildSuspensions (void)
{
	myNumSuspendedChilds++;
	if (myParent) myParent->incChildSuspensions();
}

//=============================
// decChildSuspensions
//=============================
void SuspensionBufferTree::decChildSuspensions (void)
{
	myNumSuspendedChilds--;
	assert (myNumSuspendedChilds >= 0);
	if (myParent) myParent->decChildSuspensions();
}

/*EOF*/
