/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file ThreadedMpiPlace.cpp
 *       A placement driver that is started as an MPI
 *       application and uses multi-threading to allow
 *       concurrent communication (loggs, mgmt, trace).
 *
 *
 * @author Tobias Hilbrich
 */

#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <sys/time.h>

#include <pnmpimod.h>
#include "ThreadedMpiPlace.h"
#include "GtiTraceDefines.h"
#include "GtiMacros.h"
#include "SuspensionBufferTree.h"

#ifdef GTI_VT
#include <vt_user.h>
#include <dlfcn.h>

void* theHandle = 0;

typedef void (*startP) (const char* name, const char* file, int lno);
typedef void (*endP) (const char* name);

void VT_User_start__(const char* name, const char* file, int lno)
{
	if (!theHandle)
		theHandle = dlopen ("libvt-mpi.so", RTLD_LAZY);

	startP fp = (startP) dlsym (theHandle, "VT_User_start__");
	fp (name, file, lno);
}

void VT_User_end__(const char* name)
{
	endP fp = (endP) dlsym (theHandle, "VT_User_end__");
	fp (name);
}
#endif /*GTI_VT*/


using namespace gti;

mGET_INSTANCE_FUNCTION (ThreadedMPIPlace)
mFREE_INSTANCE_FUNCTION (ThreadedMPIPlace)
mPNMPI_REGISTRATIONPOINT_FUNCTION(ThreadedMPIPlace);

GTI_RETURN getInstanceName (const char **instanceName)
{
	//get self module, get name of own instance
	int err = PNMPI_SUCCESS;
	PNMPI_modHandle_t modHandle;

	err=PNMPI_Service_GetModuleSelf(&modHandle);
	assert (err == PNMPI_SUCCESS);

	/* get own module name */
	char temp[64];
	sprintf (temp, "instanceToUse");
	err=PNMPI_Service_GetArgument(modHandle, temp, instanceName);

	if (err != PNMPI_SUCCESS)
	{
		std::cerr << "Error: threaded mpi place module needs a PnMPI argument named \"instanceToUse\" that lists a valid instance name to be used as instance." << std::endl;
		return GTI_ERROR;
	}

	return GTI_SUCCESS;
}

extern "C" int MPI_Init (int* argc, char*** argv)
{
	//init MPI first
    int err = PMPI_Init (argc, argv);

    //get name of instance to use
    const char *instanceName;
    GTI_RETURN ret = getInstanceName (&instanceName);
    assert (ret == GTI_SUCCESS);

    //Initialize and run place
    ThreadedMPIPlace* place;
    place = ThreadedMPIPlace::getInstance(instanceName);
    assert(place);
    place->run();

    return err;
}

extern "C" int MPI_Finalize (void)
{
	//get name of instance to use
	const char *instanceName;
	GTI_RETURN ret =getInstanceName (&instanceName);
	assert (ret == GTI_SUCCESS);

    ThreadedMPIPlace* place;
    place = ThreadedMPIPlace::getInstance(instanceName);
    ThreadedMPIPlace::freeInstanceForced (place);

    return PMPI_Finalize ();
}

ThreadedMPIPlace::ThreadedMPIPlace (const char* instanceName)
    : ModuleBase<ThreadedMPIPlace, I_Module> (instanceName),
      myWrapper (NULL)
{
    //add the id to all sub modules
    int rank;
    PMPI_Comm_rank (MPI_COMM_WORLD, &rank);
    char temp[64];
    sprintf (temp,"%d",rank);
    addDataToSubmodules ("id", temp);

    //create sub modules
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //save sub modules
    assert (subModInstances.size() >= 2);
    myTraceRecv = (I_CommStrategyDown*) subModInstances[0];
    myReceival = (I_PlaceReceival*) subModInstances[1];

    if (subModInstances.size() > 2)
    	myWrapper = (I_Module*) subModInstances[2];

    //trigger trace init
    setupTracing (GTI_TRACE_INIT_ROOT, NULL);
}

ThreadedMPIPlace::~ThreadedMPIPlace (void)
{
    OTF_Tracer::writeEnter (THREADED_MPI_PLACE_FUNC__Destructor);

    if (myTraceRecv)
    {
    	myTraceRecv->shutdown(GTI_FLUSH, GTI_SYNC);
    	destroySubModuleInstance ((I_Module*) myTraceRecv);
    	myTraceRecv = NULL;
    }

    if (myReceival)
    	destroySubModuleInstance ((I_Module*) myReceival);
    myReceival = NULL;

    if (myWrapper)
    	destroySubModuleInstance ((I_Module*) myWrapper);
    myWrapper = NULL;

    OTF_Tracer::writeLeave (THREADED_MPI_PLACE_FUNC__Destructor);
    OTF_Tracer::writeLeave (THREADED_MPI_PLACE_FUNC__main);
}

bool ThreadedMPIPlace::isTraceable (void)
{
    return true;
}

GTI_RETURN ThreadedMPIPlace::setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData)
{
    //This is a root module, neither supports copy, nor child
    assert (init == GTI_TRACE_INIT_ROOT);

    //Configuration data
    std::map<std::string, std::string> data = getData();
    uint32_t procID = 100; //TODO get this value from something random
    uint32_t parent = 0; //No Parent
    const char* processName = "Threaded_MPI_Place";

    //look into the module config data for data that overrides the defaults
    if (data.find("otf_proc_id") != data.end ())
        procID = atoi(data.find("otf_proc_id")->second.c_str());
    if (data.find("otf_parent") != data.end ())
        parent = atoi(data.find("otf_parent")->second.c_str());

    //Threaded MPI place adds MPI rank to proc id!
    int rank;
    PMPI_Comm_rank(MPI_COMM_WORLD,&rank);
    procID += 10*rank; //TODO "*10" to have room for extra threads ...

    //Setup tracer
    OTF_Tracer::setup (procID, parent, processName);

    //Setup Functions
    OTF_Tracer::defFunction (THREADED_MPI_PLACE_FUNC__main, "main", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (THREADED_MPI_PLACE_FUNC__setupTracing, "setupTracing", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (THREADED_MPI_PLACE_FUNC__Destructor, "~ThreadedMPIPlace", FUNCTION_GROUP_ID__PROCESSING);
    OTF_Tracer::defFunction (THREADED_MPI_PLACE_FUNC__run, "run", FUNCTION_GROUP_ID__IDLE);

    OTF_Tracer::writeEnter (THREADED_MPI_PLACE_FUNC__main);
    OTF_Tracer::writeEnter (THREADED_MPI_PLACE_FUNC__setupTracing);

    //Setup sub-modules
    if (myTraceRecv->isTraceable())
        //runs in this thread
        myTraceRecv->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    if (myReceival->isTraceable())
        //runs in this thread
        myReceival->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    if (myWrapper->isTraceable())
    	//runs in this thread
    	myWrapper->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    OTF_Tracer::writeLeave (THREADED_MPI_PLACE_FUNC__setupTracing);
    return GTI_SUCCESS;
}

void ThreadedMPIPlace::run (void)
{
	OTF_Tracer::writeEnter (THREADED_MPI_PLACE_FUNC__run);
#ifdef GTI_VT
	VT_TRACER("ThreadedMPIPlace::run");
#endif /*GTI_VT*/

	GTI_RETURN (*free_function) (void* free_data, unsigned long num_bytes, void* buf);
	int flag;
	unsigned long   size;
	void            *buf, *free_data;
	unsigned long   numClients, numChannels;
	unsigned long   numOps = 0;
	GTI_RETURN     ret;
	unsigned long fromChannel;
	SuspensionBufferTree *tree = NULL,
			*node,
			*firstSuspendedNode = NULL,
			*firstNonEmptyQ = NULL;
	unsigned long currTimeout = 0;
	const unsigned long timeoutThreshold = 5000000; //5s

#ifdef GTI_DEBUG
	//Debug
	int gRank;
	MPI_Comm_rank (MPI_COMM_WORLD, &gRank);
	char temp[256];
	sprintf (temp, "place_channel_tree_%d.dot", gRank);
	std::ofstream out (temp);
#endif

	//get number of clients
	myTraceRecv->getNumClients(&numClients);
	numChannels = numClients;
#ifdef GTI_DEBUG
	printf ("\n\nCLIENTS: %ld\n\n",numClients);
#endif

	//Loop until all channels signaled that they finished
	while (numClients)
	{
		bool receiveRecord = true; //Used to determine whehter it is necessary to receive a new record or not
		bool hasRecordToProcess = false;
		RecordInfo processRecord;

		// 1) Check whether we have any records in our queues that could be processed
		//     If so select such a record as record to process
		//================================================
		if (tree)
		{
			if (tree->getQueuedRecord (&processRecord))
			{
#ifdef GTI_VT
			VT_TRACER("ThreadedMPIPlace::run::gotQueuedRecord");
#endif /*GTI_VT*/
				//Set that there is a record to process ... (we need not receive something)
				receiveRecord = false;
				hasRecordToProcess = true;
			}
		}

		// 2) No queued records left, receive a new record
		//================================================
		if (receiveRecord)
		{
#ifdef GTI_VT
			{
			VT_TRACER("ThreadedMPIPlace::run::test");
#endif /*GTI_VT*/

			ret = myTraceRecv->test(&flag, &size, &buf, &free_data, &free_function, &fromChannel);
			//ret = myTraceRecv->wait(&size, &buf, &free_data, &free_function, &fromChannel);
#ifdef GTI_VT
			}
#endif /*GTI_VT*/

			if (ret == GTI_ERROR)
			{
				std::cerr << "Place: error while receiving, communication strategy returned error." << std::endl;
				break;
			}
			else
				if (ret == GTI_SUCCESS) //Necessary as strategy may also return it is not ready yet ...
				{
					if (flag)
					{
#ifdef GTI_VT
						VT_TRACER("ThreadedMPIPlace::run::evaluateNewRecord");
#endif /*GTI_VT*/
						I_ChannelId *channelId;

						//Get and update channel id for this record
						if (myReceival->getUpdatedChannelId(buf, size, fromChannel, numChannels, &channelId) != GTI_SUCCESS)
							break; //Abort main loop in case of some error

						//Store information on the record
						processRecord.buf = buf;
						processRecord.free_data = free_data;
						processRecord.num_bytes = size;
						processRecord.recordChannelId = channelId;
						processRecord.buf_free_function = free_function;

						//Allocate tree if this is the first channel id we got
						if (!tree) tree = new SuspensionBufferTree (channelId->getNumUsedSubIds()-1, channelId->getSubIdNumChannels(channelId->getNumUsedSubIds()-1));

						//==If we have any suspension, we must check whether it is allowed to process the record
						if (!tree->hasAnySuspension())
						{
							hasRecordToProcess = true;
						}
						else
						{
							//Find the node for this channel id in the channel id tree
							firstSuspendedNode = firstNonEmptyQ = NULL;
							node = tree->getNode(channelId, &firstSuspendedNode, &firstNonEmptyQ);

							//Is the channel of the received record ok for processing ?
							//If there is any non-empty queue, then no
							if (firstNonEmptyQ)
							{
								//If there was a non empty queue on the way, add to that node, otherwise add to the leaf node
								firstNonEmptyQ->pushBack(processRecord);

#ifdef GTI_DEBUG
								////Debug:
								//std::cout << "Test: " << processRecord.recordChannelId->toString() << std::endl;
								out << "digraph temp" << std::endl << "{" << std::endl << "myNode [label=\"Added to queue from: "<< channelId->toString() << "\"];" << std::endl << "}" << std::endl;
								tree->printAsDot(out);
#endif
							}
							else
							if (node->hasAnySuspension() || firstSuspendedNode)
							{
								//Can't process at the moment, has suspended sub-nodes or parents!
								node->pushBack(processRecord);

#ifdef GTI_DEBUG
								////Debug:
								//std::cout << "Test: " << processRecord.recordChannelId->toString() << std::endl;
								out << "digraph temp" << std::endl << "{" << std::endl << "myNode [label=\"Added to queue from(descendant blocked): "<< channelId->toString() << "\"];" << std::endl << "}" << std::endl;
								tree->printAsDot(out);
#endif
							}
							else 
                             if (node->getQueueSize() || node->getChildQueueSize())
							{
							    //Can't process at the moment, there is a queued event in the childs that needs to be processed first
							    node->pushBack(processRecord);

#ifdef GTI_DEBUG
                                ////Debug:
                                //std::cout << "Test: " << processRecord.recordChannelId->toString() << std::endl;
                                out << "digraph temp" << std::endl << "{" << std::endl << "myNode [label=\"Added to queue from(descendant has queue entry): "<< channelId->toString() << "\"];" << std::endl << "}" << std::endl;
                                tree->printAsDot(out);
#endif
							}
							else
							{
								hasRecordToProcess = true;
							}
						}//If (!tree->hasAnySuspension())
					}//test had flag=true
					else //flag
					{
#ifdef GTI_VT
						VT_TRACER("ThreadedMPIPlace::run::sleep");
#endif /*GTI_VT*/
						//we did not receive something ...
						struct timeval gStart, gEnd;
						gettimeofday(&gStart, NULL);
						usleep (10);
						gettimeofday(&gEnd, NULL);
						currTimeout += ((gEnd.tv_sec * 1000000 + gEnd.tv_usec) - (gStart.tv_sec * 1000000 + gStart.tv_usec));

						//Timeout threshold reached ?
						if (currTimeout > timeoutThreshold)
						{
							currTimeout = 0;

//							if (tree && tree->hasAnySuspension())
//							{
								myReceival->timeOutReductions();
								if (tree)
									tree->removeAllSuspensions();

#ifdef GTI_DEBUG
								//DEBUG:
								std::cout << "Timed out!" << std::endl;
								out << "digraph temp" << std::endl << "{" << std::endl << "myNode [label=\"Time out\"];" << std::endl << "}" << std::endl;
								if (tree) tree->printAsDot(out);
#endif
//							}
						}
					}//if flag
				}//return GTI_SUCCESS
				else
				{
					//Input strategy not ready yet, wait a bit ...
					usleep(500);
				}//if-else-if for return of wait call
		}//If need to receive record

		//3) Process a record if there was a valid record for processing
		//================================================
		if (hasRecordToProcess)
		{
#ifdef GTI_VT
			VT_TRACER("ThreadedMPIPlace::run::process");
#endif /*GTI_VT*/
			//reset timeout
			currTimeout = 0;

			//process
			bool wasChannelClosed = false;
			std::list<I_ChannelId*> reopenedChannels;
			std::list<I_ChannelId*>::iterator iter;
			int fromChannel = processRecord.recordChannelId->getSubId(processRecord.recordChannelId->getNumUsedSubIds()-1);

			if (myReceival->ReceiveRecord(
					processRecord.buf,
					processRecord.num_bytes,
					processRecord.free_data,
					processRecord.buf_free_function,
					&numClients,
					processRecord.recordChannelId,
					&wasChannelClosed,
					&reopenedChannels) != GTI_SUCCESS)
				break; //Abort main loop in case of some error

			myTraceRecv->acknowledge (fromChannel);

			//evaluate suspensions
			if (wasChannelClosed)
			{
				node = tree->getNode(processRecord.recordChannelId, NULL, NULL);
				node->setSuspension(true);
			}

			for (iter = reopenedChannels.begin(); iter != reopenedChannels.end(); iter++)
			{
				I_ChannelId *id = *iter;
				if (!id) continue;

				node = tree->getNode(id, NULL, NULL);
				node->setSuspension(false);
				delete (id);
			}

#ifdef GTI_DEBUG
			////Debug:
			//std::cout << "Test: " << processRecord.recordChannelId->toString() << std::endl;
			out << "digraph temp" << std::endl << "{" << std::endl << "myNode [label=\"Processed from: "<< processRecord.recordChannelId->toString() << "\"];" << std::endl << "}" << std::endl;
			tree->printAsDot(out);
#endif

			delete (processRecord.recordChannelId);
			numOps++;
		}
	}//while channels left

#ifdef GTI_DEBUG
	//Debug
	std::cout << "NumOps: " << numOps << std::endl;
#endif

	//Delete tree
	if (tree) delete (tree);

	OTF_Tracer::writeLeave (THREADED_MPI_PLACE_FUNC__run);
}//run method

/*EOF*/
