/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2010-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2010-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CompletionTree.cpp
 *       @see gti::CompletionTree
 *
 * @author Tobias Hilbrich
 * @date 17.12.2010
 */

#include <assert.h>

#include "CompletionTree.h"

using namespace gti;

//=============================
// CompletionTree
//=============================
CompletionTree::CompletionTree (int subIdIndex, long numChannels)
 : I_ChannelTree<CompletionTree> (subIdIndex, numChannels),
   myIsCompleted (false),
   myNumChildsCompleted (0)
{
	/*Nothing to do*/
}

//=============================
// ~CompletionTree
//=============================
CompletionTree::~CompletionTree ()
{
	/*Nothing to do*/
}

//=============================
// flushCompletions
//=============================
void CompletionTree::flushCompletions (void)
{
	myIsCompleted = false;
	myNumChildsCompleted = 0;

	std::map<long int, CompletionTree*>::iterator iter;
	for (iter = myChilds.begin(); iter != myChilds.end(); iter++)
	{
		iter->second->flushCompletions();
	}
}

//=============================
// isCompleted
//=============================
bool CompletionTree::isCompleted (void)
{
	if (myIsCompleted)
		return true;

	if (myNumChildsCompleted == myNumChannels && myNumChildsCompleted > 0)
		return true;

	return false;
}

//=============================
// addCompletion
//=============================
void CompletionTree::addCompletion (I_ChannelId* idCompleted)
{
	CompletionTree *node = getChildForChannel (idCompleted);

	assert (node != NULL);

	if (node == this)
	{
		myIsCompleted = true;
	}
	else
	{
		bool wasCompleted = node->isCompleted();

		node->addCompletion(idCompleted);

		bool isCompletedNow = node->isCompleted();

		if (!wasCompleted && isCompletedNow)
			myNumChildsCompleted++;
	}
}

//=============================
// getNodeColor
//=============================
std::string CompletionTree::getNodeColor (void)
{
	if (myIsCompleted)
		return "green";

	if (isCompleted())
		return "yellow";

	return "red";
}

//=============================
// wasCompleted
//=============================
bool CompletionTree::wasCompleted (I_ChannelId* id)
{
	long int subId = getChannelForId (id);

	//If this node is already completed, all sub-nodes for which the id might be will also be completed
	if (myIsCompleted)
		return true;

	//Is this the node for the id ? -> Completed if any child completed!
	if (subId == -1)
		return myNumChildsCompleted;

	//Does the child actually exists ?
	if (myChilds.find (subId) == myChilds.end())
		return false;

	//Recurse
	return getChildForChannel(id)->wasCompleted(id);
}

/*EOF*/
