/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2010-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2010-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file ModuleBase.h
 *       @see gti::ModuleBase
 *
 * @author Tobias Hilbrich
 *
 * Complete overhaul done 20.10.2010.
 *
 */

#include <mpi.h>
#include <pnmpimod.h>
#include <assert.h>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <map>
#include <list>
#include <vector>

#include "GtiDefines.h"
#include "GtiEnums.h"
#include "GtiTypes.h"

#include "I_Module.h"

#ifndef MODULE_BASE_H
#define MODULE_BASE_H

namespace gti
{
/**
 * Base class for common module behavior.
 * Includes a common initialization and destruction
 * mechanism for sub modules. As well as parsing and
 * access to module data. It uses two templates:
 * - "T" type of the final object to create; used when
 *     allocating an instance in ModuleBase::getInstance
 * - "Base" base class from which ModuleBase should
 *   inherit, e.g. I_Module. If it is a different class
 *   than I_Module, "Base" must inherit
 *   from I_Module itself or indirectly.
 *
 * See @ref ModConfPage for details.
 *
 * ModuleBase is used as follows:
 * - In PNMPI_RegistrationPoint the static member ModuleBase::readModuleInstances
 *   must be called in order to read all named instances
 *   from the PnMPI configuration. Needs to register the module with the module
 *   name given in the PnMPI configuration file. Needs to register the freeInstance
 *   and getInstance functions see below.
 * - getInstance must call ModuleBase::getInstance and pass the given instance name
 *   to it.
 * - freeInstance must call ModuleBase::freeInstance and pass the given instance
 *   pointer to it.
 * - A class using ModuleBase needs to require an instance name as an argument to
 *   its constructor.
 * - This instance name must be passed to the constructor of ModuleBase
 * - The class using ModuleBase may call:
 * -- ModuleBase::createsSubModuleInstance to initialize sub modules specified for
 *       this instance in the PnMPI configuration file. A list with pointers to the
 *       initialized sub modules is returned.
 * -- ModuleBase::destroySubModule to destroy a sub module instance.
 * -- ModuleBase::getData to receive a list of key value pairs given as data to this
 *       instance.
 *
 * Furthermore, ModuleBase provides a default implementation for the functions of
 * gti::I_Module. Defaulting to a not traceable object.
 *
 * Example:
 * @code
namespace gti
{
	class MyInstance : public ModuleBase<MyInstance, I_Module>
	{
	protected:
		std::vector<I_Module*> mySubMods;

	public:
		MyInstance (const char* InstanceName);
		~MyInstance (void);
	};
}//namespace gti

using namespace gti;

MyInstance::MyInstance (const char* InstanceName)
	: ModuleBase<MyInstance, I_Module> (InstanceName)
{
	std::cout << "Constructing instance." << std::endl;

	mySubMods = createSubModuleInstances ();
}

MyInstance::~MyInstance (void)
{
	std::cout << "Destructing instance." << std::endl;

	for (int i = 0; i < mySubMods.size(); i++)
	{
		destroySubModuleInstance (mySubMods[i]);
	}
}
	 @endcode
 *
 * Use the macros in GtiMacros.h to implement getInstance, freeInstance and
 * PnMPI_RegistrationPoint.
 *
 */
template <class T, class Base>
class ModuleBase : public Base
{
private:
	/*
	 * Typedefs.
	 */
	typedef std::map<std::string, std::string> DataMapT; /**< Type for the data map list. */
	typedef std::map<std::string, std::pair<T*, int> > ModInstancesT; /**< Type for mapping of known and used instances and their indices. */
	typedef std::list <std::pair<std::string, std::string> > SubModNamesT; /**< Type for list of sub module identifiers (pairs of ModName and InstanceName). */

	/*
	 * Static attributes.
	 * Keep in mind these two attributes need to be distinct for each type of module,
	 * as the first template parameter is going to differ for each type of module this
	 * will hold even though one might think otherwise at first.
	 */
	static ModInstancesT ourInstances; /**< Maps instance name to instance pointers and instance index, also used to store the names of defined instances.*/
	static PNMPI_modHandle_t ourModHandle; /**< The module handle for this module, set during readModuleInstances. */
	static std::string ourModName; /**< Name of this module. */
	static std::map<std::string, DataMapT > ourDataFromAncestors; /** Data passed from ancestor modules to instances of this module. */

	/*
	 * Regular attributes.
	 */
	int myRefCount; /**< Reference count for each instance of a module.*/
	SubModNamesT mySubModNames; /**< (ModuleName, InstanceName) pairs for all sub modules, first element is first sub-module.*/
	DataMapT myModData; /**< Key value pairs that describe the module data.*/
	std::string myInstanceName; /**< Name of this instance.*/

protected:
	/**
	 * Constructor.
	 * @param instanceName name of this instance.
	 */
	ModuleBase (const char* instanceName);

	/**
	 * Empty virtual destructor.
	 */
	virtual ~ModuleBase () {};

	/**
	 * Interprets the sub modules given in the PnMPI
	 * configuration file for this instance and creates
	 * all of these instances. Returns the pointer
	 * to their instances as a vector.
	 * @return vector of created instances.
	 */
	std::vector<I_Module*> createSubModuleInstances (void);

	/**
	 * Frees the given instance of a sub module.
	 * The sub-module must have been created with
	 * createSubModuleInstances.
	 * @param instance pointer to instance that shall be freed.
	 * @return GTI_SUCCESS if successfull.
	 */
	GTI_RETURN destroySubModuleInstance (I_Module* instance);

	/**
	 * Adds a specified key value pair to the module data
	 * of all sub modules and their respective sub modules
	 * (recursive).
	 * This data will only be considered if the respective
	 * sub modules where not yet created.
	 * @param key name of the data key.
	 * @param value for data key.
	 * @return GTI_SUCCESS if successful, GTI_ERROR otherwise.
	 */
	GTI_RETURN addDataToSubmodules (std::string key, std::string value);

	/**
	 * Returns the address to a function in a wrapper module.
	 * Requires a "instance<N>Wrapper" specification to be available
	 * for this module instance in the PnMPI configuration file
	 * as well as a wrapper module that specifies the named function.
	 * @param functionName name of the function to be found in the wrapper.
	 * @param pOutFunction pointer to storage for a function pointer
	 *        only set if the call is successful.
	 * @return GTI_SUCCESS if the function was found, GTI_ERROR otherwise.
	 *
	 * Example:
	 * @code
		int (*newSize) (int);
		if (getWrapperFunction ("newSize", (GTI_Fct_t**)&newSize) == GTI_SUCCESS)
		{
			for (int i = 0; i < c; i++)
				(*newSize) (array[i]);
		}
		else
		{
			std::cout << "ERROR: failed to get \"newSize\" function pointer from wrapper." << std::endl;
		}
	 @endcode
	 */
	GTI_RETURN getWrapperFunction (std::string functionName, GTI_Fct_t** pOutFunction);

public:
	/**
	 * Returns the instance with the given name.
	 * Constructs it if it is queried for the
	 * first time.
	 *
	 * SHOULD ONLY BE CALLED BY THE MACROS PROVIDED
	 * IN GtiMacros.h !
	 *
	 * @param instanceName name of the instance.
	 *        If an empty string the instance
	 *        with index 0 will be returned
	 *        (assumed there is an instance).
	 * @return the instance or NULL if an unknown
	 *         instance name was used or another
	 *         error occured.
	 */
	static T* getInstance (std::string instanceName);

	/**
	 * Frees the given instance.
	 * Destructs it when the last user of the instance
	 * calls this function.
	 *
	 * SHOULD ONLY BE CALLED BY THE MACROS PROVIDED
	 * IN GtiMacros.h !
	 *
	 * @param instance the instance to be freed.
	 * @return GTI_SUCCESS if successful.
	 */
	static GTI_RETURN freeInstance (T* instance);

	/**
	 * Frees the instance even if its reference
	 * count indicates that there are still other
	 * users of the instance.
	 *
	 * SHOULD ONLY BE CALLED BY THE MACROS PROVIDED
	 * IN GtiMacros.h !
	 *
	 * @param instance the instance to be freed.
	 * @return GTI_SUCCESS if successful.
	 */
	static GTI_RETURN freeInstanceForced (T* instance);

	/**
	 * Reads the names of all instances used for this module.
	 * Must be called when PNMPI_RegistrationPoint is entered.
	 * If it is not called no instances can be retrieved from
	 * getInstance and freeInstance.
	 *
	 * SHOULD ONLY BE CALLED BY THE MACROS PROVIDED
	 * IN GtiMacros.h !
	 *
	 * @param modHandle handle to this module.
	 * @return GTI_SUCCESS if successful.
	 */
	static GTI_RETURN readModuleInstances (PNMPI_modHandle_t modHandle);

	/**
	 * Adds data to a module instance.
	 * The data will only be considered if the named instance exists and
	 * was not yet constructed. If the data uses the same key as a
	 * previously added data item or a data item in the P^nMPI configuration
	 * than the last call to addData will set the value for this key.
	 * @param instanceName name of the module instance to add the data
	 *        to.
	 * @param key name of key used for the data.
	 * @param value of the data.
	 * @return GTI_SUCCESS if successful, GTI_ERROR otherwise.
	 */
	static GTI_RETURN addData (std::string instanceName, std::string key, std::string value);

	/**
	 * Returns a list of all currently active instances.
	 * Active refers to allocated and initialized while not
	 * yet destroyed.
	 */
	static std::map <std::string, T*> getActiveInstances (void);

	/**
	 * Returns the name of this module.
	 * Not the instance name!
	 * @return name of this module.
	 */
	std::string getName (void);

	/**
	 * Gives an overwriteable implementation of
	 * this function that defaults to false.
	 * @see I_Module::isTraceable
	 */
	virtual bool isTraceable (void);

	/**
	 * Gives an overwriteable implementation of
	 * this function that causes an assert.
	 * @see I_Module::setupTracing
	 */
	virtual GTI_RETURN setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData);

	/**
	 * Returns the data associated with this module.
	 * The data is returned as a map, were key is the
	 * data attribute name and value the data value.
	 * E.g.
	 * @code
		           ModuleBase<MyModule, I_Module> mod (INSTANCE_NAME);
		           std::map<std::string,std::string> map;
		           map = mod.getData ();
		           //check for the "ip" data attribute
		           map::iterator i = map.find ("ip");
		           if (i != map::end())
		           {
		               printf ("My module IP is: %s\n", (i->second).c_str());
		           }
		           else
		           {
		               printf ("No IP specified for this module.\n");
		           }
		   @endcode
	 *
	 * @return map for the module data attributes.
	 */
	DataMapT getData (void);

	}; /*class ModuleBase*/
}/*namespace gti*/

/*Include implementation*/
#include "ModuleBase.hxx"

#endif /* MODULE_BASE_H */
