/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2010-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2010-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file module_base.hxx
 *       @see gti::ModuleBase
 *
 * @author Tobias Hilbrich
 *
 * Complete overhaul done 20.10.2010.
 *
 */

namespace gti
{
	//=============================
	// Static attribute construction
	//=============================
	template <class T, class Base> typename ModuleBase<T, Base>::ModInstancesT ModuleBase<T, Base>::ourInstances;
	template <class T, class Base> PNMPI_modHandle_t ModuleBase<T, Base>::ourModHandle = 0; /**< TODO: PNMPI should offer a PNMPI_MODHANDLE_NULL ...*/
	template <class T, class Base> std::string ModuleBase<T, Base>::ourModName;
	template <class T, class Base> std::map<std::string, typename ModuleBase<T, Base>::DataMapT > ModuleBase<T, Base>::ourDataFromAncestors;

	//=============================
	// Constructor
	//=============================
	template <class T, class Base>
	ModuleBase<T, Base>::ModuleBase (const char* instanceName)
		: myRefCount (1),
  		  mySubModNames (),
		  myModData (),
		  myInstanceName (instanceName)
	{
		//== Tasks
		//Read PNMPI Arguments
		// -> sub modules
		// -> instance data

		int index;
		const char* argument;
		int err;
		char temp[64];

		//== Find the index of this instance
		typename ModInstancesT::iterator instancePos;
		instancePos = ourInstances.find (instanceName);
		assert (instancePos != ourInstances.end());
		index = instancePos->second.second;

		//== sub modules
		sprintf (temp, "instance%dSubMods", index);
		err=PNMPI_Service_GetArgument(ourModHandle, temp, &argument);

		if (err == PNMPI_SUCCESS)
		{
			std::string subModArg (argument);
			size_t pos = 0, posLast;

			/*
			 * subModArg is of the form:
			 * x:y,z:w,...
			 * each pair a:b is a module and instance name pair
			 *
			 * This loop is used to extract each pair.
			 */
			do
			{
				std::string pairDesc, modName, instanceName;
				size_t columnPos;

				//find pair
				posLast = pos;
				pos = subModArg.find_first_of (',', pos);

				//Use remaining string if no ',' was found
				if (pos == std::string::npos)
					pos = subModArg.length();//emulates a ',' after the actual string

				//extract substring of form:
				//MOD_NAME:INSTANCE_NAME
				pairDesc.assign (subModArg,posLast,pos-posLast);

				//find the column character and split accordingly
				columnPos = pairDesc.find_first_of (':');

				if (columnPos == std::string::npos)
				{
					std::cerr
						<< "Error: a sub module string is malformed moduleName,instanceName pair was \""
						<< pairDesc
						<< "\" it needs to be of form MOD_NAME:INSTANCE_NAME, complete sub module argument was \""
						<< subModArg
						<< "\"."
						<< std::endl;
					assert (0);
				}

				modName.assign (pairDesc, 0, columnPos);
				instanceName.assign (pairDesc, columnPos+1, pairDesc.length() - columnPos - 1);
				mySubModNames.push_back (std::make_pair (modName, instanceName));

				//next
				pos++;

			} while (pos != std::string::npos && pos < subModArg.length());
		}/*found sub module argument*/

		//== Module data
		sprintf (temp, "instance%dData", index);
		err=PNMPI_Service_GetArgument(ourModHandle, temp, &argument);

		if (err == PNMPI_SUCCESS)
		{
			std::string dataArg (argument);
			size_t pos = 0, posLast;

			/*
			 * dataArg is of the form:
			 * x=y,z=w,...
			 * each pair a=b is a key value pair
			 *
			 * This loop is used to extract each pair.
			 */
			do
			{
				std::string pairDesc, key, value;
				size_t columnPos;

				//find pair
				posLast = pos;
				pos = dataArg.find_first_of (',', pos);

				//Use remaining string if no ',' was found
				if (pos == std::string::npos)
					pos = dataArg.length();//emulates a ',' after the actual string

				//extract substring of form:
				//MOD_NAME:INSTANCE_NAME
				pairDesc.assign (dataArg,posLast,pos-posLast);

				//find the column character and split accordingly
				columnPos = pairDesc.find_first_of ('=');

				if (columnPos == std::string::npos)
				{
					std::cerr
					<< "Error: a sub module string is malformed moduleName,instanceName pair was \""
					<< pairDesc
					<< "\" it needs to be of form MOD_NAME:INSTANCE_NAME, complete sub module argument was \""
					<< dataArg
					<< "\"."
					<< std::endl;
					assert (0);
				}

				key.assign (pairDesc, 0, columnPos);
				value.assign (pairDesc, columnPos+1, pairDesc.length() - columnPos - 1);
				myModData.insert (std::make_pair(key, value));

				//next
				pos++;

			} while (pos != std::string::npos && pos < dataArg.length());
		}

		//== Add data given from ancestors
		DataMapT extraData =  ourDataFromAncestors[instanceName];
		DataMapT::iterator dataIter;

		for (dataIter = extraData.begin(); dataIter != extraData.end(); dataIter++)
		{
			if (myModData.find (dataIter->first) == myModData.end())
				myModData.insert (*dataIter);
			else
				myModData[dataIter->first] = dataIter->second;
		}

		//== Add ancestor data to my sub modules
		for (dataIter = extraData.begin(); dataIter != extraData.end(); dataIter++)
		{
			addDataToSubmodules (dataIter->first, dataIter->second);
		}

/*		//DEBUG
		std::cerr << "constructed: " << getName() << ":" << myInstanceName << std::endl;
		for (dataIter = myModData.begin(); dataIter != myModData.end(); dataIter++)
		{
			std::cerr << "-->" << dataIter->first << "=" << dataIter->second << std::endl;
		}
*/
	}

	//=============================
	// getInstance
	//=============================
	template <class T, class Base>
	T* ModuleBase<T, Base>::getInstance (std::string instanceName)
	{
		typename ModInstancesT::iterator pos;

		/*Dummy call -> will cause init of PnMPI if that hasn't happened yet ...*/
		PNMPI_modHandle_t modHandle;
		PNMPI_Service_GetModuleSelf(&modHandle);

		//If name is an empty string, we return the first instance
		if (instanceName == "")
		{
			for (pos = ourInstances.begin(); pos != ourInstances.end(); pos++)
			{
				if (pos->second.second == 0)
				{
					instanceName = pos->first;
					break;
				}
			}
		}

		//find instance name
		pos = ourInstances.find (instanceName);

		//instance name known ?
		if (pos == ourInstances.end())
		{
			std::cerr << "Unknown instance name \"" << instanceName << "\"" << std:: endl;

			std::cerr << "Known instances: " << std::endl;
			for (pos = ourInstances.begin(); pos != ourInstances.end(); pos++)
			{
				std::cerr << " * " << pos->first << std::endl;
			}

			return NULL;
		}

		//existing instance ?
		if (pos->second.first != NULL)
			pos->second.first->myRefCount = pos->second.first->myRefCount + 1;
		else
			pos->second.first = new T (instanceName.c_str());

/*//DEBUG
		{
		typename ModInstancesT::iterator iter;

		for (iter = ourInstances.begin(); iter != ourInstances.end(); iter++)
				{
		std::cerr << "-->: " << (unsigned long long) iter->second.first << std::endl;
				}
		}
*/
		return pos->second.first;
	}

	//=============================
	// freeInstance
	//=============================
	template <class T, class Base>
	GTI_RETURN ModuleBase<T, Base>::freeInstance (T* instance)
	{
		assert (instance);

		typename ModInstancesT::iterator iter;

/*//DEBUG
std::cerr << "freeInstance: " << (unsigned long long) instance << " numInstances: " << ourInstances.size() << " (" << instance->getName() << ")" << std::endl;
*/
		for (iter = ourInstances.begin(); iter != ourInstances.end(); iter++)
		{
/*//DEBUG
std::cerr << "-->: " << (unsigned long long) iter->second.first << std::endl;
*/
			if (!iter->second.first)
				continue;

			if (iter->second.first->myInstanceName != instance->myInstanceName)
				continue;

			instance->myRefCount = instance->myRefCount - 1;

			if (instance->myRefCount == 0)
			{
				//Instance was consumed, completely remove it from registration
				ourInstances.erase (iter);

				delete instance;
			}

			return GTI_SUCCESS;
		}

		//error if not found
		return GTI_ERROR;
	}

	//=============================
	// freeInstanceForced
	//=============================
	template <class T, class Base>
	GTI_RETURN ModuleBase<T, Base>::freeInstanceForced (T* instance)
	{
		assert (instance);
		instance->myRefCount = 1;
		return freeInstance (instance);
	}

	//=============================
	// readModuleInstances
	//=============================
	template <class T, class Base>
	GTI_RETURN ModuleBase<T, Base>::readModuleInstances (PNMPI_modHandle_t modHandle)
	{
		int err;
		const char *argument;
		const char *modName;
		int numInstances;
		char temp[64];

		// store our static module handle
		ourModHandle = modHandle;

		// get own module name
		sprintf (temp, "moduleName");
		err=PNMPI_Service_GetArgument(modHandle, temp, &modName);
		assert (err == PNMPI_SUCCESS);
		ourModName = modName;

		// get number of instances
		sprintf (temp, "numInstances");
		err=PNMPI_Service_GetArgument(modHandle, temp, &argument);

		if (err != PNMPI_SUCCESS)
		{
			//No instances given
			std::cerr << "Warning: module named \"" << modName << "\" has no numInstances argument, thus it can't be instantiated." << std::endl;
			return GTI_SUCCESS;
		}

		numInstances = atoi (argument);

		// read all the instance names
		for (int i = 0; i < numInstances; i++)
		{
			sprintf (temp, "instance%d", i);

			// Read name of instance
			err=PNMPI_Service_GetArgument(modHandle, temp, &argument);

			if (err != PNMPI_SUCCESS)
			{
				std::cerr << "Error: module named \"" << modName << "\" specifies " << numInstances << " instances but no name for instance of index " << i << " is given." << std::endl;
				return GTI_ERROR;
			}

			// Add to map of known instances
			ourInstances.insert (std::make_pair(argument, std::make_pair((T*)NULL, i)));

			// Add an item to static attribute for extra module instance data
			std::map<std::string, std::string> empty;
			ourDataFromAncestors.insert (std::make_pair(argument, empty));
		}

		return GTI_SUCCESS;
	}

	//=============================
	// addData
	//=============================
	template <class T, class Base>
	GTI_RETURN ModuleBase<T, Base>::addData (
			std::string instanceName,
			std::string key,
			std::string value)
	{
		typename std::map<std::string, DataMapT >::iterator pos;

		//Find map entry for instance
		pos = ourDataFromAncestors.find (instanceName);

		if (pos == ourDataFromAncestors.end())
		{
			std::cerr << "In ModuleBase::addData: invalid instance Name given. (name=" << instanceName << ")" << std::endl;
			return GTI_ERROR;
		}

		//Find entry for key (if present) and set value
		DataMapT::iterator i;

		i = pos->second.find (key);

		if (i == pos->second.end())
			pos->second.insert (std::make_pair (key, value));
		else
			(pos->second)[key] = value;

		return GTI_SUCCESS;
	}

	//=============================
	// getActiveInstances
	//=============================
	template <class T, class Base>
	std::map <std::string, T*> ModuleBase<T, Base>::getActiveInstances (void)
	{
		std::map <std::string, T*> ret;

		typename ModInstancesT::iterator i;

		for (i = ourInstances.begin(); i != ourInstances.end(); i++)
		{
			if (i->second.first) //only insert if active !
				ret.insert (std::make_pair (i->first, i->second.first));
		}

		return ret;
	}

	//=============================
	// createSubModuleInstances
	//=============================
	template <class T, class Base>
	std::vector<I_Module*> ModuleBase<T, Base>::createSubModuleInstances (void)
	{
		PNMPI_modHandle_t handle;
		PNMPI_Service_descriptor_t service;
		PNMPI_Service_Fct_t fct;
		int err;
		std::vector<I_Module*> subModInstances;

		//Create the given sub modules and initialize them
		for (SubModNamesT::iterator i = mySubModNames.begin(); i != mySubModNames.end(); i++)
		{
#ifdef PNMPI_FIXED
			err = PNMPI_Service_GetModuleByName(i->first.c_str(), &handle);

#else
			char string[512];
			sprintf (string, "%s",i->first.c_str());
			err = PNMPI_Service_GetModuleByName(string, &handle);
#endif
			if (err != PNMPI_SUCCESS)
			{
				std::cerr
				<< "Failed to get a handle for the P^nMPI module \""
				<< i->first
				<< "\""
				<< std::endl
				<< "(Failed in module:instance \""
				<< getName()
				<< ":"
				<< myInstanceName
				<< "\")"
				<< std::endl;
				assert (0);
			}

#ifdef PNMPI_FIXED
			err = PNMPI_Service_GetServiceByName(handle, "getInstance", "pp", &service);
#else
			char string2[3];
			sprintf (string2, "pp");
			sprintf (string, "getInstance");
			err = PNMPI_Service_GetServiceByName(handle, string, string2, &service);
#endif
			assert (err == PNMPI_SUCCESS);

			I_Module* instance;
			((getInstance_t) service.fct) (&instance, i->second.c_str());
			subModInstances.push_back(instance);
		}

		return subModInstances;
	}

	//=============================
	// destroySubModuleInstance
	//=============================
	template <class T, class Base>
	GTI_RETURN ModuleBase<T, Base>::destroySubModuleInstance (I_Module* instance)
	{
		PNMPI_modHandle_t handle;
		PNMPI_Service_descriptor_t service;
		PNMPI_Service_Fct_t fct;
		int err;

/*//DEBUG
std::cerr << "Calling destroySubModuleInstance: " << (unsigned long long) instance << std::endl;
*/
#ifdef PNMPI_FIXED
		err = PNMPI_Service_GetModuleByName(instance->getName().c_str(), &handle);

#else
		char string[512];
		sprintf (string, "%s",((I_Module*)instance)->getName().c_str());
		err = PNMPI_Service_GetModuleByName(string, &handle);
#endif
		assert (err == PNMPI_SUCCESS);

#ifdef PNMPI_FIXED
		err = PNMPI_Service_GetServiceByName(handle, "freeInstance", "p", &service);
#else
		char string2[2];
		sprintf (string2, "p");
		sprintf (string, "freeInstance");
		err = PNMPI_Service_GetServiceByName(handle, string, string2, &service);
#endif

		assert (err == PNMPI_SUCCESS);
		((freeInstance_t) service.fct) (instance);

		return GTI_SUCCESS;
	}

	//=============================
	// getData
	//=============================
	template <class T, class Base>
	typename ModuleBase<T, Base>::DataMapT ModuleBase<T, Base>::getData (void)
	{
		return myModData;
	}

	//=============================
	// addDataToSubmodules
	//=============================
	template <class T, class Base>
	GTI_RETURN ModuleBase<T, Base>::addDataToSubmodules (std::string key, std::string value)
	{
		SubModNamesT::iterator i;

		for (i = mySubModNames.begin(); i != mySubModNames.end(); i++)
		{
			PNMPI_modHandle_t handle;
			PNMPI_Service_descriptor_t service;
			PNMPI_Service_Fct_t *fct;
			int err;

			std::string
				moduleName = i->first,
				instanceName = i->second;

#ifdef PNMPI_FIXED
			err = PNMPI_Service_GetModuleByName(moduleName.c_str(), &handle);

#else
			char string[512];
			sprintf (string, "%s",moduleName.c_str());
			err = PNMPI_Service_GetModuleByName(string, &handle);
#endif
			if (err != PNMPI_SUCCESS)
			{
				std::cerr
				<< "Failed to get a handle for the P^nMPI module \""
				<< moduleName
				<< "\""
				<< std::endl
				<< "(Failed in module:instance \""
				<< getName()
				<< ":"
				<< myInstanceName
				<< "\")"
				<< std::endl;
				assert (0);
			}

#ifdef PNMPI_FIXED
			err = PNMPI_Service_GetServiceByName(handle, "addDataHandler", "ppp", &service);
#else
			char string2[4];
			sprintf (string2, "ppp");
			sprintf (string, "addDataHandler");
			err = PNMPI_Service_GetServiceByName(handle, string, string2, &service);
#endif
			assert (err == PNMPI_SUCCESS);

			((addDataHandler_t) service.fct) (instanceName.c_str(), key.c_str(), value.c_str());
		}

		return GTI_SUCCESS;
	}

	//=============================
	// getWrapperFunction
	//=============================
	template <class T, class Base>
	GTI_RETURN ModuleBase<T, Base>::getWrapperFunction (
			std::string functionName,
			GTI_Fct_t** pOutFunction)
	{
		PNMPI_modHandle_t handle;
		PNMPI_Service_descriptor_t service;
		PNMPI_Service_Fct_t *fct;
		int err;
		int instanceIndex;
		const char *wrappModName;

		if (pOutFunction) *pOutFunction = NULL;

		//Get module handle
#ifdef PNMPI_FIXED
		err = PNMPI_Service_GetModuleByName(ourModName.c_str(), &handle);
#else
		char temp[64];
		sprintf (temp, "%s", ourModName.c_str());
		err = PNMPI_Service_GetModuleByName(temp, &handle);
#endif
		if (err != PNMPI_SUCCESS) return GTI_ERROR;

		//Get own module index
		typename ModInstancesT::iterator pos = ourInstances.find (myInstanceName);
		assert (pos != ourInstances.end()); //should never happen
		instanceIndex = pos->second.second;

		//Get wrapper argument
		char wrapperArg[128];
		sprintf (wrapperArg, "instance%dWrapper", instanceIndex);
		err=PNMPI_Service_GetArgument(handle, wrapperArg, &wrappModName);
		if (err != PNMPI_SUCCESS) return GTI_ERROR;

		//Get wrapper module
		err = PNMPI_Service_GetModuleByName(wrappModName, &handle);

		//Get "getFunction" service
#ifdef PNMPI_FIXED
		err = PNMPI_Service_GetServiceByName(handle, "getFunction", "pp", &service);
#else
		char temp2[3];
		sprintf (temp2, "pp");
		sprintf (temp, "getFunction");
		err = PNMPI_Service_GetServiceByName(handle, temp, temp2, &service);
#endif
		if (err != PNMPI_SUCCESS) return GTI_ERROR;

		//Call getFunction to get symbol
		assert (err == PNMPI_SUCCESS);
		if (((getFunction_t) service.fct) (functionName.c_str(), pOutFunction) != PNMPI_SUCCESS)
			return GTI_ERROR;

		return GTI_SUCCESS;
	}

	//=============================
	// getName
	//=============================
	template <class T, class Base>
	std::string ModuleBase<T, Base>::getName (void)
	{
		return ourModName;
	}

	//=============================
	// isTraceable
	//=============================
	template <class T, class Base>
	bool ModuleBase<T, Base>::isTraceable (void)
	{
		return false;
	}

	//=============================
	// setupTracing
	//=============================
	template <class T, class Base>
	GTI_RETURN ModuleBase<T, Base>::setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData)
	{
		//must not be called as isTraceable returns false per default
		assert (0);
		return GTI_ERROR;
	}
} /*namespace gti*/

/*EOF*/
