/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file OtfTracer.cpp
 *       Implementation for the OTF tracer.
 *
 * @author Tobias Hilbrich
 */

#include <assert.h>
#include <stdio.h>
#include "OtfTracer.h"
#include <sys/time.h>
#include <time.h>
#include "GtiTraceDefines.h"


using namespace gti;

OTF_Tracer::OTF_Tracer (void)
{
    myManager   = NULL;
    myWStream   = NULL;
    myProcessId = 0;
    myIsCopy    = true;
    myIsInitialized = false;

    struct tm *ltime;
    struct timeval sTime;
    uint64_t time;

    gettimeofday(&sTime, NULL);
    ltime = localtime (&sTime.tv_sec);

    myBaseTime = ltime->tm_yday*86400 +
         (ltime->tm_year-70)*31536000 + ((ltime->tm_year-69)/4)*86400 -
        ((ltime->tm_year-1)/100)*86400 + ((ltime->tm_year+299)/400)*86400;
}

OTF_Tracer::~OTF_Tracer (void)
{
    if (!myIsCopy)
    {
        if (myWStream)
            OTF_WStream_close (myWStream);
        myWStream = NULL;

        if (myManager)
            OTF_FileManager_close (myManager);
        myManager = NULL;
    }
}

uint32_t OTF_Tracer::getProcessId (void)
{
    return myProcessId;
}

GTI_RETURN OTF_Tracer::setup (uint32_t processId, uint32_t parent, const char* processName)
{
	//If already initialized, do nothing
	if (myIsInitialized)
		return GTI_SUCCESS;

    myIsInitialized = true;
    myIsCopy    = false;
    myManager   = OTF_FileManager_open (100);
    myWStream   = OTF_WStream_open ("must_otf_trace", processId, myManager);
    myProcessId = processId;

    OTF_WStream_writeDefTimerResolution (myWStream, 1000000); /*Ticks per second*/

    OTF_WStream_writeDefProcess (
            myWStream,
            myProcessId,      /*my process id*/
            processName,      /*Name*/
            parent       );   /*Parent*/

    //Setup function groups for new tracer
    OTF_Tracer::defFunctionGroup (FUNCTION_GROUP_ID__COMMUNICATE, FUNCTION_GROUP_NAME__COMMUNICATE);
    OTF_Tracer::defFunctionGroup (FUNCTION_GROUP_ID__IDLE, FUNCTION_GROUP_NAME__IDLE);
    OTF_Tracer::defFunctionGroup (FUNCTION_GROUP_ID__PROCESSING, FUNCTION_GROUP_NAME__PROCESSING);
    OTF_Tracer::defFunctionGroup (FUNCTION_GROUP_ID__APP_MPI, FUNCTION_GROUP_NAME__APP_MPI);
    OTF_Tracer::defFunctionGroup (FUNCTION_GROUP_ID__APP_PROCESSING, FUNCTION_GROUP_NAME__APP_PROCESSING);

    return GTI_SUCCESS;
}

GTI_RETURN OTF_Tracer::setup (OTF_Tracer* copy_from)
{
	//If already initialized, do nothing
	if (myIsInitialized)
		return GTI_SUCCESS;

    myIsCopy    = true;
    myManager   = copy_from->myManager;
    myWStream   = copy_from->myWStream;
    myProcessId = copy_from->myProcessId;
    myIsInitialized = true;

    return GTI_SUCCESS;
}

GTI_RETURN OTF_Tracer::defFunctionGroup (uint32_t id, const char* name)
{
    if (!myIsInitialized)
        return GTI_SUCCESS;

    OTF_WStream_writeDefFunctionGroup (
            myWStream, id, name);

    return GTI_SUCCESS;
}

GTI_RETURN OTF_Tracer::defFunction (uint32_t id, const char* name, uint32_t group_id)
{
    if (!myIsInitialized)
        return GTI_SUCCESS;

    OTF_WStream_writeDefFunction (
            myWStream, id, name, group_id,
            0 /*Source*/);

    return GTI_SUCCESS;
}

GTI_RETURN OTF_Tracer::writeEnter (uint32_t function_id)
{
    if (!myIsInitialized)
        return GTI_SUCCESS;

    struct timeval sTime;
    uint64_t time;

    gettimeofday(&sTime, NULL);
    time = sTime.tv_sec * 1000000 + sTime.tv_usec - myBaseTime* 1000000;

    OTF_WStream_writeEnter (
            myWStream, time, function_id, myProcessId,
            0 /*Source Code Ref*/);

    return GTI_SUCCESS;
}

GTI_RETURN OTF_Tracer::writeLeave (uint32_t function_id)
{
    if (!myIsInitialized)
        return GTI_SUCCESS;

    struct timeval sTime;
    uint64_t time;

    gettimeofday(&sTime, NULL);
    time = sTime.tv_sec * 1000000 + sTime.tv_usec - myBaseTime* 1000000;

    OTF_WStream_writeLeave (
            myWStream, time, function_id, myProcessId,
            0 /*Source Code Ref*/);

    return GTI_SUCCESS;
}

GTI_RETURN OTF_Tracer::writeSendMsg (uint32_t receiver, uint32_t tag, uint32_t length)
{
    if (!myIsInitialized)
        return GTI_SUCCESS;

    struct timeval sTime;
    uint64_t time;

    gettimeofday(&sTime, NULL);
    time = sTime.tv_sec * 1000000 + sTime.tv_usec - myBaseTime* 1000000;

    OTF_WStream_writeSendMsg (
            myWStream, time, myProcessId, receiver,
            0, /*process group, 0 = no group*/
            tag, length,
            0 /*source*/);

    return GTI_SUCCESS;
}

GTI_RETURN OTF_Tracer::writeRecvMsg (uint32_t sender, uint32_t tag, uint32_t length)
{
    if (!myIsInitialized)
        return GTI_SUCCESS;

    struct timeval sTime;
    uint64_t time;

    gettimeofday(&sTime, NULL);
    time = sTime.tv_sec * 1000000 + sTime.tv_usec - myBaseTime* 1000000;

    OTF_WStream_writeRecvMsg (
            myWStream, time, myProcessId, sender,
            0, /*process group, 0 = no group*/
            tag, length,
            0 /*source*/);

    return GTI_SUCCESS;
}

/*EOF*/
