/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file OtfTracer.h
 *       Implements functionality for module activity tracing.
 *
 * @author Tobias Hilbrich
 */

#include "GtiDefines.h"
#include "GtiEnums.h"
#include "GtiTypes.h"

#ifdef USE_OTF
#include <otf.h>
#endif

#ifndef OTF_TRACER_H
#define OTF_TRACER_H

namespace gti
{
#ifndef USE_OTF
    typedef unsigned int uint32_t;
#endif

    /**
     * Class for OTF based activity tracing.
     * @todo obviously there is a lack of a tracing
     *       functionality interface, something like
     *       that should be implemented in case there
     *       is a need for a second tracing mechanism.
     */
    class OTF_Tracer
    {
    protected:
#ifdef USE_OTF
        /** The OTF file manager. */
        OTF_FileManager*    myManager;
        /** The OTF stream. */
        OTF_WStream*        myWStream;
        /** My process ID. */
        uint32_t            myProcessId;
        /** Is it a copy ?. */
        bool                myIsCopy;
        /** Is it initialized ? */
        bool                myIsInitialized;

        uint64_t            myBaseTime;
#endif

    public:
        /** Constructor. */
        OTF_Tracer (void);
        /** Destructor. */
        virtual ~OTF_Tracer (void);

        /**
         * Returns the process ID associated with this
         * instance.
         * @return process id associated with this instance.
         */
        uint32_t getProcessId (void);

        /**
         * Sets-up the OTF tracer as new process using
         * given process id.
         * @param processId the id of the new process.
         * @param parent the parent processID.
         * @param processName the name of this process.
         * @return @see GTI_RETURN
         */
        GTI_RETURN setup (uint32_t processId, uint32_t parent, const char* processName);

        /**
         * Sets-up the OTF tracer as a copy of an already
         * initialized tracer.
         * @param setup an already initialized tracer.
         * @return @see GTI_RETURN
         */
        GTI_RETURN setup (OTF_Tracer* setup);

        /**
         * Creates a function group with the given id.
         * The requested id must not be 0.
         * @param id the id used for the new function group.
         * @param name the name of the new function group.
         * @return GTI_SUCCESS if successful.
         */
        GTI_RETURN defFunctionGroup (uint32_t id, const char* name);

        /**
         * Creates a new function with the given id.
         * The requested id must not be 0.
         * @param id the id used for the new function.
         * @param name the name of the new function.
         * @param group_id the function group for the new
         *        function, 0 for default group.
         * @return GTI_SUCCESS if successful.
         */
        GTI_RETURN defFunction (uint32_t id, const char* name, uint32_t group_id);

        /**
         * Creates an enter event.
         * @param function_id the id of the function being entered.
         * @return GTI_SUCCESS if successful.
         */
        GTI_RETURN writeEnter (uint32_t function_id);

        /**
         * Creates a leave event.
         * @param function_id the id of the function being left.
         * @return GTI_SUCCESS if successful.
         */
        GTI_RETURN writeLeave (uint32_t function_id);

        /**
         * Creates a send message event.
         * Should be called between the enter and leave of the
         * sending function.
         * @param receiver the OTF process_id of the receiver.
         * @param tag used to perform message matching, use this
         *        value to differentiate distinct non-matching messages.
         * @param length the data length being send.
         * @todo do we need to make a process group specifiable ?
         */
        GTI_RETURN writeSendMsg (uint32_t receiver, uint32_t tag, uint32_t length);

        /**
         * Creates a receive message event.
         * Should be called between the enter and leave of the
         * receive function.
         * @param sender the OTF process_id of the sender.
         * @param tag used to perform message matching, use this
         *        value to differentiate distinct non-matching messages.
         * @param length the data length being received.
         * @see writeSendMsg
         */
        GTI_RETURN writeRecvMsg (uint32_t sender, uint32_t tag, uint32_t length);

    }; /*class OTF_Tracer*/
} /*namespace gti*/

#endif /* OTF_TRACER_H */
