/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2010-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2010-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file ModuleConfig.cpp
 * 		@see gti::weaver::generation::ModuleConfig
 *
 * @author Tobias Hilbrich
 * @date 1.11.2010
 */

#include <fstream>
#include <assert.h>

#include "ModuleConfig.h"
#include "ApiCalls.h"

using namespace gti::weaver::generation;

//=============================
// ModuleConfig
//=============================
ModuleConfig::ModuleConfig (
		Level *level,
		std::string fileName)
	: GenerationBase (level, fileName, std::ios_base::out | std::ios_base::app)
{
	/*
	 * Do the generation directly in the constructor.
	 */
	out
		<< "<level index=\"" << myOrder << "\" size=\"" << mySize << "\">" << std::endl;

    //---------------------------------
	// Gather information on analyses and comms
	//---------------------------------
	if (!calculateComms())
		return;
	if (!calculateAnalyses())
		return;

	//---------------------------------
	// Print information on all the modules
	//---------------------------------
	out
		<< "\t<modules>" << std::endl;

	//== Place (if not application)
	if (myOrder != 0)
		printModule (
				myPlace,
				myPlace,
				"place",
				"place");

	//== Incoming comm strategy + protocol (if not application)
	if (myInList.size() > 0)
	{
		printAdjacency (myInList[0], "incoming_", "protocol-down", true);
	}

	//== Wrapper
	std::string wrapperConfigName = myWrapperSourceName;
	size_t pos = wrapperConfigName.find_last_of(".");
	wrapperConfigName.replace (pos, wrapperConfigName.length()-pos, "");
	Module wrapper ("wrapper", (std::string)"lib" + wrapperConfigName);

	printModule (
			&wrapper,
			"wrapper",
			"wrapper");

	//== Receival (if not application)
	std::string receivalConfigName = myReceivalSourceName;
	pos = receivalConfigName.find_last_of(".");
	receivalConfigName.replace (pos, receivalConfigName.length()-pos, "");
	Module receival ("receival", (std::string)"lib" + receivalConfigName);

	if (myOrder != 0)
		printModule (
				&receival,
				"receival",
				"receival");

	//== Analyses
	/*
	 * The myAnalysisMap has analysis functions, we
	 * need analysis modules here, as a module may have
	 * multiple functions we must make sure to not
	 * add duplicates. We simply identify analysis modules
	 * with the id of their FIRST analysis function that we
	 * encountered.
	 *
	 * Order is crucial here, the order in which we list
	 * dependencies must reflect what the receival/wrapper
	 * generators do, so we need to stick to the order in
	 * myAnalysisMap!
	 */
	std::map<Analysis*, int>::iterator anIter;
	std::map<AnalysisModule*, int> usedModules; //map of analysis modules and their ids
	std::map<AnalysisModule*, int>::iterator uAnIter;
	std::list<AnalysisModule*> orderedModules;
	std::list<AnalysisModule*>::iterator orderedIter;

	for (anIter = myAnalysisMap.begin(); anIter != myAnalysisMap.end(); anIter++)
	{
		Analysis *a = anIter->first;
		int id = anIter->second;

		//Already printed this module ?
		if (usedModules.find (a->getModule()) != usedModules.end())
			continue;

		usedModules.insert (std::make_pair(a->getModule(), id));
		orderedModules.push_back (a->getModule());

		char anName[128];
		sprintf (anName, "analysis%d", id);

		printModule (
				a->getModule(),
				"analysis",
				anName);
	}

	//== Outgoing communications
	std::map <Adjacency*, int>::iterator commIter;

	for (commIter = myComms.begin(); commIter != myComms.end(); commIter++)
	{
		Adjacency *a = commIter->first;
		int commId = commIter->second;
		char prefix[128];
		sprintf (prefix, "outgoing%d_", commId);

		printAdjacency (a, prefix, "protocol-up", false);
	}

	out
		<< "\t</modules>" << std::endl;

	//---------------------------------
	// Print information on module relationships
	//---------------------------------
	out
		<< "\t<relationships>" << std::endl;

	//== place
	//   -> incoming_strategy
	//   -> receival
	//   -> wrapper
	if (myOrder != 0)
	{
		out
			<< "\t\t<relationship user=\"place\">incoming_strategy</relationship>" << std::endl
			<< "\t\t<relationship user=\"place\">receival</relationship>" << std::endl
			<< "\t\t<relationship user=\"place\">wrapper</relationship>" << std::endl;
	}

	//== incoming_strategy
	//   -> incoming_protocol
	if (myOrder != 0)
	{
		out << "\t\t<relationship user=\"incoming_strategy\">incoming_protocol</relationship>" << std::endl;
	}

	//== receival/wrapper
	//   -> All strats: outgoing<i>_strategy
	//   -> All analyses: analysis<i>
	for (commIter = myComms.begin(); commIter != myComms.end(); commIter++)
	{
		Adjacency *a = commIter->first;
		int commId = commIter->second;
		char stratName[128];
		sprintf (stratName, "outgoing%d_strategy", commId);

		if (myOrder != 0)
			out << "\t\t<relationship user=\"receival\">" << stratName << "</relationship>" << std::endl;

		out << "\t\t<relationship user=\"wrapper\">" << stratName << "</relationship>" << std::endl;
	}

	for (orderedIter = orderedModules.begin(); orderedIter != orderedModules.end(); orderedIter++)
	{
		int id = usedModules[*orderedIter];

		char anName[128];
		sprintf (anName, "analysis%d", id);

		if (myOrder != 0)
			out << "\t\t<relationship user=\"receival\">" << anName << "</relationship>" << std::endl;

		out	<< "\t\t<relationship user=\"wrapper\">" << anName << "</relationship>" << std::endl;
	}

	//== All strats: outgoing<i>_strategy
	//   -> outgoing<i>_protocol
	for (commIter = myComms.begin(); commIter != myComms.end(); commIter++)
	{
		Adjacency *a = commIter->first;
		int commId = commIter->second;
		char stratName[128];
		char protName[128];
		sprintf (protName, "outgoing%d_protocol", commId);
		sprintf (stratName, "outgoing%d_strategy", commId);

		out	<< "\t\t<relationship user=\"" << stratName << "\">" << protName << "</relationship>" << std::endl;
	}

	//== Analysis dependencies
	// For this loop order of modules is of no importance
	for (uAnIter = usedModules.begin(); uAnIter != usedModules.end(); uAnIter++)
	{
		AnalysisModule *module = uAnIter->first;
		std::list<AnalysisModule*> deps = module->getDependencies();
		std::list<AnalysisModule*>::iterator depIter;

		int id = uAnIter->second;

		//Here we need to stick to the order given in the dependencies
		for (depIter = deps.begin(); depIter != deps.end(); depIter++)
		{
			AnalysisModule *depMod = *depIter;

			//Check whether this is just a soft dependency, if so the dependent module is no child of this module!
			if (module->isSoftDependency(depMod))
				continue;

			assert (usedModules.find (depMod) != usedModules.end());

			int idDep = usedModules[depMod];
			out	<< "\t\t<relationship user=\"analysis" << id << "\">analysis" << idDep << "</relationship>" << std::endl;
		}
	}

	out
		<< "\t</relationships>" << std::endl;

	//---------------------------------
	// Finalize output
	//---------------------------------
	out
		<< "\t</level>" << std::endl;
}

//=============================
// ~ModuleConfig
//=============================
ModuleConfig::~ModuleConfig (void)
{
	/*Nothing to do*/
}

//=============================
// printAdjacency
//=============================
bool ModuleConfig::printAdjacency (
		Adjacency *a,
		std::string moduleNamePrefix,
		std::string protocolType,
		bool isDown)
{
	//Strategy
	Module *strat = a->getComm()->getCommStrategy();

	if (isDown)
		strat = a->getComm()->getCommStrategy()->getDownModule();

	if (!printModule (
			strat,
			"strategy",
			moduleNamePrefix + "strategy",
			a->getComm()->getStrategySettings()))
		return false;

	//Protocol
	if (!printModule (
			a->getComm()->getCommProtocol(),
			a->getComm()->getCommProtocol(),
			protocolType,
			moduleNamePrefix + "protocol",
			a->getComm()->getProtocolSettings(),
			true,
			a->getTarget()->getOrder()))
		return false;
	return true;
}

//=============================
// printModule
//=============================
bool ModuleConfig::printModule (
	Module *module,
	std::string type,
	std::string name,
	std::vector<Setting*> settings,
	bool needsToLevel,
	int toLevel)
{
	out
		<< "\t\t<module "
		<< "type=\"" << type << "\" "
		<< "name=\"" << name << "\" "
		<< "pnmpi-module-name=\"" << module->getConfigName() << "\"";

	if (needsToLevel)
		out << " to-level=\"" << toLevel << "\"";

	out
		<< ">" << std::endl
		<< "\t\t\t<data>" << std::endl;

    for (int i = 0; i < settings.size(); i++)
    {
    	Setting* s = settings[i];

    	out << "\t\t\t\t<setting key=\"" << s->getName() << "\">" << s->getValue() << "</setting>" << std::endl;
    }

	out
		<< "\t\t\t</data>" << std::endl
		<< "\t\t</module>" << std::endl;

	return true;
}

//=============================
// printModule
//=============================
bool ModuleConfig::printModule (
	Module *module,
	Prepended *prepModule,
	std::string type,
	std::string name,
	std::vector<Setting*> settings,
	bool needsToLevel,
	int toLevel)
{
	static int prepId = 0;

	//print prepended modules
	std::list<Module*> modules = prepModule->getPrependedModules();
	std::list<Module*>::iterator i;

	for (i = modules.begin(); i != modules.end(); i++)
	{
		char temp[128];
		sprintf (temp, "prepended%d", prepId);
		prepId++;

		printModule (*i, "analysis", temp);
	}

	//print the actual module
	printModule (module, type, name, settings, needsToLevel, toLevel);

	return true;
}

/*EOF*/
