/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2010-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2010-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file WrapperGenerator.cpp
 *		@see gti::codegen::WrapperGenerator
 *
 * @author Tobias Hilbrich
 * @date 11.08.2010
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iostream>
#include <list>
#include <assert.h>

#include "Verbose.h"

#include "WrapperGenerator.h"

using namespace gti::codegen;

/**
 * WrapperGenerator main function.
 */
int main (int argc, char** argv)
{
	//=========== 0 read input or print help ===========
	//print help if requested
	if ((argc == 2) &&
		(
				(strncmp(argv[1], "--help", strlen("--help")) == 0 ) ||
				(strncmp(argv[1], "-help", strlen("--help")) == 0 ) ||
				(strncmp(argv[1], "-h", strlen("--help")) == 0 )
		)
	)
	{
		printUsage (argv[0], std::cout);
		return 0;
	}

	//enough arguments ?
	if (argc < 2)
	{
		std::cerr << "Error: Not enough arguments!" << std::endl << std::endl;
		printUsage (argv[0], std::cerr);
		return 1;
	}

	std::string fileName = argv[1];
	int retVal = 0;
	WrapperGenerator generator (fileName, &retVal);

	return retVal;
}

//=============================
// printUsage
//=============================
void printUsage (std::string execName, std::ostream &out)
{
	out
		<< "Usage: "
		<< execName
		<< " <WrapperGeneratorInputXML>" << std::endl
		<< std::endl
		<< "E.g.: " << execName << " wrappGenIn.xml"
		<< std::endl;
}

//=============================
// myGetGeneratorName
//=============================
std::string WrapperGenerator::myGetGeneratorName (void)
{
	return "wrapper generator";
}

//=============================
// myGetRootNodeName
//=============================
std::string WrapperGenerator::myGetRootNodeName (void)
{
	return "wrapper-specification";
}

//=============================
// WrapperGenerator
//=============================
WrapperGenerator::WrapperGenerator (std::string inputFile, int* retVal)
	: 	GeneratorBase (),
	  	myOrder (0)
{
	VERBOSE(1) << "Processing input file " << inputFile << " ..." << std::endl;

	if (retVal)
		*retVal = 1;

	//Open the input XML
	//=================================
	SpecificationNode currentPointer;
	SpecificationNode child;

	currentPointer = openInput (inputFile);
	if (!currentPointer) return;
	std::string orderStr;
	if (!currentPointer.getAttributeOrErr("order", "|  |--> Error: root node has no \"order\" attribute.", &orderStr))
		return;
	myOrder = atoi(orderStr.c_str());

	//Read: settings
	//=================================
	VERBOSE(1) << "|- Reading settings ..." << std::endl;
	child = currentPointer.findChildNodeNamedOrErr(
			"settings",
			"|  |--> Error: root node has no \"settings\" child.");
	if (!readSettings (child))
		return;

	VERBOSE(1) << "|  |--> SUCCESS" << std::endl;

	//Get record generation implementation
	//=================================
	if (!getRecordGenerationImplementation(&myImpl))
		return;

	//Init record generation
	//=================================
	std::list<std::string>	sources1,
							sources2,
							libs,
							headers;

	myImpl->initBegin(myOutDir,mySourceName);
	myImpl->initGetIntermediateSourceFileNames (&sources1);
	myImpl->initGetStaticSourceFileNames (&sources2);
	myImpl->initGetLibNames (&libs);
	myImpl->initGetHeaderFileNames (&myRecordHeaders);
	myImpl->initEnd ();

	//Read: headers
	//=================================
	VERBOSE(1) << "|- Reading headers ..." << std::endl;
	child = currentPointer.findChildNodeNamedOrErr(
			"headers",
			"|  |--> Error: root node has no \"headers\" child.");
	if (!readAndPrintHeaders (child))
		return;

	VERBOSE(1) << "|  |--> SUCCESS" << std::endl;

	//Read: communications
	//=================================
	VERBOSE(1) << "|- Reading communications ..." << std::endl;
	child = currentPointer.findChildNodeNamedOrErr(
			"communications",
			"|  |--> Error: root node has no \"communications\" child.");
	if (!readCommunications (child))
		return;

	VERBOSE(1) << "|  |--> SUCCESS" << std::endl;

	//Read: analyses
	//=================================
	VERBOSE(1) << "|- Reading analyses ..." << std::endl;
	child = currentPointer.findChildNodeNamedOrErr(
			"analyses",
			"|  |--> Error: root node has no \"analyses\" child.");
	if (!readAnalyses (child))
		return;

	VERBOSE(1) << "|  |--> SUCCESS" << std::endl;

	//Write basic parts of source
	//=================================
	if (!writeSourceHeaderBasics())
		return;

	//Read: calls
	//=================================
	VERBOSE(1) << "|- Reading calls ..." << std::endl;
	child = currentPointer.findChildNodeNamedOrErr(
			"calls",
			"|  |--> Error: root node has no \"calls\" child.");
	if (!readCalls (child))
		return;

	VERBOSE(1) << "|  |--> SUCCESS" << std::endl;

	//Write bye bye to the sources
	mySourceOut << std::endl << "/*EOF*/" << std::endl << std::endl;

	//Clean up
	//=================================
	VERBOSE(1) << "--> SUCCESS" << std::endl;
	if (retVal)
		*retVal = 0;

	//Write output XML
	//=================================
	std::list<std::string>::iterator lIter;

	myLogOut
		<< "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
		<< "<wrappgen-output>" << std::endl
		<< "\t<extra-sources>" << std::endl;

	for (lIter = sources1.begin(); lIter != sources1.end(); lIter++)
	{
		myLogOut << "<source>" << *lIter << "</source>" << std::endl;
	}

	for (lIter = sources2.begin(); lIter != sources2.end(); lIter++)
	{
		myLogOut << "<source>" << *lIter << "</source>" << std::endl;
	}

	myLogOut
		<< "\t</extra-sources>" << std::endl
		<< "\t<extra-libs>" << std::endl;

	for (lIter = libs.begin(); lIter != libs.end(); lIter++)
	{
		myLogOut << "<lib>" << *lIter << "</lib>" << std::endl;
	}

	myLogOut
		<< "\t</extra-libs>" << std::endl
		<< "</wrappgen-output>";
}

//=============================
// ~WrapperGenerator
//=============================
WrapperGenerator::~WrapperGenerator (void)
{

}

//=============================
// writeSourceHeaderBasics
//=============================
bool WrapperGenerator::writeSourceHeaderBasics (void)
{
	mySourceOut
		<< "using namespace gti;"
		<< "" << std::endl
		<< "extern \"C\" int getFunctionGenWrapper_" << myClassName << " (const char* functionName, GTI_Fct_t** pOutFunctionAddr);" << std::endl
		<< "" << std::endl
		<< "mGET_INSTANCE_FUNCTION (GenWrapper_" << myClassName << ");" << std::endl
		<< "mFREE_INSTANCE_FUNCTION_FORCED (GenWrapper_" << myClassName << ");" << std::endl
		<< "mPNMPI_REGISTRATIONPOINT_FUNCTION_WRAPPER (GenWrapper_" << myClassName << ");" << std::endl
		<< "" << std::endl
		<< "GenWrapper_" << myClassName << "::GenWrapper_" << myClassName << " (const char* instanceName)" << std::endl
		<< "    : ModuleBase<GenWrapper_" << myClassName << ", I_Module> (instanceName)" << std::endl
		<< "{" << std::endl
		<< "    //add the id to all sub modules" << std::endl
		<< "    //TODO" << std::endl
		<< "    //int rank;" << std::endl
		<< "    //PMPI_Comm_rank (MPI_COMM_WORLD, &rank);" << std::endl
		<< "    //char temp[64];" << std::endl
		<< "    //sprintf (temp,\"%d\",rank);" << std::endl
		<< "    //addDataToSubmodules (\"id\", temp);" << std::endl
		<< "" << std::endl
		<< "    //create sub modules" << std::endl
		<< "    std::vector<I_Module*> subModInstances;" << std::endl
		<< "    subModInstances = createSubModuleInstances ();" << std::endl
		<< "" << std::endl
		<< "    //save sub modules" << std::endl;

	//Add module saving
	//0) check enough modules given
	mySourceOut
		<< "    assert (subModInstances.size() == " << myCommMods.size() + myAnalysisMods.size() << ");" << std::endl;

	//1) comm modules
	std::map <int, std::string>::iterator commIter;
	int i = 0;
	for (commIter = myCommMods.begin(); commIter != myCommMods.end(); commIter++, i++)
	{
		int id = commIter->first;
		std::string name = commIter->second;

		mySourceOut	<< "    myCStrats[" << i << "] = (I_CommStrategyUp*) subModInstances[" << i << "];" << std::endl;
	}

	//2) analyses
	std::list <ModuleInfo*>::iterator anIter;
	for (anIter = myAnalysisMods.begin(); anIter != myAnalysisMods.end(); anIter++, i++)
	{
		ModuleInfo* info = *anIter;

		mySourceOut << "    analysis" << info->index << " = (" << info->datatype << "*) subModInstances[" << i << "];" << std::endl;
	}

	//3) profiling variables for profiling (if enabled)
	if (myProfiling)
	{
	    std::map <std::string, AnalysisFunction>::iterator anFnIter;
	    for (anFnIter = myAnalyses.begin(); anFnIter != myAnalyses.end();anFnIter++)
	    {
	        mySourceOut << "    " << getProfilingVariableName(&(anFnIter->second)) << " = 0;" << std::endl;
	        mySourceOut << "    " << getProfilingCountVariableName(&(anFnIter->second)) << " = 0;" << std::endl;
	    }
	    mySourceOut << "    usec_communicating = 0;" << std::endl;
	    mySourceOut << "    count_communicating = 0;" << std::endl;
	}

	mySourceOut
		<< "" << std::endl
		<< "    myInWrapper=false;" << std::endl
		<< "" << std::endl
		<< "    //trigger trace init" << std::endl;

	//Only as Application layer the wrapper is root module
	if (myOrder == 0)
		mySourceOut << "    setupTracing (GTI_TRACE_INIT_ROOT, NULL);" << std::endl;

	mySourceOut
		<< "}" << std::endl
		<< std::endl
		<< "bool GenWrapper_" << myClassName << "::isTraceable (void)" << std::endl
		<< "{" << std::endl
		<< "    return true;" << std::endl
		<< "}" << std::endl
		<< std::endl
		<< "GenWrapper_" << myClassName << "::~GenWrapper_" << myClassName << " (void)"  << std::endl
		<< "{" << std::endl;

	//1) comm modules
	i = 0;
	for (commIter = myCommMods.begin(); commIter != myCommMods.end(); commIter++, i++)
	{
		int id = commIter->first;
		std::string name = commIter->second;

		mySourceOut
			<< "    myCStrats[" << i << "]->shutdown(GTI_FLUSH, GTI_SYNC);" << std::endl
			<< "    destroySubModuleInstance ((I_Module*) myCStrats[" << i << "]);" << std::endl;

	}

	//2) analyses
	for (anIter = myAnalysisMods.begin(); anIter != myAnalysisMods.end(); anIter++, i++)
	{
		ModuleInfo *info = *anIter;

		mySourceOut
			<< "    destroySubModuleInstance ((I_Module*) analysis" << info->index << ");" << std::endl;
	}

	//3) profiling: print the results (if enabled)
	if (myProfiling)
	{
	    mySourceOut
	    << " std::stringstream stream;" << std::endl
	    << " stream << \"place_wrapper_profile_\" << getpid () << \".profile\";" << std::endl
	    << " std::ofstream profileOut;" << std::endl
	    << " profileOut.open(stream.str().c_str());" << std::endl;

	    std::map <std::string, AnalysisFunction>::iterator anFnIter;
	    for (anFnIter = myAnalyses.begin(); anFnIter != myAnalyses.end();anFnIter++)
	    {
	        mySourceOut << "   profileOut << \"" << anFnIter->second.info->name << "::" << anFnIter->second.function << " \" <<" << getProfilingVariableName(&(anFnIter->second)) << " << \" \" << " << getProfilingCountVariableName(&(anFnIter->second)) << "  << std::endl;" << std::endl;
	    }
	    mySourceOut
            << "   profileOut << \"Communication \" << usec_communicating << \" \" << count_communicating << std::endl;" << std::endl
            << "   profileOut.close();" << std::endl;
	}

	mySourceOut
		<< "}" << std::endl
		<< std::endl
		<< "uint64_t GenWrapper_" << myClassName << "::getUsecTime (void)" <<  std::endl
		<< "{" << std::endl
		<< "    struct timeval t;" << std::endl
		<< "    gettimeofday(&t, NULL);" << std::endl
		<< "    return t.tv_sec * 1000000 + t.tv_usec;" << std::endl
		<< "}" << std::endl
		<< "" << std::endl
		<< "GTI_RETURN GenWrapper_" << myClassName << "::setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData)" << std::endl
		<< "{" << std::endl
		<< "    //Init based on given init type" << std::endl
		<< "    if (init == GTI_TRACE_INIT_COPY)" << std::endl
		<< "    {" << std::endl
		<< "        OTF_Tracer::setup ((OTF_Tracer*) pSetupData);" << std::endl
		<< "    }" << std::endl
		<< "    else if (init == GTI_TRACE_INIT_CHILD)" << std::endl
		<< "    {" << std::endl
		<< "        uint32_t    parent = *((uint32_t*)pSetupData)," << std::endl
		<< "        proc_id = parent;" << std::endl
		<< "        //Configuration data" << std::endl
		<< "        std::map<std::string, std::string> data = getData();" << std::endl
		<< "        if (data.find(\"otf_proc_offset\") != data.end ())" << std::endl
		<< "            proc_id += atoi(data.find(\"otf_proc_offset\")->second.c_str());" << std::endl
		<< "        else" << std::endl
		<< "            proc_id += 1; //TODO should warn or abort ?" << std::endl
		<< "" << std::endl
		<< "        OTF_Tracer::setup (proc_id, parent, \"GenWrapper\");" << std::endl
		<< "    }" << std::endl
		<< "    else if (init == GTI_TRACE_INIT_ROOT)" << std::endl
		<< "    {" << std::endl
		<< "    	//Configuration data" << std::endl
		<< "    	std::map<std::string,std::string> data = getData();" << std::endl
		<< "    	uint32_t procID = 100; //TODO get this value from something random" << std::endl
		<< "    	uint32_t parent = 0; //No Parent" << std::endl
		<< "    	const char* processName = \"Gen_Wrapper\";" << std::endl
		<< "" << std::endl
		<< "    	//look into the module config data for data that overrides the defaults" << std::endl
		<< "    	if (data.find(\"otf_proc_id\") != data.end ())" << std::endl
		<< "       	    procID = atoi(data.find(\"otf_proc_id\")->second.c_str());" << std::endl
		<< "    	if (data.find(\"otf_parent\") != data.end ())" << std::endl
		<< "        	parent = atoi(data.find(\"otf_parent\")->second.c_str());" << std::endl
		<< "" << std::endl
		<< "   		//Wrapper adds MPI rank to proc id!" << std::endl
		<< "    	//TODO use PID ?" << std::endl
		<< "    	int rank = 0;" << std::endl
		<< "    	int flag;" << std::endl
		<< "    	PMPI_Initialized (&flag);" << std::endl
		<< "    	if (flag)" << std::endl
		<< "        	PMPI_Comm_rank(MPI_COMM_WORLD,&rank);" << std::endl
		<< "    	else " << std::endl
		<< "        	std::cout << \"Warning: MPI not initialized yet OTF tracing may fail or cause errors!\" << std::endl;" << std::endl
		<< "    	procID += 10*rank; //TODO \"*10\" to have room for extra threads ..." << std::endl
		<< "" << std::endl
		<< "    	//Setup tracer" << std::endl
		<< "    	OTF_Tracer::setup (procID, parent, processName);" << std::endl
		<< "    } /*INIT_ROOT*/" << std::endl
		<< "" << std::endl
		<< "    //Setup Functions" << std::endl
		<< "    OTF_Tracer::defFunction (WRAPPER__main, \"Application\", FUNCTION_GROUP_ID__PROCESSING);" << std::endl
		<< "    OTF_Tracer::defFunction (WRAPPER__APP_PROCESSING, \"App\", FUNCTION_GROUP_ID__APP_PROCESSING);" << std::endl
		<< "    OTF_Tracer::defFunction (WRAPPER__setupTracing, \"setupTracing\", FUNCTION_GROUP_ID__PROCESSING);" << std::endl
		<< "    OTF_Tracer::defFunction (WRAPPER__Destructor, \"Destructor\", FUNCTION_GROUP_ID__PROCESSING);" << std::endl
		<< "    OTF_Tracer::defFunction (WRAPPER__MPI_Abort, \"MPI_Abort\", FUNCTION_GROUP_ID__PROCESSING);" << std::endl
		<< "    //TODO list all wrapped calls" << std::endl
		<< "" << std::endl
		<< "    OTF_Tracer::writeEnter (WRAPPER__main);" << std::endl
		<< "    OTF_Tracer::writeEnter (WRAPPER__setupTracing);" << std::endl
		<< "" << std::endl
		<< "    //Setup sub-modules" << std::endl;

	//==setup tracing for sub modules
	//1) comm modules
	i = 0;
	for (commIter = myCommMods.begin(); commIter != myCommMods.end(); commIter++, i++)
	{
		int id = commIter->first;
		std::string name = commIter->second;

		mySourceOut
			<< "    if (myCStrats[" << i << "]->isTraceable())" << std::endl
			<< "        //runs in this thread" << std::endl
			<< "        myCStrats[" << i << "]->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);" << std::endl;
	}

	//2) analyses
	for (anIter = myAnalysisMods.begin(); anIter != myAnalysisMods.end(); anIter++, i++)
	{
		ModuleInfo *info = *anIter;

		mySourceOut
			<< "    if (analysis" << info->index << "->isTraceable())" << std::endl
			<< "        //runs in this thread" << std::endl
			<< "        analysis" << info->index << "->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);" << std::endl;
	}

	mySourceOut
		<< "" << std::endl
		<< "    OTF_Tracer::writeLeave (WRAPPER__setupTracing);" << std::endl
		<< "" << std::endl
		<< "    OTF_Tracer::writeEnter (WRAPPER__APP_PROCESSING);" << std::endl
		<< "" << std::endl
		<< "    return GTI_SUCCESS;" << std::endl
		<< "}" << std::endl
		<< "" << std::endl
		<< "GTI_RETURN free_serialized_buf (void* free_data, unsigned long num_bytes, void* buf)" << std::endl
		<< "{" << std::endl
		<< "    if (!free_data)" << std::endl
		<< "    {" << std::endl
		<< "        delete[] ((char*)buf);" << std::endl
		<< "        return GTI_SUCCESS;" << std::endl
		<< "    }" << std::endl
		<< std::endl
		<< "    int *refCount = (int*) free_data;" << std::endl
		<< "    *refCount = *refCount - 1;" << std::endl
		<< "    if (*refCount == 0)" << std::endl
		<< "    {" << std::endl
		<< "        free ((char*)buf);" << std::endl
///*C stuff in record gen uses malloc ...*/		<< "        delete[] ((char*)buf);" << std::endl
		<< "        delete refCount;" << std::endl
		<< "    }" << std::endl
		<< std::endl
		<< "    return GTI_SUCCESS;" << std::endl
		<< "}" << std::endl << std::endl;

	//Stuff for header
	myHeaderOut
		<< "#include \"GtiDefines.h\"" << std::endl
		<< "#include \"GtiEnums.h\"" << std::endl
		<< "#include \"GtiMacros.h\"" << std::endl
		<< "#include \"GtiTypes.h\"" << std::endl
		<< "#include \"GtiTraceDefines.h\"" << std::endl
		<< "#include \"I_Module.h\"" << std::endl
		<< "#include \"ModuleBase.h\"" << std::endl
		<< "#include \"OtfTracer.h\"" << std::endl
		<< "#include <iostream>" << std::endl
		<< "#include <stdio.h>" << std::endl
		<< "#include <stdlib.h>" << std::endl
		<< "#include <sstream>" << std::endl
		<< "#include <fstream>" << std::endl
        << "#include <sys/time.h>" << std::endl;


	std::list<std::string>::iterator headerIter;
	for (headerIter = myRecordHeaders.begin(); headerIter != myRecordHeaders.end(); headerIter++)
	{
		myHeaderOut << "#include \"" << *headerIter <<  "\"" << std::endl;
	}

	myHeaderOut
		<< "" << std::endl
		<< "namespace gti" << std::endl
		<< "{" << std::endl
		<< "    class GenWrapper_" << myClassName << " : public ModuleBase<GenWrapper_" << myClassName << ", I_Module>, public OTF_Tracer" << std::endl
		<< "    {" << std::endl
		<< "    public:" << std::endl
		<< "        bool myInWrapper;" << std::endl;

	//1) comm modules
	myHeaderOut	<< "        I_CommStrategyUp* myCStrats[" << myCommMods.size() << "];" << std::endl;

	//2) analyses
	for (anIter = myAnalysisMods.begin(); anIter != myAnalysisMods.end(); anIter++)
	{
		ModuleInfo *info = *anIter;

		myHeaderOut << "        " << info->datatype << "* analysis" << info->index << ";" << std::endl;
	}

	//3) profiling variables for profiling (if enabled)
	if (myProfiling)
	{
	    std::map <std::string, AnalysisFunction>::iterator anFnIter;
	    for (anFnIter = myAnalyses.begin(); anFnIter != myAnalyses.end();anFnIter++)
	    {
	        myHeaderOut << "        uint64_t " <<getProfilingVariableName(&(anFnIter->second)) << ";" << std::endl;
	        myHeaderOut << "        uint64_t " <<getProfilingCountVariableName(&(anFnIter->second)) << ";" << std::endl;
	    }
	    myHeaderOut << "        uint64_t usec_communicating;" << std::endl;
	    myHeaderOut << "        uint64_t count_communicating;" << std::endl;
	}

	myHeaderOut
		<< "" << std::endl
		<< "    public:" << std::endl
		<< "        /**" << std::endl
		<< "         * Constructor." << std::endl
		<< "         * Sets up the this tool place with all its modules." << std::endl
		<< "         * @param instanceName name of this module instance." << std::endl
		<< "         */" << std::endl
		<< "        GenWrapper_" << myClassName << " (const char* moduleConf);" << std::endl
		<< "" << std::endl
		<< "        /**" << std::endl
		<< "         * Destructor." << std::endl
		<< "         */" << std::endl
		<< "        ~GenWrapper_" << myClassName << " (void);" << std::endl
		<< "" << std::endl
		<< "        /**" << std::endl
		<< "         * @see I_Module::isTraceable" << std::endl
		<< "         */" << std::endl
		<< "        bool isTraceable (void);" << std::endl
		<< "" << std::endl
		<< "        /**" << std::endl
		<< "         * @see I_Traceable::setupTracing" << std::endl
		<< "         */" << std::endl
		<< "        GTI_RETURN setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData);" << std::endl
		<< "" << std::endl
		<< "        uint64_t getUsecTime (void);" <<  std::endl
		<< "    };" << std::endl
		<< "}" << std::endl;

	return true;
}

//=============================
// readCalls
//=============================
bool WrapperGenerator::readCalls (SpecificationNode node)
{
	SpecificationNode child;
	std::list <std::string> wrappedCalls; //used to generate code that returns pointers to wrapped calls

	//==Childs: call
	child = node.findChildNodeNamed("call");

	while (child)
	{
		std::string callName;

		if (!readCall (child, &callName))
			return false;

		wrappedCalls.push_back (callName);

		//next
		child = child.findSiblingNamed("call");
	}

	//==Create the "getFunction" function that returns addresses
	//  to wrapper functions
	mySourceOut
		<< std::endl
		<< "extern \"C\" int getFunctionGenWrapper_" << myClassName << " (const char* functionName, GTI_Fct_t** pOutFunctionAddr)" << std::endl
		<< "{" << std::endl;

	std::list <std::string>::iterator iter;
	for (iter = wrappedCalls.begin(); iter != wrappedCalls.end(); iter++)
	{
		if (iter != wrappedCalls.begin())
			mySourceOut << "    else" << std::endl;

		mySourceOut
			<< "    if ((std::string)functionName == (std::string)\"" << *iter << "\")" << std::endl
			<< "    {" << std::endl
			<< "        *pOutFunctionAddr = (GTI_Fct_t*) &" << *iter << ";" << std::endl
			<< "        return PNMPI_SUCCESS;" << std::endl
			<< "    }" << std::endl;
	}

	mySourceOut
		<< "    " << std::endl
		<< "    return PNMPI_NOSERVICE;" << std::endl
		<< "}" << std::endl
		<< std::endl;

	return true;
}

//=============================
// readCall
//=============================
bool WrapperGenerator::readCall (SpecificationNode node, std::string *pOutCallName)
{
	SpecificationNode child, subchild, subsubchild;
	std::string retType;
	std::string name;
	bool isFinalizer = false;
	std::string isFinalizerString;
	std::list< std::pair<std::string, std::string> > argsnTypes; // List:(type,arg)
	std::list< std::pair<std::string, std::string> >::iterator iter;
	std::list<std::string> typesAfterArg;
	std::list<std::string>::iterator afterIter;

	//==Optional attribute: is-finalizer
	if (node.getAttribute("is-finalizer", &isFinalizerString))
	{
		if (isFinalizerString == "yes")
			isFinalizer = true;

		if (isFinalizerString != "yes" && isFinalizerString != "no")
		{
			std::cerr << "|  |-->Error: a call uses the \"is-finalizer\" attribute with an unknown value \"" << isFinalizerString << "\", valid values are \"yes\" and \"no\"." << std::endl;
			return false;
		}
	}

	//==Child: return-type
	child = node.findChildNodeNamedOrErr(
			"return-type",
			"|  |-->Error: a call node has no \"return-type\" child.");
	if (!child) return false;
	retType = child.getNodeContent();

	//==Child: call-name
	child = node.findChildNodeNamedOrErr(
			"call-name",
			"|  |-->Error: a call node has no \"call-name\" child.");
	if (!child) return false;
	name = child.getNodeContent();

	if (pOutCallName)
		*pOutCallName = name;

	//==Child: arguments and argument*
	child = node.findChildNodeNamedOrErr(
			"arguments",
			"|  |-->Error: a call node has no \"arguments\" child.");
	if (!child) return false;

	subchild = child.findChildNodeNamed("argument");

	while (subchild)
	{
		std::string type,arg, typeAfterArg = "";

		//==Attribute: typeAfterArg
		subchild.getAttribute("typeAfterArg", &typeAfterArg);
		typesAfterArg.push_back (typeAfterArg);

		//==Child: type
		subsubchild = subchild.findChildNodeNamedOrErr(
				"type",
				"|  |-->Error: an argument node has no \"type\" child.");
		if (!subsubchild) return false;
		type = subsubchild.getNodeContent();

		//==Child: arg
		subsubchild = subchild.findChildNodeNamedOrErr(
				"arg",
				"|  |-->Error: an argument node has no \"arg\" child.");
		if (!subsubchild) return false;
		arg = subsubchild.getNodeContent();

		//Add to list of arguments
		argsnTypes.push_back (std::make_pair(type,arg));

		//next
		subchild = subchild.findSiblingNamed("argument");
	}

	//==Open up the call!
	mySourceOut
		<< "extern \"C\" " << retType << " " << name << " (";
	afterIter = typesAfterArg.begin();
	for (iter = argsnTypes.begin(); iter != argsnTypes.end(); iter++,afterIter++)
	{
		if (iter != argsnTypes.begin())
			mySourceOut << ", ";
		mySourceOut << iter->first << " " << iter->second << *afterIter;
	}
	mySourceOut
		<< ")" << std::endl
		<< "{" << std::endl
		<< "    " << retType << " ret" << name << ";" << std::endl
		<< "    bool " << name << "wrapped = false;" << std::endl
		<< "" << std::endl
		<< "    GenWrapper_" << myClassName << "* place;" << std::endl
		<< "    place = GenWrapper_" << myClassName << "::getInstance(\"\");" << std::endl
		<< "" << std::endl
	    << "    /*********** Place alive ? && Recursion Guard ***********/" << std::endl
	    << "    if (place";

	//TODO recursion guards disabled, do we need them ? Will PnMPI handle this for MPI ?
	//if (??) mySourceOut << " && !place->myInWrapper";

	mySourceOut
	    << ")" << std::endl
		<< "    {" << std::endl
	    << "        place->myInWrapper = true;" << std::endl
		<< "" << std::endl
		<< "#ifdef GTI_VERBOSE" << std::endl
		<< "        std::cout << \"Entering " << name << "\" << std::endl;" << std::endl
		<< "#endif" << std::endl
	    << "" << std::endl;

	//==Child: pre
	mySourceOut << "        /*********** Pre ***********/" << std::endl;
	child = node.findChildNodeNamedOrErr(
			"pre",
			"|  |-->Error: a call node has no \"pre\" child.");
	if (!child) return false;

	//=Childs: (source-piece|exec-analysis)*
	if (!printAnalysesPieces (child))
		return false;

	//=Sub-Child: forwarding
	subchild = child.findChildNodeNamedOrErr(
			"forwarding",
			"|  |-->Error: a pre node has no \"forwarding\" child.");
	if (!subchild) return false;
	if (!printForwarding(subchild, "place->", mySourceOut)) return false;

	//==If this is a finalizer we shutdown here !
	//// Before the actual call
	if (isFinalizer)
	{
		mySourceOut
			<< "    //This call finalizes, we shutdown now !" << std::endl
			<< "    GenWrapper_" << myClassName << "::freeInstanceForced (place);" << std::endl
			<< "    place = NULL;" << std::endl
			<< std::endl;
	}

	//==Do the actual call
	mySourceOut
		<< std::endl << "    /*********** Actual Call ***********/" << std::endl
		<< "    ret" << name << " = P" << name << "(";
	for (iter = argsnTypes.begin(); iter != argsnTypes.end(); iter++)
	{
		if (iter != argsnTypes.begin())
			mySourceOut << ", ";
		mySourceOut << iter->second;
	}
	mySourceOut << ");" << std::endl;

	//==If this is a finalizer we must not do anything in Post
	//place is now NULL!
	if (isFinalizer)
	{
		mySourceOut << "    return ret" << name << ";" << std::endl;
	}

	//==Child: post
	mySourceOut << std::endl << "        /*********** Post ***********/" << std::endl;
	child = node.findChildNodeNamedOrErr(
			"post",
			"|  |-->Error: a call node has no \"post\" child.");
	if (!child) return false;

	//=Childs: (source-piece|exec-analysis)*
	if (!printAnalysesPieces (child))
		return false;

	//=Sub-Child: forwarding
	subchild = child.findChildNodeNamedOrErr(
			"forwarding",
			"|  |-->Error: a post node has no \"forwarding\" child.");
	if (!subchild) return false;
	if (!printForwarding(subchild, "place->", mySourceOut)) return false;

	//==Child: cleanup
	child = node.findChildNodeNamedOrErr(
			"cleanup",
			"|  |-->Error: a call node has no \"cleanup\" child.");
	if (!child) return false;

	subchild = child.findChildNodeNamed("source-piece");
	while (subchild)
	{
		if (!printSourcePiece (subchild))
			return false;

		//next
		subchild = subchild.findSiblingNamed("source-piece");
	}

	//==Close the if wrapped part
	mySourceOut
		<< std::endl
		<< "#ifdef GTI_VERBOSE" << std::endl
		<< "        std::cout << \"Leaving " << name << "\" << std::endl;" << std::endl
		<< "#endif" << std::endl
		<< std::endl
		<< "        place->myInWrapper = false;" << std::endl
		<< "    } /*End place alive and recursion guard (if used)*/" << std::endl;


	//==Do the actual call for not wrapped case
	mySourceOut
		<< "    else" << std::endl
		<< "    {" << std::endl
		<< "        ret" << name << " = P" << name << "(";
	for (iter = argsnTypes.begin(); iter != argsnTypes.end(); iter++)
	{
		if (iter != argsnTypes.begin())
			mySourceOut << ", ";
		mySourceOut << iter->second;
	}
	mySourceOut << ");" << std::endl;


	//==Close the call
	mySourceOut
		<< "    }"<< std::endl
		<< std::endl
		<< "    return ret" << name << ";" << std::endl
		<< "}" << std::endl << std::endl;

	return true;
}

//=============================
// printAnalysesPieces
//=============================
bool WrapperGenerator::printAnalysesPieces (SpecificationNode node)
{
	SpecificationNode child;

	std::list<std::string> toFind;
	toFind.push_back("source-piece");
	toFind.push_back("exec-analysis");
	std::string got;

	child = node.findChildNodeNamed (toFind, &got);

	while (child)
	{
		if (got == "source-piece")
		{
			if (!printSourcePiece (child))
				return false;
		}
		else
		{
			if (!printExecAnalysis(child))
				return false;
		}

		//next
		child = child.findSiblingNamed(toFind, &got);
	}

	return true;
}

//=============================
// printSourcePiece
//=============================
bool WrapperGenerator::printSourcePiece (SpecificationNode node)
{
	mySourceOut
		<< "        //source-piece" << std::endl
		<< "        " << node.getNodeContent()
		<< std::endl
		<< std::endl;
	return true;
}

//=============================
// printExecAnalysis
//=============================
bool WrapperGenerator::printExecAnalysis (SpecificationNode node)
{
    static int x = 0;//For profiling

	SpecificationNode child, subchild;
	std::string analysisId;
	AnalysisFunction info;

	//==Child: analysis-id
	child = node.findChildNodeNamedOrErr(
			"analysis-id",
			"|  |-->Error: an exec-analysis node has no \"analysis-id\" child.");
	if (!child) return false;
	analysisId = child.getNodeContent();

	if (myAnalyses.find(analysisId) == myAnalyses.end())
	{
		std::cerr << "|  |-->Error: an exec-analysis node specifies an invalid analysis-id, no analysis with id \"" << analysisId << "\" known." << std::endl;
		return false;
	}

	info = myAnalyses[analysisId];

	//Start timing (if profiling)
	if (myProfiling)
	{
	    mySourceOut
	        << "uint64_t tStart_" << getProfilingVariableName(&info) << x << " = place->getUsecTime();" << std::endl;
	}

	mySourceOut
		<< "        //exec-analysis" << std::endl
		<< "        place->analysis" << info.info->index << "->" << info.function << "(";

	//==Child: inputs and input*
	child = node.findChildNodeNamedOrErr(
			"inputs",
			"|  |-->Error: an exec-analysis node has no \"inputs\" child.");
	if (!child) return false;

	subchild = child.findChildNodeNamed("input");

	bool isFirst = true;
	while (subchild)
	{
		if (!isFirst)
			mySourceOut << ", ";
		if (isFirst)
			isFirst= false;

		mySourceOut << subchild.getNodeContent();

		//next
		subchild = subchild.findSiblingNamed("input");
	}

	if (info.needsChannelId)
	{
		if (!isFirst)
			mySourceOut << ", ";

		mySourceOut << "NULL";
	}

	mySourceOut << ");" << std::endl << std::endl;

	//Stop timing (if profiling)
	if (myProfiling)
	{
	    mySourceOut
	        << "place->" << getProfilingVariableName(&info) << " += place->getUsecTime() - tStart_" << getProfilingVariableName(&info)<< x << ";" << std::endl
	        << "place->" << getProfilingCountVariableName(&info) << " += 1;" << std::endl;
	    x++;
	}

	return true;
}

//=============================
// myGetRecordName
//=============================
std::string WrapperGenerator::myGetRecordName (void)
{
	return "recordIn";
}

/*EOF*/
