/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file test_cstrat_b.cxx
 *       A module that tests communication strategies and a
 *       communication protocol.
 *
 * @data 24.04.2009
 * @author Tobias Hilbrich
 */

#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <pnmpimod.h>
#include <stdlib.h>
#include <string.h>
#include "GtiTraceDefines.h"
#include "GtiDefines.h"
#include "GtiEnums.h"
#include "GtiTypes.h"
#include "I_Module.h"
#include "ModuleBase.h"
#include "OtfTracer.h"
#include "I_CommStrategyDown.h"
#include "I_CommStrategyUp.h"
#include "GtiMacros.h"

namespace gti
{
    /**
     * Class used to drive the comm strat & comm protocol
     * testing.
     */
    class TestCStratB : public ModuleBase<TestCStratB, I_Module>, public OTF_Tracer
    {
    protected:
        I_CommStrategyUp *strat;

    public:
        TestCStratB(const char*);
        ~TestCStratB(void);
        void run (void); /**< Runs the tests. */
        bool isTraceable (void);
        GTI_RETURN setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData);
    };
}

using namespace gti;

GTI_RETURN my_free_function (void* free_data, unsigned long num_bytes, void* buf)
{
    delete [] ((char*)buf);
    return GTI_SUCCESS;
}

mGET_INSTANCE_FUNCTION(TestCStratB);
mFREE_INSTANCE_FUNCTION(TestCStratB);
mPNMPI_REGISTRATIONPOINT_FUNCTION(TestCStratB);

extern "C" int MPI_Init (int* argc, char*** argv)
{
    //init MPI first
    int err = PMPI_Init (argc, argv);

    //Create the place with the given module configuration
    TestCStratB* place;
    place = TestCStratB::getInstance("");
    assert(place);
    place->run();

    return err;
}

extern "C" int MPI_Finalize (void)
{
    TestCStratB* place;
    place = TestCStratB::getInstance("");
    TestCStratB::freeInstanceForced(place);

    return PMPI_Finalize ();
}

TestCStratB::TestCStratB (const char* instanceName)
    : ModuleBase<TestCStratB, I_Module> (instanceName)
{
    //add the id to all sub modules
    int rank;
    MPI_Comm_rank (MPI_COMM_WORLD, &rank);
    char temp[64];
    sprintf (temp,"%d",rank);
    addDataToSubmodules ("id", temp);

    //create sub-modules
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //save sub modules
    assert (subModInstances.size() == 1);
    strat = (I_CommStrategyUp*) subModInstances[0];

    //trigger trace init
    setupTracing (GTI_TRACE_INIT_ROOT, NULL);
}

TestCStratB::~TestCStratB (void)
{
	if (strat)
	{
		strat->shutdown(GTI_FLUSH, GTI_SYNC);
		destroySubModuleInstance ((I_Module*)strat);
		strat = NULL;
	}

    OTF_Tracer::writeLeave (THREADED_MPI_PLACE_FUNC__main);
}

bool TestCStratB::isTraceable (void)
{
    return true;
}

GTI_RETURN TestCStratB::setupTracing (GTI_TRACE_INIT_TYPE init, void *pSetupData)
{
    //This is a root module, neither supports copy, nor child
    assert (init == GTI_TRACE_INIT_ROOT);

    //Configuration data
    std::map<std::string, std::string> data = getData();
    uint32_t procID = 100;
    uint32_t parent = 0; //No Parent
    const char* processName = "TestCStratB";

    //look into the module config data for data that overrides the defaults
    if (data.find("otf_proc_id") != data.end ())
        procID = atoi(data.find("otf_proc_id")->second.c_str());
    if (data.find("otf_parent") != data.end ())
        parent = atoi(data.find("otf_parent")->second.c_str());

    //Threaded MPI place adds MPI rank to proc id!
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);
    procID += 10*rank; //"*10" to have room for extra threads ...

    //Setup tracer
    OTF_Tracer::setup (procID, parent, processName);

    //Setup Functions
    OTF_Tracer::defFunction (THREADED_MPI_PLACE_FUNC__main, "main", FUNCTION_GROUP_ID__PROCESSING);

    OTF_Tracer::writeEnter (THREADED_MPI_PLACE_FUNC__main);

    //Setup sub-modules
    if (strat->isTraceable())
        strat->setupTracing(GTI_TRACE_INIT_COPY, (OTF_Tracer*)this);

    return GTI_SUCCESS;
}

void TestCStratB::run (void)
{
    unsigned long   size;
    void            *buf,
                    *free_data;
    int             flag,
                    i,
                    rank;
    char            *text;
    GTI_RETURN (*free_function) (void* free_data, unsigned long num_bytes, void* buf);

    text = new char [256];

    // (1) Receive something from T
    strat->wait(&size, &buf, &free_data,&free_function);
    printf("Got a message: {size=%ld, conntent=\"%s\"}\n",size,(char*)buf);
    free_function (free_data, size, buf);

    // (2) Test till something arrives from T
    i = 0;
    do
    {
        i++;
        strat->test(&flag,&size,&buf,&free_data,&free_function);
    } while (!flag);
    printf("Got a message after %d tests: {size=%ld, conntent=\"%s\"}\n",i,size,(char*)buf);
    free_function (free_data, size, buf);

    // (3) Send something to the master
    MPI_Comm_rank (MPI_COMM_WORLD, &rank);
    sprintf (text, "Hallo3 von %d nach oben!",rank);
    strat->send(text, strlen(text)+1, NULL, my_free_function);
}

/*EOF*/
