/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2010-2011
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2010-2011
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file test_wrappgen.cpp
 * Simple test case for trace record generation interface implementations.
 * Creates a simple source file and a CMake module to build a shared
 * library that wraps a call of the signature "mycall (int a, float* b)".
 *
 * @author Tobias Hilbrich
 */

#include <mpi.h>

#include <pnmpimod.h>
#include <iostream>
#include <fstream>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "I_RecordGenerator.h"

int main (int argc, char** argv)
{
    gti::I_RecordGenerator *pImpl;
    std::list<std::string> sources1,
        sources2, libs, headers;
    std::list<std::string>::iterator iter;

    //Get implementation of Record Generator Interface
    MPI_Init (&argc,&argv);
    MPI_Finalize();
    PNMPI_modHandle_t handle;
    PNMPI_Service_descriptor_t service;
    int err;

#ifdef PNMPI_FIXED
	err = PNMPI_Service_GetModuleByName("record_gen_implementation", &handle);
#else
	char string[512];
	sprintf (string, "record_gen_implementation");
	err = PNMPI_Service_GetModuleByName(string, &handle);
#endif

	if (err != PNMPI_SUCCESS)
    {
    	std::cerr
			<< "Failed to get implementation for record generator!"
    		<< std::endl;
    	assert (0);
    }

	err = PNMPI_Service_GetServiceByName(handle, "getInstance", "pp", &service);
    assert (err == PNMPI_SUCCESS);

	((getInstance_t) service.fct) (&pImpl, "");

    //Initialize Record Generator
    pImpl->initBegin ("/tmp", "gti-intermediate-code");
    pImpl->initGetIntermediateSourceFileNames (&sources1);
    pImpl->initGetStaticSourceFileNames (&sources2);
    pImpl->initGetLibNames (&libs);
    pImpl->initGetHeaderFileNames (&headers);
    pImpl->initEnd ();

    //Create a record
    gti::I_RecordDescription *pDesc;
    gti::I_RecordType *pRecord;
    pImpl->createRecordDescription (&pDesc);
    pDesc->addArgument ("arg1", "int", "myCall", "arg1");
    pDesc->addArrayArgument (
		"arg2", "float", "myCall", "arg2", "arg1");
    pDesc->createRecord (1234, &pRecord);

    //Get source code for some basic record usage
    std::string
        definitionCode,
        definitionCode2,
        freeCode,
        freeCode2,
        initCode,
        setArg1Code,
        setArg2Code,
        serializeCode,
        deserializeCode,
        returnArg1Code,
        returnArg2Code,
        getUidCode;
    pRecord->createInstance ("myRecord", &definitionCode);
    pRecord->initInstance ("myRecord", &initCode);
    pRecord->freeInstance ("myRecord", &freeCode);
    pRecord->writeArgument ("myRecord", "arg1", "myArg1", &setArg1Code);
    pRecord->writeArrayArgument ("myRecord", "arg2", "myArg2", &setArg2Code);
    pRecord->serialize ("myRecord", "serBuf", "serLen", &serializeCode);
    pRecord->createInstance ("myRecord2", &definitionCode2);
    pRecord->freeInstance ("myRecord", &freeCode2);
    pRecord->deserialize ("serBuf", "serLen", "myRecord2", &deserializeCode);
    pRecord->returnArgument ("myRecord2", "arg1", &returnArg1Code);
    pRecord->returnArrayArgument ("myRecord2", "arg2", &returnArg2Code);
    pImpl->returnUidFromSerialized ("serBuf", &getUidCode); //Reads Uid (1234) from serialized record.

    //Create a tiny source file that uses the record
    std::ofstream out ("test.cc");

    out << "#include <stdio.h>" << std::endl;
    out << "#include <iostream>" << std::endl;

    for (iter = headers.begin(); iter != headers.end (); iter++)
        out << "#include <" << *iter << ">" << std::endl;

    out << std::endl
        << "extern \"C\" int myCall (int myArg1, float *myArg2)" << std::endl
        << "{" << std::endl
        << "    void *serBuf = NULL;" << std::endl
        << "    unsigned long serLen;" << std::endl
        << "    " << definitionCode << ";" << std::endl
        << "    " << definitionCode2 << ";" << std::endl
        << std::endl
        << "    " << initCode << std::endl
        << "    " << setArg1Code << std::endl
        << "    " << setArg2Code << std::endl
        << "    " << serializeCode << std::endl
        << std::endl
        << "    if (" << getUidCode << " == 1234)" << std::endl
        << "    {" << std::endl
        << "        " << deserializeCode << std::endl
        << std::endl
        << "        std::cout << \"Wrapped myCall (\" << myArg1 << \", myArg2={\";" << std::endl
        << "        for (int i = 0; i < myArg1; i++)" << std::endl
        << "            std::cout << myArg2[i] << \"; \";" << std::endl
        << "        std::cout << \"})\" << std::endl;" << std::endl
        << "        std::cout << \"myRecord2 has: myArg1=\" <<"
            << returnArg1Code
            << "<< \" myArg2={\";" << std::endl
        << "        for (int i = 0; i < myArg1; i++)" << std::endl
        << "            std::cout << " << returnArg2Code << "[i] << \"; \";" << std::endl
        << "        std::cout << \"}\" << std::endl;" << std::endl
        << "    }" << std::endl
        << std::endl
        << "    free (serBuf);" << std::endl
        << "    " << freeCode << std::endl
        << "    " << freeCode2 << std::endl
        << "    return 0;" << std::endl
        << "}" << std::endl;

    //Clean up
    pRecord->deleteObject ();
    pDesc->deleteObject ();
    err = PNMPI_Service_GetServiceByName(handle, "freeInstance", "p", &service);
    assert (err == PNMPI_SUCCESS);

    ((freeInstance_t) service.fct) (pImpl);
    pImpl = NULL;

    //Write a CMake file for the generated source file
    //TODO: Missing is the include for the gti Macro file and the import
    //      of the flags, libraries, ... used to compile gti
    std::ofstream make ("CMakeLists.txt");
    make
        << "CMAKE_MINIMUM_REQUIRED(VERSION 2.6)" << std::endl
        << "PROJECT (MyCallWrapper C CXX)" << std::endl
        << "SET (SOURCES test.cc";

    for (iter = sources1.begin(); iter != sources1.end (); iter++)
        make << " " << *iter;
    for (iter = sources2.begin(); iter != sources2.end (); iter++)
        make << " " << *iter;

    make
        << ")" << std::endl
        << "ADD_LIBRARY (myCallWrapper MODULE ${SOURCES})" << std::endl
        << "SET_TARGET_PROPERTIES (myCallWrapper PROPERTIES" << std::endl
        << "    VERSION 1.0.0 SOVERSION 1" << std::endl
        << "    )" << std::endl
        << "TARGET_LINK_LIBRARIES (myCallWrapper ";

    for (iter = libs.begin(); iter != libs.end (); iter++)
        make << " " << *iter;

    make
        << ")" << std::endl
        << "INSTALL(TARGETS myCallWrapper LIBRARY DESTINATION ${PROJECT_BINARY_DIR}/lib)" << std::endl
        << "#Optional#GTI_MAC_GET_TARGET_MODULE_NAME (name myCallWrapper 1.0.0)" << std::endl
        << "#Optional#GTI_MAC_ADD_PNMPI_PATCH (${name} ${PNMPI_PATCHER})" << std::endl;

    //close files
    out.close ();
    make.close ();

    return 0;
}
